/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	BASANF:					*/
/*	Implement all routine to manage IP	*/
/*	number conversion.			*/
/*						*/
/************************************************/
#include	<errno.h>
#include	<string.h>
#include	<stdio.h>
#include	<netdb.h>
#include	<arpa/inet.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"basafn.h"
/*

*/
/************************************************/
/*						*/
/*	Routine to compare 2 IP num, can be 	*/
/*	IPV4 or IPV6 IP number.			*/
/*						*/
/************************************************/
static int cmpip(AFNTYP *ipnum1,AFNTYP *ipnum2,int mask)

{
int areequal;
int maskbyt;

areequal=false;
if (mask>128)
  mask=128;
maskbyt=mask/8;
if (memcmp(ipnum1->ip,ipnum2->ip,maskbyt)==0) {
  areequal=true;
  mask%=8;
  if (mask!=0) {
    int maskbit;
    int b1;
    int b2;

    maskbit=(~0U<<(8-mask)); 
    b1=((int)ipnum1->ip[maskbyt])&maskbit;
    b2=((int)ipnum2->ip[maskbyt])&maskbit;
    if (b1!=b2)
      areequal=false;
    }
  }
return areequal;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert a addr struct to an	*/
/*	IP number in AFNTYP structure.		*/
/*						*/
/************************************************/
AFNTYP *afn_getaddrinfo(struct addrinfo *rp)

{
AFNTYP *afnnum;

afnnum=(AFNTYP *)0;
if (rp!=(struct addrinfo *)0) {
  u_char *ptr;

  ptr=(u_char *)0;
  afnnum=calloc(1,sizeof(AFNTYP));
  afnnum->afntype=rp->ai_family;
  switch (rp->ai_family) {
    case AF_INET	:
      afnnum->afnmask=32;
      ptr=(u_char *)&(((struct sockaddr_in *)rp->ai_addr)->sin_addr.s_addr);
      break;
    case AF_INET6	:
      afnnum->afnmask=128;
      ptr=(u_char *)&(((struct sockaddr_in6 *)rp->ai_addr)->sin6_addr.s6_addr);
      break;
    default		:	/*trouble trouble	*/
      afnnum=afn_freeipnum(afnnum);
      break;
    }
  if (ptr!=(u_char *)0)
    (void) memmove(afnnum->ip,ptr,afnnum->afnmask/8);
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to convert clean string to be	*/
/*	IP number in AFNTYP structure.		*/
/*						*/
/************************************************/
AFNTYP *afn_getipnum(char *cleanipstr)

{
AFNTYP *afnnum;
int phase;
int proceed;

afnnum=(AFNTYP *)0;
phase=0;
proceed=(cleanipstr!=(char *)0);
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets say ip format ok	*/
      afnnum=(AFNTYP *)calloc(1,sizeof(AFNTYP));
      afnnum->afntype=AF_INET;
      afnnum->afnmask=32;
      if (strchr(cleanipstr,':')==(char *)0) 
	phase++;	/*must be IPV4		*/
      break;
    case 1	:	/*is it declared IPV6	*/
      if (strcasestr(cleanipstr,STRIPV6)==cleanipstr) 
	cleanipstr+=strlen(STRIPV6);
      afnnum->afntype=AF_INET6;
      afnnum->afnmask=128;
      break;
    case 2	:	/*checking '0' situation*/
      if (strcmp(cleanipstr,"0")==0) 
	cleanipstr="0.0.0.0";
      break;
    case 3	:	/*lets convert it	*/
      if (inet_pton(afnnum->afntype,cleanipstr,afnnum->ip)<=0) 
	afnnum=afn_freeipnum(afnnum);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to mask IPNUM according the 	*/
/*	number of by needed			*/
/*						*/
/************************************************/
AFNTYP *afn_maskip(AFNTYP *afnnum,int cidr)

{
if (afnnum!=(AFNTYP *)0) {
  int num;

  if (cidr>128)
    cidr=128;
  afnnum->afnmask=cidr;
  num=cidr/8;
  cidr%=8;
  if (cidr!=0) {
    afnnum->ip[num]&=(~0U<<(8-cidr));
    num++;
    }
  if (num<AFT)
    (void) memset(afnnum->ip+num,'\000',AFT-num);
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to fee memory used by an AFNTYP	*/
/*						*/
/************************************************/
AFNTYP *afn_freeipnum(AFNTYP *afnnum)

{
if (afnnum!=(AFNTYP *)0) {
  (void) free(afnnum);
  afnnum=(AFNTYP *)0;
  }
return afnnum;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to compare two ipnum by taking	*/
/*	mask ip_ipbits, return true if ipbits	*/
/*	are identical, false otherwise.		*/
/*	return -1 if unexpected type		*/
/*						*/
/************************************************/
int afn_cmpipnum(AFNTYP *afnnum1,AFNTYP *afnnum2,int mask)

{
int areequal;
int phase;
int proceed;

areequal=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*two ipnum really	*/
      if ((afnnum1==(AFNTYP *)0)||(afnnum2==(AFNTYP *)0))
	proceed=false;	/*no need to go further	*/
      break;
    case 1	:	/*is mask 0, all IP ok	*/
      if (mask==0) {
	areequal=true;
	proceed=false;
	}
      break;
    case 2	:	/*two ipnum same type	*/
      if (afnnum1->afntype!=afnnum2->afntype)
	proceed=false;	/*no need to go further	*/
      break;
    case 3	:	/*lets compare ip	*/
      switch(afnnum1->afntype) {
	case AF_INET	:
	  if (mask>32)
	    mask=32;	/*NO break		*/
	case AF_INET6	:
	  areequal=cmpip(afnnum1,afnnum2,mask);
	  break;
	default		:
	  errno=EAFNOSUPPORT;
	  areequal=-1;
	  break;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return areequal;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a string with	*/
/*	ipnum address in clear format		*/
/*	Address need to	freed.			*/
/*						*/
/************************************************/
char *afn_stripnum(AFNTYP *afnnum)

{
char *strip;

strip=(char *)0;
if (afnnum!=(AFNTYP *)0) {
  char *markr;
  char host[NI_MAXHOST];

  markr="";
  switch (afnnum->afntype) {
    case AF_INET6	:
      markr=STRIPV6;		/*NO Break	*/
    case AF_INET	:
      if (inet_ntop(afnnum->afntype,&(afnnum->ip),host,sizeof(host))!=(char *)0) 
        (void) asprintf(&strip,"%s%s",markr,host);
      else
	(void) asprintf(&strip,"[afn_stripnum error=<%s>]",strerror(errno));
      break;
    default		:
      (void) asprintf(&strip,"[ipvx:0.%d]",afnnum->afntype);
      break;
    }
  }
return strip;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a revers address as	*/
/*	dynamically allocated string		*/
/*						*/
/************************************************/
char *afn_reversipnum(AFNTYP *afnnum)

{
char * reversip;

reversip=(char *)0;
if (afnnum!=(AFNTYP *)0) {
  int i;

  reversip=strdup("");
  switch (afnnum->afntype) {
    case AF_INET6	:
      for (i=16;i>0;i--) {
	char *ptr;

	ptr=reversip;
	(void) asprintf(&reversip,"%s%x.%x.",ptr,afnnum->ip[i-1]&0xf,afnnum->ip[i-1]>>4);
	(void) free(ptr);
	}
      break;
    case AF_INET	:
      for (i=4;i>0;i--) {
	char *ptr;

	ptr=reversip;
	(void) asprintf(&reversip,"%s%d.",ptr,afnnum->ip[i-1]);
	(void) free(ptr);
	}
      break;
    default		:
      (void) asprintf(&reversip,"%d.0.0.0.0.ukn.",afnnum->afntype);
      break;
    }
  reversip[strlen(reversip)-1]='\000';
  }
return reversip;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a revers generic	*/
/*	domain related to ipnum			*/
/*	dynamically allocated string		*/
/*						*/
/************************************************/
const char *afn_reversdom(AFNTYP *afnnum)

{
const char *reversdom;

reversdom=(const char *)0;
if (afnnum!=(AFNTYP *)0) {
  switch (afnnum->afntype) {
    case AF_INET6	:
      reversdom="ip6.arpa";
      break;
    case AF_INET	:
      reversdom="in-addr.arpa";
      break;
    default		:
      reversdom="ukn.arpa";
      break;
    }
  }
return reversdom;
}
