/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	COMEML:					*/
/*	Define all routine to manage email 	*/
/*	communication exchange.			*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unijrl.h"
#include	"comeml.h"

#define MXTIME	300	/*wait for a 300 sec max*/
#define	CRLF	"\r\n"
/*

*/
/************************************************/
/*						*/
/*	Procedure to display connection status	*/
/*	if doing command 'ps ax' from console	*/
/*						*/
/************************************************/
static void emlsettitle(CTTYP *contact)

{
(void) prc_settitle("%s: Reading on %s:%s since %s, from '%s/[%s]'",
                    appname,contact->locip,contact->servport,contact->attime,
		    contact->peername,contact->peerip);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a contact structure*/
/*						*/
/************************************************/
static CTTYP *setcontact(TLSFILE *tls,PROTYP proto,char *port)

{
#define	MXWAIT	300	/*contact maximum idle time	*/
#define	OPSE	"comeml.c:setcontact,"
CTTYP *contact;

contact=(CTTYP *)calloc(1,sizeof(CTTYP));
contact->tls=tls;
contact->locname=tls_getaddrname(tls,pr_locname,"NO_LOCAL_HOST_NAME");
contact->locip=tls_getaddrname(tls,pr_locip,"?.?.?.?");
contact->servport=strdup(port);
contact->peername=tls_getaddrname(tls,pr_remname,NORVERS);
contact->peerip=tls_getaddrname(tls,pr_remip,"?.?.?.?");
contact->proto=proto;
contact->sequence=0;
contact->lastdata=time((time_t)0);
contact->start=contact->lastdata;
contact->stop=contact->lastdata;
contact->attime=strdup(clm_ascsystime(contact->lastdata));
contact->mtaille=(u_long)0;
contact->maxwait=MXWAIT;
contact->founddata=false;
contact->session=clm_uniquename(contact->sequence);
if (debug>0) {
  char *dbgname;
  FILE *dbgfile;

  dbgname=clm_uniquename((unsigned int)0);
  contact->bckstdout=stdout;
  (void) asprintf(&(contact->dbgname),"%s/%s.dbgs",clm_appdir(d_dbgs),dbgname);
  if ((dbgfile=fopen(contact->dbgname,"w"))==(FILE *)0)
    (void) clm_core_dump("%s Unable to fopen file <%s> (error=<%s>)",
		          OPSE,contact->dbgname,strerror(errno));
  stdout=dbgfile;	/*now STDOUT is the file	*/
  if (debug>3) {
    char *strloc;

    (void) tls_trace(tls,true);
    (void) asprintf(&strloc,"Opening session '%s'\n",contact->session);
    (void) tls_traceout(contact->tls,"#--",strloc,strlen(strloc));
    (void) free(strloc);
    }
  dbgname=clm_freestr(dbgname);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	contact structure.			*/
/*						*/
/************************************************/
static CTTYP *freecontact(CTTYP *contact)

{
if (contact->dbgname!=(char *)0) {
  if (fclose(stdout)<0)
    (void) clm_core_dump("%s Unable to fclose stdout (error=<%s>)",
		          "comeml.c:freecontact",strerror(errno));
  stdout=contact->bckstdout;
  if (debug<3)
    (void) unlink(contact->dbgname);
  (void) free(contact->dbgname);
  }
contact->authname=clm_freestr(contact->authname);
contact->attime=clm_freestr(contact->attime);
contact->info=clm_freestr(contact->info);
contact->cipherid=clm_freestr(contact->cipherid);
contact->peerip=clm_freestr(contact->peerip);
contact->servport=clm_freestr(contact->servport);
contact->peername=clm_freestr(contact->peername);
contact->heloname=clm_freestr(contact->heloname);
contact->locip=clm_freestr(contact->locip);
contact->locname=clm_freestr(contact->locname);
contact->session=clm_freestr(contact->session);
(void) free(contact);
contact=(CTTYP *)0;
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a string ended with	*/
/*	CRLF on the TCP line			*/
/*						*/
/************************************************/
int printfcrlf(CTTYP *contact,const char *fmt,va_list args)

{
va_list ap;
int written;
int taille;
char stdq[500];

va_copy(ap,args);

if ((taille=vsnprintf(stdq,sizeof(stdq),fmt,ap))>(sizeof(stdq)-5)) {
  /*case of a very long formated line		*/
  char *local;		

  taille=vasprintf(&local,fmt,args);
  local=(char *)realloc(local,taille+10);
  (void) strcpy(local+taille,CRLF);
  taille+=strlen(CRLF);
  written=tls_write(contact->tls,local,taille);
  (void) free(local);
  }
else {
  /*usual length  write				*/
  (void) strcpy(stdq+taille,CRLF);
  taille+=strlen(CRLF);
  written=tls_write(contact->tls,stdq,taille);
  }
va_end(ap);
return written;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an "HELO" message	*/
/*	if EHLO message is not accepted.	*/
/*						*/
/************************************************/
static int dohelo(CTTYP *contact,char *hostname)

{
int done;
char *buffer;

done=false;
(void) eml_printfcrlf(contact,HELO,contact->locname);
if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))!=(char *)0) {
  int numret;

  (void) clm_cleanstring(buffer);
  if (sscanf(buffer,"%d",&numret)==1) {
    done=true;
    if (numret!=CMDOK) {
      (void) jrl_journalling(false,EVTRT,"Unable to carry HELO with '%s' returned status=<%s>",
		         		  hostname,buffer);
      done=false;
      }
    }
  (void) free(buffer);
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an "EHLO" message	*/
/*	an collect working parameters		*/
/*						*/
/************************************************/
static int doehlo(CTTYP *contact,char *hostname)

{
#define	OPEH	"comeml.c:doehlo,"

typedef	enum	{
	h_tls,		/*STARTTLS capable	*/
	h_size,		/*SIZE capable		*/
	h_unknown	/*unknown code		*/
	}HTYP;

static VOCTYP vochelo[]={
        {h_tls,STRTTLS,(const void *)0},
        {h_size,MSIZE,(const void *)0},
        {h_unknown,(char *)0,(const void *)0}
        };

int done;
int numret;
int sofar;
char *buffer;
int phase;
int proceed;

done=true;
numret=0;
sofar=0;
buffer=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending EHLO		*/
      (void) eml_printfcrlf(contact,EHLO,contact->locname);
      if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))==(char *)0) {
	done=false;
	phase=999;	/*trouble! no answer	*/
	}
      break;
    case 1	:	/*getting the answer	*/
      if (sscanf(buffer,"%d%n",&numret,&sofar)!=1) {
        (void) clm_alert(0,"%s (pid=%05d) Unable to carry EHLO with '%s'",
		           OPEH,getpid(),hostname);
        done=false;
        phase=999;	/*trouble! no answer	*/
	(void) free(buffer);
	}
      break;
    case 2	:	/*proper answer		*/
      phase++;
      if (numret!=CMDOK) {
	(void) free(buffer);
	done=dohelo(contact,hostname);/*backup solution	*/
	phase=999;
	}
      break;
    case 3	:	/*lets wait full answer	*/
      if ((buffer=tls_waitforaction(contact->tls,contact->maxwait))==(char *)0) {
	done=false;
	phase=999;	/*trouble! no answer	*/
	}
      break;
    case 4	:	/*lets scanreply	*/
      if (buffer!=(char *)0) {	/*always	*/
	VOCTYP *voc;

        if (buffer[sofar]=='-') { /*EHLO next	*/
	  phase--;	/*next wait for anwser	*/
	  phase--;
	  }
        (void) memmove(buffer,buffer+sofar+1,strlen(buffer+sofar+1)+1);
        (void) clm_cleanstring(buffer);
        if ((voc=clm_getvoca(vochelo,buffer))!=(VOCTYP *)0) {
	  switch (voc->code) {
	    case h_tls	:
	      contact->capable|=CPTLS;
	      break;
	    case h_size	:
	      contact->capable|=CPSIZE;
	      break;
	    default	:
	      break;
	    }
	  }
	(void) free(buffer);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	advice request or a take note message.	*/
/*						*/
/************************************************/
static char *ctttomsg(register CTTYP *cta,char *msghdr)

{
char *advmsg;
char *from;

advmsg=(char *)0;
from=cta->frm->efrom;
if ((cta->evl!=(EVLTYP *)0)&&(cta->evl->realfrom!=(char *)0))
  from=cta->evl->realfrom;
(void) clm_fasprintf(&advmsg,"%s %s\n",msghdr,cta->session);
(void) clm_fasprintf(&advmsg,"%s %s %s %lu %d\n",SORIG,
			     cta->peerip,from,cta->mtaille,cta->relayable);
if (cta->evl!=(EVLTYP *)0) {
  if (cta->evl->subject!=(char *)0)
    (void) clm_fasprintf(&advmsg,"%s %s\n",SSUBJ,cta->evl->subject);
  if (cta->relayable==false)
    (void) clm_fasprintf(&advmsg,"%s %s\n",SSPM,cta->evl->spamlvl);
  }
if (cta->virname!=(char *)0) 
  (void) clm_fasprintf(&advmsg,"%s %s\n",CHKVIR,cta->virname);
if (cta->rcpts!=(RCPTYP **)0) {
  int i;
  RCPTYP **rcploc;

  i=0;
  rcploc=cta->rcpts;
  while (*rcploc!=(RCPTYP *)0) {
    (void) clm_fasprintf(&advmsg,"%s %d %s\n",SCHK,(*rcploc)->status,(*rcploc)->rcpt);
    rcploc++;
    i++;
    }
  }
return advmsg;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	take note message.			*/
/*						*/
/************************************************/
char *eml_cttotknmsg(register CTTYP *cta)

{
return ctttomsg(cta,TKNOTE);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert CTTYP data to an	*/
/*	advice request message.			*/
/*						*/
/************************************************/
char *eml_cttoadvmsg(register CTTYP *cta)

{
return ctttomsg(cta,GVADV);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert an advice answer	*/
/*	message to an RCPT structure		*/
/*						*/
/************************************************/
void eml_advmsgupdrcpt(char *advice,RCPTYP **rcpts)

{
#define	OPAD	"comeml.c:eml_advmsgupdrcpt,"

char **split;
int phase;
int proceed;

split=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*data to scan ?	*/
      if ((advice==(char *)0)||((split=clm_splitstring(advice,'\n'))==(char **)0))
	phase=999;	/*Humm... strange	*/
      break;
    case 1	:	/*proper data		*/
      if (strstr(split[0],MYADV)!=split[0]) {
	(void) clm_alert(0,"%s Unexpected advice <%s>",OPAD,advice);
	phase++;	/*NO! scan		*/
	}
      break;
    case 2	:	/*scanning given advice	*/
      if (split!=(char **)0) {	/*always	*/
        char **ptr;

        ptr=split;
	while (*ptr!=(char *)0) {
	  if (strstr(*ptr,SCHK)==*ptr) {
	    int ind;
	    int status;

	    ind=0;
	    status=0;
	    if (sscanf(*ptr,"%*s %d %d",&ind,&status)==2) {
	      rcpts[ind]->status=status;
	      } 
	    } 
          ptr++;
	  }
        }
      break;
    case 3	:	/*cleanning act		*/
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a string to remote	*/
/*	SMTP channel but add an CRLF at the	*/
/*	of the line.				*/
/*						*/
/************************************************/
int eml_printfcrlf(CTTYP *contact,const char *fmt,...)

{
va_list args;
int written;

va_start(args,fmt);
written=printfcrlf(contact,fmt,args);
va_end(args);
return written;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage CMD exhange with	*/
/*	a remote SMTP server.			*/
/*						*/
/************************************************/
int eml_sendcmd(CTTYP *contact,int expected,char *fmt,...)

{
#define CMDMRK	"comeml.c:eml_sendcmd, "

#define	DBGCM	5	/*routine debug level 		*/

int status;
va_list args;

status=ERRPROC;
va_start(args,fmt);
if (printfcrlf(contact,fmt,args)>0) {
  contact->info=clm_freestr(contact->info);
  status=tls_getreply(contact->tls,contact->maxwait,pro_smtp,&(contact->info));
  switch (status) {
    case ERRPROC	:
      if (sigterm==true) {
	status=CMDOK;	/*status override	*/
	break;		/* it was a TERM signal	*/
	}
      if (contact->info==(char *)0)
        contact->info=eml_ascstat(status,"Reply processing error");
      if (tls_iseof(contact->tls)==false) {
	contact->info=clm_freestr(contact->info);
	(void) asprintf(&(contact->info),"no answer from smtp server <%s/%s:%s>",
				         contact->peername,contact->peerip,
					 contact->servport);
	}
      (void) clm_alert(0,"%05d %s ERRPROC: %s",getpid(),CMDMRK,contact->info);
      break;
    default		:
      contact->lastdata=time((time_t *)0);
      if ((status!=expected)&&(debug>=DBGCM)) {
	char *local;

	(void) vasprintf(&local,fmt,args);
        (void) clm_alert(DBGCM,"%05d %s Unexpected answer <%s> to command <%s> from <%s>",
			       getpid(),CMDMRK,contact->info,local,contact->peername);
	(void) free(local);
	}
      break;
    }
  }
va_end(args);
return status; 
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open and SMTP connection	*/
/*	according server information		*/
/*						*/
/************************************************/
CTTYP *eml_opencontact(PROTYP proto,char *hostname,char *port,int maxwait)

{
#define	OPSR	"comeml.c:eml_opencontact,"

CTTYP *cta;
TLSFILE *tls;

cta=(CTTYP *)0;
if ((tls=tls_opentcp(hostname,port,maxwait))!=(TLSFILE *)0) {
  int proceed;
  int phase;

  phase=1;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*trouble, trouble		*/
        cta=eml_lostcontact(cta);
	proceed=false;
        break;
      case 1	:	/*Lets prepare contact info	*/
	if ((cta=setcontact(tls,proto,port))!=(CTTYP *)0) {
	  char *strloc;

	  (void) asprintf(&strloc,"Contact established with <%s[%s]:%s>, waiting '%d' seconds",
				   cta->peername,cta->peerip,port,maxwait);
	  (void) tls_traceout(cta->tls,"#--",strloc,strlen(strloc));
	  (void) free(strloc);
	  cta->maxwait=maxwait;
	  }
        break;
      case 2	: 	/*Lets get first prompt		*/
        if (tls_waitsignon(cta->tls,cta->maxwait,proto,&(cta->info))!=true) {
          (void) clm_alert(0,"%s Invalid prompt from '%s' (prompt=<%s>)",
			     OPSR,hostname,cta->info);
	  phase=999;	/*contact not successful	*/
	  }
        break;
      case 3	:	/*doing Hello if needed		*/
	switch (proto) {
	  case (pro_ment)	:	/*Mentor HELO	*/
	  case (pro_smtp)	:	/*SMTP HELO	*/
	    if (doehlo(cta,hostname)==false) {
	      phase=999;/*trouble trouble		*/
	      }
	    break;
	  default		:	/*nothing to do	*/
	    break;
	  }
	break;
      case 4	:	/*if remote system is TLS, lets go TLS	*/
	if ((cta->capable&CPTLS)!=0) {
	  int statusloc;

	  statusloc=eml_sendcmd(cta,SIGNON,STRTTLS);
	  switch (statusloc) {
	    case SIGNON		:
	      (void) tls_gotls(cta->tls,false);
	      switch (proto) {
		case (pro_smtp)	:
	          if (doehlo(cta,hostname)==false) {
	      	    phase=999;		/*big trouble	*/
		    }
		  break;
		default		:	/*nothing to do	*/
		  break;
		}
	      break;
	    case TLSNRDY	:	/*TLS not ready	*/
	      break;
	    default		:
	      phase=999;		/*big trouble	*/
	      break;
	    }
	  }
        break;
      case 5	:	/*contact well established	*/
	cta->established=true;
        break;
      default	:	/*everything well, lets go	*/
	proceed=false;
        break;
      }
    phase++;
    }
  }
return cta;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to terminate an SMTP connexion*/
/*						*/
/************************************************/
CTTYP *eml_closecontact(CTTYP *contact)

{
if (contact!=(CTTYP *)0) {
  (void) eml_sendcmd(contact,QUITOK,SQUIT);
  (void) usleep(10000);
  contact=eml_lostcontact(contact);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait and acknowledge	*/
/*	contact from a remote site.		*/
/*						*/
/************************************************/
CTTYP *eml_acceptcontact(LSTTYP *channel)

{
#define	OPTR	"comeml.c:eml_acceptcontact,"

CTTYP *contact;
TLSFILE *tls;

contact=(CTTYP *)0;
if ((tls=tls_accept(channel))!=(TLSFILE *)0) {
  contact=setcontact(tls,channel->binding->proto,channel->binding->port);
  switch (channel->binding->proto) {
    case pro_smtp	:
      (void) eml_printfcrlf(contact,"%d %s %s %s-%s; %s",
                            SIGNON,contact->locname,"ESMTP",appname,
                            clm_getvers(),clm_ascsysstamp(time((time_t *)0)));
      (void) emlsettitle(contact);
      break;
    case pro_pop3	:
      contact=eml_lostcontact(contact);	
      (void) clm_alert(0,"%s proto pop3 not yet implemented",OPTR);
      break;
    case pro_ment	:
      (void) eml_printfcrlf(contact,"%d %s %s %s-%s; %s",
                            SIGNON,contact->locname,"MENTOR",appname,
                            clm_getvers(),clm_ascsysstamp(time((time_t *)0)));
      (void) emlsettitle(contact);
      break;
    case pro_ukn	:
      (void) clm_alert(0,"%s undefined protocol",OPTR);
      contact=eml_lostcontact(contact);	
      break;
    }
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close a contact with a 	*/
/*	remote site.				*/
/*						*/
/************************************************/
CTTYP *eml_lostcontact(CTTYP *contact)

{
if (contact!=(CTTYP *)0) {
  (void) tls_close(contact->tls);
  contact=freecontact(contact);
  (void) fclose(stdout);
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to format the status report	*/
/*	conveniently for clement logs		*/
/*						*/
/************************************************/
char *eml_forlogs(char *logs,char *note)

{
char *forlogs;
char **lines;

forlogs=(char *)0;
lines=(char **)0;
(void) clm_cutchar(logs,'\r');
if ((lines=clm_splitstring(logs,'\n'))!=(char **)0) {
  int i;

  for (i=0;lines[i]!=(char *)0;i++) {
    if (forlogs!=(char *)0)
      forlogs=clm_astrcat(forlogs,"\n\t  ");
    forlogs=clm_astrcat(forlogs,lines[i]);
    }
  lines=(char **)vec_freelstlst((void **)lines,VECFREE);
  }
if (note!=(char *)0) {
  forlogs=clm_astrcat(forlogs,"\n\t  Note: ");
  forlogs=clm_astrcat(forlogs,note);
  }
return forlogs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a "MAIL FROM:" line	*/
/*	to a remote SMTP host.			*/
/*						*/
/************************************************/
int eml_sendmfrom(CTTYP *contact,char *originator,u_long size)

{
char *fmsm;

fmsm=FMSM;
if ((contact->capable&CPSIZE)!=0)
  fmsm=FMSM " %s=%lu";
return eml_sendcmd(contact,CMDOK,fmsm,MAILF,originator,MSIZE,size);
}
