/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Implement very sub level procedure to 	*/
/*	debug potential memory leak (or other)	*/
/*	trouble.				*/
/*						*/
/************************************************/
#include	<malloc.h>
#include	<signal.h>
#include	<errno.h>
#include	<search.h>
#include	<stdio.h>
#include	<unistd.h>
#include	<time.h>
#include	<syslog.h>
#include	<execinfo.h>
#include	<sys/time.h>
#include	<memory.h>
#include	"dbgmem.h"

#undef	calloc
#undef	malloc
#undef	realloc
#undef	free
#undef	putenv
#undef	setenv
#undef	unsetenv
#undef	strdup
#undef	asprintf
#undef	vasprintf


#define MXTB	10	/*max trace deep		*/
#define	MXD	200	/*maximun path size		*/

extern char **environ;	/*application env varirable	*/


typedef	struct timeval	TIMEVAL;

typedef struct 	{	/*leak detection structure	*/
	void *ptr;	/*allocated memory		*/
	TIMEVAL date;	/*allocation date		*/
	size_t size;	/*memory size			*/
	void *tb[MXTB]; /*back trace			*/
	int noleak;	/*No leak consideration flag	*/
	}LEAKTYP;

int memleak=DMFALSE;	/*flag to check for memory leak	*/

static void *memory;	/*allocated memory structure	*/
static u_long totmem;	/*total requested memory size	*/
static u_long nummem;	/*number of leak structure	*/
static int flgleak;	/*leak mod flag			*/
static u_long collected;/*number of memory still open	*/
static LEAKTYP **memsort;
static char cpath[MXD];	/*Application curpath		*/
static char cbase[MXD];	/*Application base		*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	ptr, used by search routine		*/
/*						*/
/************************************************/
static int cmpptr(const void *v1,const void *v2)

{
register u_long p1;
register u_long p2;

if (v1==(void *)0) {
  if (v2==(void *)0)
    return 0;
  else
    return -1;
  }
if (v2==(void *)0)
  return 1;
p1=(u_long)(((LEAKTYP *)v1)->ptr);
p2=(u_long)(((LEAKTYP *)v2)->ptr);
if (p1>p2)
  return 1;
if (p1<p2)
  return -1;
return 0;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to flag the leak	*/
/*	value withing collected open memory.	*/
/*						*/
/************************************************/
static void setleakflg(const void *v1,const VISIT which,const int depth)

{
switch (which) {
  case preorder:	/*no remote dump	*/
  case endorder:
    break;
  case postorder:	/*remote dump		*/
  case leaf:
    (*((LEAKTYP **)v1))->noleak=flgleak;
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to count collected "leaking"	*/
/*	memory.					*/
/*						*/
/************************************************/
static void collectleakptr(const void *v1,const VISIT which,const int depth)

{
switch (which) {
  case preorder:	/*no remote dump	*/
  case endorder:
    break;
  case postorder:	/*remote dump		*/
  case leaf:
    if ((*((LEAKTYP **)v1))->noleak==0) {
      memsort[collected]=*((LEAKTYP **)v1);
      collected++;
      }
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information, from the backtrace stand	*/
/*	point.					*/
/*						*/
/************************************************/
static int sametrace(const LEAKTYP *l1,const LEAKTYP *l2)

{
int inx;
long diff;

inx=0;
diff=0;
while ((inx<MXTB)&&(diff==(long)0)) {
  diff=(long)(l1->tb[inx])-(long)(l2->tb[inx]);
  inx++;
  }
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare to "leaking" memory*/
/*	information				*/
/*						*/
/************************************************/
static int cmpleak(const void *p1,const void *p2)

{
int diff;
LEAKTYP *l1;
LEAKTYP *l2;

l1=((LEAKTYP **)p1)[0];
l2=((LEAKTYP **)p2)[0];
if ((diff=sametrace(l1,l2))==0)
  diff=l1->date.tv_sec-l2->date.tv_sec;
  if (diff==0)
    diff=l1->date.tv_usec-l2->date.tv_usec;
return diff;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert back trace address	*/
/*	to sources line information.		*/
/*						*/
/************************************************/
static void disline(LEAKTYP *leak,const char *path,FILE *fichier)

{
#define ADDR2LINE	"addr2line"

int i;
FILE *canal;
char command[1000];

(void) sprintf(command,"%s -e %s",ADDR2LINE,path);
for (i=0;i<MXTB;i++) {
  char str[20];

  if (leak->tb[i]==0)
    break;
  (void) sprintf(str," %08lx",(long)leak->tb[i]);
  (void) strcat(command,str);
  }
if ((canal=popen(command,"r"))!=(FILE *)0) {
  char line[200];

  while (fgets(line,sizeof(line)-1,canal)!=(char *)0) {
    (void) fprintf(fichier,"%s",line);
    }
  (void) fclose(canal);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne a Backtrace stack	*/
/*	(to trace back leak) to memory structure*/
/*						*/
/************************************************/
static void setbtrace(LEAKTYP *leak,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i,j;
struct timezone tz;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (j=0,i=step+1;(j<MXTB)&&(i<num);i++,j++) {
  leak->tb[j]=zone[i];
  }
(void) gettimeofday(&(leak->date),&tz);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report backtrace via syslog*/
/*						*/
/************************************************/
static void logbacktrace(int loglev,int step)

{
#define MXSTEP	10

void *zone[MXSTEP];
int num;
int i;

(void) memset(zone,'\000',sizeof(zone));
num=backtrace(zone,MXSTEP);
for (i=step;i<num;i++) {
  (void) syslog(loglev,"\tbt--> %08lx",(long)zone[i]);
  }
}
/*

*/
/************************************************/
/*						*/
/*	procedure to get a "leak proof" memory	*/
/*	email to a shell script			*/
/*						*/
/************************************************/
static void *getmemory(size_t size,int step)

{
register LEAKTYP *leak;
register void **leakptr;

leak=(LEAKTYP *)calloc(1,sizeof(LEAKTYP));
leak->size=size;
leak->ptr=malloc(size);
(void) setbtrace(leak,step+1);
leakptr=tsearch((void *)leak,&memory,cmpptr);
if (*leakptr!=leak) {
  (void) syslog(LOG_DEBUG,"getmemory trouble PTR already within tsearch");
  }
totmem+=size;
nummem++;
return leak->ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the total number of	*/
/*	memory requested.			*/
/*						*/
/************************************************/
u_long dbg_gettotmem()

{
return totmem;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the total number of	*/
/*	memory segement requested.		*/
/*						*/
/************************************************/
u_long dbg_getnummem()

{
return nummem;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application path, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void dbg_setcurpath(char *path)

{
(void) memset(cpath,'\000',sizeof(cpath));
(void) strncpy(cpath,path,sizeof(cpath)-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to memorise the current	*/
/*	application base, will be used to 	*/
/*	dump the memory stat.			*/
/*						*/
/************************************************/
void dbg_setcurbase(char *path)

{
(void) memset(cbase,'\000',sizeof(cbase));
(void) strncpy(cbase,path,sizeof(cbase)-1);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne new memory area	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_malloc(size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(size,sizeof(char));
  }
else
  ptr=malloc(size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to assigne clean new memory 	*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_calloc(size_t nmemb,size_t size)

{
void *ptr;

if (memleak==DMTRUE) {
  ptr=getmemory(nmemb*size,sizeof(char));
  (void) memset(ptr,'\000',nmemb*size);
  }
else
  ptr=calloc(nmemb,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory from memry pool*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void dbg_free(void *ptr)

{
if (memleak==DMTRUE) {
  LEAKTYP **leakpt;
  LEAKTYP leak;

  leak.ptr=ptr;
  if ((leakpt=(LEAKTYP **)tfind(&leak,&memory,cmpptr))!=(LEAKTYP **)0) {
    LEAKTYP *lk;

    lk=*leakpt;
    (void) tdelete(*leakpt,&memory,cmpptr);
    (void) free(lk->ptr);
    totmem-=lk->size;
    (void) free(lk);
    nummem--;
    }
  else {
    char *strloc="subsys.c:dbg_free, Unable to find 'leak memory' "
		 "for ptr '%08lx'=<%s>\n";

    (void) syslog(LOG_INFO,strloc,(long)ptr,(char *)ptr);
    (void) setbtrace(&leak,0);
    (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
    (void) disline(&leak,cpath,stderr);
    (void) fprintf(stderr,"\n");
    }
  }
else
  (void) free(ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a variable from the	*/
/*	environement list			*/
/*						*/
/************************************************/
int dbg_unsetenv(const char *name)

{
int status;

status=-1;
errno=EINVAL;
if (memleak==DMTRUE) {
  if (environ!=(char **)0) {
    int i;

    for (i=0;environ[i]!=(char *)0;i++) {
      char *ptr;

      if ((ptr=strstr(environ[i],name))==(char *)0)
	continue;
      if (*(environ[i]+strlen(name))=='=') {
	(void) dbg_free(ptr);
        do {
          environ[i]=environ[i+1];
	  i++;
	  }
        while (environ[i]!=(char *)0);
	status=0;
	break;
	}
      }
    }
  }
else 
  status=unsetenv(name);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to install variable within the*/
/*	environnement				*/
/*						*/
/************************************************/
int dbg_putenv(char *valeur)

{
int status;

status=-1;
if (memleak==DMTRUE) {
  register char *ptr;
 
  if ((ptr=strchr(valeur,'='))!=(char *)0) {
    register int found;
    register int taille;
    register int tocheck;

    found=0;
    taille=0;
    if ((tocheck=(ptr-valeur))<=0)
      return status;
    status=0;
    if (environ!=(char **)0) {
      for (;environ[taille]!=(char *)0;taille++) {
        if (strncmp(valeur,environ[taille],tocheck)==0) {
	  (void) dbg_free(environ[taille]);
	  environ[taille]=valeur;
	  found=1;
	  break;
	  }
	}
      }
    else
      environ=dbg_calloc(1,sizeof(char *));
    if (found==0) {
      environ=(char **)dbg_realloc((void *)environ,(taille+2)*sizeof(char *));
      environ[taille]=valeur;
      environ[taille+1]=(char *)0;
      }
    }
  }
else {
  status=putenv(valeur);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set variable within the	*/
/*	environement				*/
/*						*/
/************************************************/
int dbg_setenv(const char *name,const char *value,int overwrite)

{
int status;

status=0;
if (memleak==DMTRUE) {
  char *newvar;
  char *ptr;

  (void) dbg_asprintf(&newvar,"%s=%s",name,value);
  if ((ptr=getenv(name))!=(char *)0) {
    if (overwrite!=0) 
      (void) dbg_unsetenv(name);
    else {
      (void) dbg_free(newvar);
      newvar=(char *)0;
      }
    }
  if (newvar!=(char *)0) {
    (void) dbg_putenv(newvar);
    }
  }
else {
  status=setenv(name,value,overwrite);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reallocate memory from pole*/
/*	while checking (if needed) for memory	*/
/*	leakage.				*/
/*						*/
/************************************************/
void *dbg_realloc(void *ptr,size_t size)

{
if (memleak==DMTRUE) {
  if (ptr==(void *)0)
    ptr=getmemory(size,sizeof(char));
  else {
    LEAKTYP **leakpt;
    LEAKTYP leak;

    leak.ptr=ptr;
    if ((leakpt=(LEAKTYP **)tfind(&leak,&memory,cmpptr))!=(LEAKTYP **)0) {
      LEAKTYP *lk;

      lk=*leakpt;
      totmem-=lk->size;
      (void) tdelete(*leakpt,&memory,cmpptr);
      ptr=realloc(ptr,size);
      lk->ptr=ptr;
      lk->size=size;
      totmem+=lk->size;
      (void) setbtrace(lk,1);
      (void) tsearch(lk,&memory,cmpptr);
      }
    else {
      char *strloc="subsys.c:dbg_realloc, Unable to find 'leak memory' "
		   "for ptr '%08lx'=<%s>\n";
      (void) fprintf(stderr,strloc,(long)ptr,(char *)ptr);
      (void) syslog(LOG_DEBUG,strloc,(long)ptr,(char *)ptr);
      (void) logbacktrace(LOG_DEBUG,2);
      ptr=(void *)0;
      }
    }
  }
else
  ptr=realloc(ptr,size);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string.			*/
/*						*/
/************************************************/
char *dbg_strdup(const char *str)

{
char *ptr;

if (str==(char *)0) {
  /*NULL PTR to duplicate?! crashing HARD	*/
  (void) kill(getpid(),SIGSEGV);
  }
if (memleak==DMTRUE) {
  register int taille;

  taille=strlen(str)+1;
  ptr=getmemory(taille,1);
  (void) memmove(ptr,str,taille);
  }
else
  ptr=strdup(str);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate memory to		*/
/*	duplicate a string but only for a max	*/
/*	number of character.			*/
/*						*/
/************************************************/
char *dbg_strndup(const char *str,size_t n)

{
char *ptr;

if (str==(char *)0) {
  /*NULL PTR to duplicate?! crashing HARD	*/
  (void) kill(getpid(),SIGSEGV);
  }
if (memleak==DMTRUE) {
  register int taille;

  taille=strlen(str);
  if (taille>n)
    taille=n;
  ptr=getmemory(taille+1,sizeof(char));
  (void) memmove(ptr,str,taille);
  ptr[taille+1]='\000';
  }
else
  ptr=strndup(str,n);
return ptr;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory using a va_list	*/
/*						*/
/************************************************/
int dbg_vasprintf(char **strp, const char *fmt,va_list args)

{
int count;

count=-1;
if (memleak==DMTRUE) {
  char *strloc;

  count=vasprintf(&strloc,fmt,args);
  *strp=dbg_strdup(strloc);
  (void) free(strloc);
  }
else {
  count=vasprintf(strp,fmt,args);
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to allocate a string within	*/
/*	a dynamic memory			*/
/*						*/
/************************************************/
int dbg_asprintf(char **strp, const char *fmt, ...)

{
int count;
va_list args;

count=-1;
va_start(args,fmt);
if (memleak==DMTRUE)
  count=dbg_vasprintf(strp,fmt,args);
else 
  count=vasprintf(strp,fmt,args);
va_end(args);
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory allocated by	*/
/*	a library procedure (as scandir)	*/
/*	Memory allocated by such library routine*/
/*	are not known by leak detection.	*/
/*						*/
/************************************************/
void dbg_sysfree(void *ptr)

{
(void) free(ptr);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to flag current memory as not	*/
/*	to be checked against memory leak	*/
/*	Implementation is such memory can be	*/
/*	flagged/unflagged as a no memory leak.	*/
/*	noleak value should always be positive.	*/
/*						*/
/************************************************/
void dbg_setnoleak(register int value)

{
flgleak=value;
(void) twalk(memory,setleakflg);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dump current memory	*/
/*	allocation and detect memory allocation	*/
/*						*/
/*	Data will be dumped on /tmp under	*/
/*	application name and process pid.	*/
/*						*/
/************************************************/
void dbg_dumpmem(const char *apname,const char *extension)

{
#define	DUMPFRM	"%s/var/tmp/%s%s%s.%05d"

if (memleak==DMTRUE) {
  char *sep;
  FILE *fichier;
  char *name;
  LEAKTYP *leak;

  sep="-";
  collected=0;
  if ((extension==(char *)0)||(strlen(extension)==0)) {
    sep="";
    extension="";
    }
  (void) asprintf(&name,DUMPFRM,cbase,apname,sep,extension,getpid());
  if ((fichier=fopen(name,"w"))==(FILE *)0) {
    (void) fprintf(stderr,"dbg_dumpmem unable to open file <%s> (error=<%s>)\n",
			  name,strerror(errno));
    fichier=stderr;
    }
  memsort=(LEAKTYP **)0;
  if (nummem>0) {
    memsort=(LEAKTYP **)calloc(nummem,sizeof(LEAKTYP *));
    (void) twalk(memory,collectleakptr);
    }
  if (collected>0) {
    register int i;

    (void) fprintf(fichier,"%05ld memory allocation still open\n",collected);
    (void) qsort(memsort,collected,sizeof(LEAKTYP *),cmpleak);
    leak=(LEAKTYP *)0;
    for (i=0;i<collected;i++) {
      #define STRTIME	"%Y-%m-%d %H:%M:%S"
      struct tm *tm;
      char strtime[40];

      if ((leak==(LEAKTYP *)0)||(sametrace(leak,memsort[i])!=0)) {
        (void) fprintf(fichier,"\n");
	leak=memsort[i];
	(void) disline(leak,cpath,fichier);
	}
      tm=localtime(&(memsort[i]->date.tv_sec));
      (void) strftime(strtime,sizeof(strtime),STRTIME,tm);
      (void) fprintf(fichier,"\t size=%05ld age='%s.%06ld' ptr=%08lx->'%s'\n",
		     	     (long)(memsort[i]->size),
			     strtime,memsort[i]->date.tv_usec,
			     (long)(memsort[i]->ptr),
			     (char *)(memsort[i]->ptr));
      }
    }
  else {
#ifdef DEBUGMEM
    (void) fprintf(fichier,"No memory leak detected\n");
#else
    (void) fprintf(fichier,"No memory leak detector\n(comment in "
			   "'#define DEBUGMEM' within dbgmem.h)\n");
#endif
    }
  (void) fclose(fichier);
  (void) free(name);
  }
}
