/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Define sub level procedure to debug	*/
/*	memory problem				*/
/*						*/
/************************************************/
#ifndef DBGMEM
#define DBGMEM
#include	<stdlib.h>
#include	<stdarg.h>

/*define it at compile if you want to look for 	*/
/*memory leak, report is displayed each time	*/
/*dbg_dumpmem is called.			*/
#ifdef	DEBUG
#define DEBUGMEM
/*
*/
#endif

/*to activate Memory leak detector		*/
#define	DMTRUE	1	/*Boolean		*/
#define	DMFALSE	0	/*Boolean		*/

#ifdef DEBUGMEM
#ifndef DEBUG
#error "Caution, See dbgmem.h,"
#error "you should use DEBUGMEM in DEBUG compilation mode"
#endif
#define DMTRACE	DMTRUE	
#else
#define DMTRACE	DMFALSE	
#endif

#ifdef	DEBUGMEM	/*debugging mode	*/
#define	calloc		dbg_calloc
#define	malloc		dbg_malloc
#define	realloc		dbg_realloc
#define	free		dbg_free
#define	putenv		dbg_putenv
#define	setenv		dbg_setenv
#define	unsetenv	dbg_unsetenv
#define	asprintf	dbg_asprintf
#define	vasprintf	dbg_vasprintf
#define	SYSfree		dbg_sysfree
#undef strdup
#define	strdup		dbg_strdup
#define	strndup		dbg_strndup
#else	/*not in debugging mode			*/
#define	SYSfree		free
#endif

extern int memleak;

extern u_long dbg_gettotmem();
extern u_long dbg_getnummem();


extern void dbg_setcurpath(char *path);
extern void dbg_setcurbase(char *path);
extern void *dbg_malloc(size_t size);
extern void *dbg_calloc(size_t nmemb,size_t size);
extern void *dbg_realloc(void *ptr,size_t size);
extern char *dbg_strdup(const char *str);
extern char *dbg_strndup(const char *str,size_t n);
extern void dbg_free(void *ptr);
extern int dbg_putenv(char *valeur);
extern int dbg_setenv(const char *name,const char *value,int overwrite);
extern int dbg_unsetenv(const char *name);
extern int dbg_vasprintf(char **strp, const char *fmt,va_list args);
extern int dbg_asprintf(char **strp, const char *fmt, ...);
extern void dbg_sysfree(void *ptr);
extern void dbg_dumpmem(const char *apname,const char *extension);
extern void dbg_setnoleak(int value);
#endif
