/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIDNS:					*/
/*	Take Care of all access to DNS data	*/
/*						*/
/************************************************/
#include	<string.h>
#include	<resolv.h> 
#include	<netdb.h> 
#include	<stdlib.h> 
#include	<ctype.h> 
#include	<errno.h> 
#include	<time.h> 
#include	<arpa/inet.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"gesadv.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract RCPT information	*/
/*	from an advice request			*/
/*						*/
/************************************************/
static void getrcpt(ADVTYP *adv,char *msg)

{
RCPTYP *rcpt;

rcpt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
rcpt->rcpt=(char *)calloc(strlen(msg),sizeof(char));
if (sscanf(msg,"%d %s",(int *)&rcpt->status,rcpt->rcpt)!=2) {
  rcpt=eml_freercpt(rcpt);
  (void) clm_alert(0,"moddbd.c:getrcpt, Unable to scan <%s> (bug?)",msg);
  }
else {
  adv->rcpts=(RCPTYP **)vec_addlstlst((void **)adv->rcpts,(void *)rcpt);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract origin information	*/
/*	from an advice request			*/
/*						*/
/************************************************/
static void getsubject(ADVTYP *adv,char *subject)

{
#define	OPSU	"moddbd.c:getsubject,"

if (adv->subject==(char *)0) 
  adv->subject=strdup(subject);
else
  (void) clm_alert(0,"%s Multiple subject data <%s> (bug?)",OPSU,subject);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract origin information	*/
/*	from an advice request			*/
/*						*/
/************************************************/
static void getorigin(ADVTYP *adv,char *orig)

{
#define	OPGT	"gesadv.c:getorigin,"

#define	MXIPS	100	/*max ip number size	*/
#define	FRMORI	"%99s %s %lu %d"

if (adv->hfrom==(char *)0) {
  char *from;

  adv->clip=(char *)calloc(MXIPS,sizeof(char));
  from=(char *)calloc(strlen(orig),sizeof(char));
  if (sscanf(orig,FRMORI,adv->clip,from,&adv->size,&adv->relayed)!=4) {
    from=clm_freestr(from);
    adv->clip=clm_freestr(adv->clip);
    (void) clm_alert(0,"%s unexpected format in <%s> (bug?)",OPGT,orig);
    }
  if (from!=(char *)0) {
    adv->hfrom=strdup(from);
    adv->hdom=strdup(eml_getdomain(from));
    (void) free(from);
    }
  }
else 
  (void) clm_alert(0,"%s Multiple origin data <%s> (bug?)",OPGT,orig);
}
/*

*/
/************************************************/
/*						*/
/*	procedure to free memory used by an	*/
/*	ADVICE structure.			*/
/*						*/
/************************************************/
ADVTYP *adv_freeadvice(ADVTYP *adv)

{
if (adv!=(ADVTYP *)0) {
  adv->rcpts=eml_freercpts(adv->rcpts);
  adv->virname=clm_freestr(adv->virname);
  adv->hdom=clm_freestr(adv->hdom);
  adv->hfrom=clm_freestr(adv->hfrom);
  adv->clip=clm_freestr(adv->clip);
  adv->subject=clm_freestr(adv->subject);
  adv->session=clm_freestr(adv->session);
  (void) free(adv);
  adv=(ADVTYP *)0;
  }
return adv;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to generate an advice request	*/
/*	from a message "GVADV" contents		*/
/*						*/
/************************************************/
ADVTYP *adv_mkadvice(char **msgs)

{
ADVTYP *adv;

adv=(ADVTYP *)0;
if ((msgs!=(char **)0)&&(msgs[0]!=(char *)0)) {
  int i;

  adv=(ADVTYP *)calloc(1,sizeof(ADVTYP));
  adv->session=strdup(msgs[0]);
  adv->curtime=time((time_t *)0);
  for (i=1;msgs[i]!=(char *)0;i++) {
    typedef	enum	{
	a_origin,	/*origin data 	 	*/
	a_subject,	/*Email Subject  	*/
	a_spmlvl,	/*Email spam level  	*/
	a_time,		/*effective time	*/
	a_chk,		/*recipient entry	*/
	a_unknown	/*unknown		*/
	}ASKTYP;

    static VOCTYP vocable[]={
        {a_origin,SORIG,(const void *)0},
        {a_subject,SSUBJ,(const void *)0},
        {a_spmlvl,SSPM,(const void *)0},
        {a_time,CURTIME,(const void *)0},
        {a_chk,SCHK,(const void *)0},
        {a_unknown,(char *)0,(const void *)0}
        };

    VOCTYP *voc;

    if ((voc=clm_getvoca(vocable,msgs[i]))!=(VOCTYP *)0) {
      switch (voc->code) {
	case a_origin	:
	  (void) getorigin(adv,msgs[i]);
	  break;
	case a_subject	:
	  (void) getsubject(adv,msgs[i]);
	  break;
	case a_spmlvl	:
	  (void) sscanf(msgs[i],"%f/%f",&(adv->spmlvl),&(adv->spmlim));
	  break;
	case a_time	:
	  adv->curtime=clm_datetimesysasc(msgs[i],"");
	  break;
	case a_chk	:
	  (void) getrcpt(adv,msgs[i]);
	  break;
	case a_unknown	:
	  break;
	}
      }
    }
  }
return adv;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect all status 1 or 2	*/
/*	in the advice struct and generate an	*/
/*	advice answer				*/
/*						*/
/************************************************/
char *adv_advicemsg(ADVTYP *adv)

{
char *answ;

(void) asprintf(&answ,"%s %s",MYADV,adv->session);
if (adv->rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;adv->rcpts[i]!=(RCPTYP *)0;i++) {
    if (adv->rcpts[i]->smtpstatus==RETADV) {
      (void) clm_fasprintf(&answ,"\n%s %d %d",SCHK,i,adv->rcpts[i]->status);
      }
    }
  }
return answ;
}
