/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESDBC:					*/
/*	Take Care of all clement data-base	*/
/*	access and management.			*/
/*						*/
/************************************************/
#include	<string.h>
#include	<math.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"unijrl.h"
#include	"gesdbc.h"

#define	BINDING	"binding"	/*binding table	*/
#define	SRVTBL	"services"	/*services table*/
#define	CLMGRP	"clmgroup"	/*clmgroup table*/
#define	CLMBIND	"clmbind"	/*clmbind table	*/
#define	DOMS	"domain"	/*domain table	*/
#define	USERS	"users"		/*users table	*/
#define	INTUSR	"intusers"	/*internal users*/
#define	ALIASES	"aliases"	/*aliases table	*/
#define	BLSHST	"blshosts"	/*BLS servers	*/
#define	EMLOGS	"emlogs"	/*sessions logs	*/
#define	EMLTRT	"emltrt"	/*transit logs	*/
#define	USEPRF	"userprofile"	/*usage profile	*/
#define	OVRRIDE	"overrides"	/*overriding set*/
#define	EXTRACT	"extract"	/*extract list	*/
#define	INQUEUE	"inqueue"	/*transit queue	*/
#define	DOMAL	"domainalias"	/*dom alias list*/

#define	SELFROM	"SELECT * FROM"
#define	SELDOM	"%s %s WHERE domainname=%s"

#define	DFLTLNG	1		/*default langue*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage active status	*/
/*	return an strcmp value.			*/
/*						*/
/************************************************/
static int mngactive(char *username,int active)

#define	OPMN	"gesdbc.c:mngactive,"

{
switch (active) {
  case 0	:	/*user is not active	*/
  case 1	:	/*user is active	*/
  case 2	:	/*to rebounce group	*/
    break;		/*nothing to do		*/
  case 3	:	/*deleted user		*/
    active=0;		/*set user unactive	*/
    break;
  default	:	/*unexpect case		*/
    (void) clm_alert(0,"%s Unexpected active level '%d' for user <%s> within '%s' table!",
			OPMN,active,username,USERS);
    active=0;		/*set user unactive	*/
    break;
  }
return active;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to compare 2 aliases and	*/
/*	return an strcmp value.			*/
/*						*/
/************************************************/
static int cmpali(const void *p1,const void *p2)

{
return strcmp(*(char *const *)p1,*(char *const *) p2);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get an 'upto' date		*/
/*	including the grace period		*/
/*						*/
/************************************************/
static time_t getuptograce(SQLFILE *sql,SQLRESULT *rs,int recpos)

{
time_t upto;

upto=sql_tounixtime(sql,sql_getvalue(sql,rs,recpos,"upto"));
upto+=atoi(sql_getvalue(sql,rs,recpos,"grace"))*24*3600;
return upto;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to make the username search	*/
/*	string.					*/
/*						*/
/************************************************/
static char *getnsearch(SQLFILE *sql,char *rcpt)

{
char *nsearch;
char *username;

username=strdup(rcpt);
(void) eml_splitdomain(username);
if (strlen(username)>0) {
  char *cleanstr;

  cleanstr=sql_cleanstr(sql,username);
  (void) asprintf(&nsearch,"=%s",cleanstr);
  (void) free(cleanstr);
  }
else
  (void) asprintf(&nsearch,"%s","IS NULL");
(void) free(username);
return nsearch;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to make the domain search	*/
/*	string.					*/
/*						*/
/************************************************/
static char *getdsearch(char *name)

{
char *dsearch;
char *domain;

dsearch=(char *)0;
if ((domain=eml_getdomain(name))!=(char *)0) 
  (void) asprintf(&dsearch,"='%s'",domain);
else
  (void) asprintf(&dsearch,"%s","IS NULL");
return dsearch;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve group ID from	*/
/*	domain name, if the domain is unknown	*/
/*	return 0.				*/
/*						*/
/************************************************/
static u_int getgroupid(SQLFILE *sql,char *email)

{
u_int groupid;
char *domain;
SQLRESULT *rs;

groupid=0;
domain=sql_cleanstr(sql,eml_getdomain(email));
if ((rs=sql_tupple(sql,SELDOM,SELFROM,DOMS,domain))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    groupid=(atoi)(sql_getvalue(sql,rs,0,"groupid"));
  (void) sql_dropresult(sql,rs);
  }
(void) free(domain);
return groupid;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the number of	*/
/*	email sent to outside world by an	*/
/*	authenticated user.			*/
/*						*/
/************************************************/
static u_int getcount(SQLFILE *sql,char *authemail,u_int grpid)

{
#define	SELCNT	\
   "SELECT count(*) FROM %s WHERE sender='%s' AND status=%d AND grpin=%d AND grpout!=%d AND stamp>'%s'"

register char *YDAY;

u_int count;
SQLRESULT *rs;

count=0;
YDAY=sql_fromunixtime(sql,time((time_t *)0)-H24);
if ((rs=sql_tupple(sql,SELCNT,EMLOGS,authemail,m_delivered,grpid,grpid,YDAY))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    count=(atoi)(sql_getfield(sql,rs,0,0));
  (void) sql_dropresult(sql,rs);
  }
return count;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the emlog status	*/
/*	for a specific session+recipient.	*/
/*	If record is not found, return status	*/
/*	m_undefined.				*/
/*						*/
/************************************************/
static MENTYP getemlogstatus(SQLFILE *sql,char *session,char *rcpt)

{
#define	SELSTA	"SELECT status FROM %s WHERE session='%s' AND recipient=%s"

MENTYP status;
SQLRESULT *rs;

status=m_undefined;
if ((rs=sql_tupple(sql,SELSTA,EMLOGS,session,rcpt))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0)
    status=(atoi)(sql_getvalue(sql,rs,0,"status"));
  (void) sql_dropresult(sql,rs);
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to insert a queue entry within*/
/*	the table 'inqueue'			*/
/*	return "true" if not trouble, false	*/
/*	otherwise.				*/
/*	available.				*/
/*						*/
/************************************************/
static int setqueue(SQLFILE *sql,QUETYP *q)

{
#define	OPSQ	"gesdbc.c:setqueue,"

#define INSFLD1	"session,transit,outip,smtpmx,smtpport,subject,rejectto,orgeml,orgdom"
#define INSFLD2	"grpin,recname,recdom,mode,start,last,next,upto,orgstatus"
#define	INSQUE	\
  "INSERT INTO %s (%s,%s) VALUES('%s',%d,'%s','%s',%d,%s,%s,%s,'%s',%d,%s,'%s',%d,NOW(),NOW(),NOW(),'%s',%d)"

int status;
u_int grpin;
char *subject;
char *rejectto;
char *orgeml;
char *receml;
register char *TWODAY;

status=true;
subject=sql_cleanstr(sql,q->subject);
grpin=getgroupid(sql,q->orgeml);
rejectto=sql_cleanstr(sql,q->rejectto);
orgeml=sql_cleanstr(sql,q->orgeml);
receml=sql_cleanstr(sql,q->recname);
TWODAY=sql_fromunixtime(sql,time((time_t *)0)+(2*H24));
if (sql_action(sql,INSQUE,INQUEUE,INSFLD1,INSFLD2,q->session,q->transit,
					 	  q->outip,q->smtpmx,atoi(q->smtpport),subject,
					 	  rejectto,orgeml,eml_getdomain(q->orgeml),grpin,
					 	  receml,q->recdom,(int)q->mode,TWODAY,q->orgstatus)!=1) {
  (void) clm_alert(0,"%s Unable store session-iteration <%s-%03d> (BUG?)",
		      OPSQ,q->session,q->transit);
  status=false;
  }
(void) free(receml);
(void) free(orgeml);
(void) free(rejectto);
(void) free(subject);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract One Queue from	*/
/*	data-base.				*/
/*						*/
/************************************************/
static QUETYP *getonequeue(SQLFILE *sql,SQLRESULT *rs,int pos)

{
QUETYP *queue;
register char *session;
register char *recname;
register char *recdom;
register MMOENU mode;
register int orgstatus;
register char *outip;
register char *smtpmx;
register char *smtpport;
register u_long transit;
register char *subject;
register char *rejectto;
register char *orgeml;

session=sql_getvalue(sql,rs,pos,"session");
recname=sql_getvalue(sql,rs,pos,"recname");
recdom=sql_getvalue(sql,rs,pos,"recdom");
mode=(MMOENU)atoi(sql_getvalue(sql,rs,pos,"mode"));
orgstatus=atoi(sql_getvalue(sql,rs,pos,"orgstatus"));
outip=sql_getvalue(sql,rs,pos,"outip");
smtpmx=sql_getvalue(sql,rs,pos,"smtpmx");
smtpport=sql_getvalue(sql,rs,pos,"smtpport");
transit=atol(sql_getvalue(sql,rs,pos,"transit"));
subject=sql_getvalue(sql,rs,pos,"subject");
rejectto=sql_getvalue(sql,rs,pos,"rejectto");
orgeml=sql_getvalue(sql,rs,pos,"orgeml");
queue=que_mkqueue(outip,smtpmx,smtpport,session,transit,subject,rejectto,orgeml,recname,recdom,mode,orgstatus);
queue->start=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"start"));
queue->upto=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"upto"));
queue->nexttime=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"next"));
queue->lasttime=sql_tounixtime(sql,sql_getvalue(sql,rs,pos,"last"));
queue->retry=atoi(sql_getvalue(sql,rs,pos,"retry"));
queue->mode=atoi(sql_getvalue(sql,rs,pos,"mode"));
queue->numtrace=atol(sql_getvalue(sql,rs,pos,"numtrace"));
return queue;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a list of queued	*/
/*	email ready to be processed.		*/
/*						*/
/************************************************/
static QUETYP **getpendingqueue(SQLFILE *sql,int max)

{
#define SELQUE	"SELECT * FROM %s WHERE next<'%s' AND progress=0 ORDER BY recdom,session LIMIT %d"
QUETYP **queues;
SQLRESULT *rs;

queues=(QUETYP **)0;
if ((rs=sql_tupple(sql,SELQUE,INQUEUE,sql_fromunixtime(sql,time((time_t *)0)),max))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    register char *session;
    register char *recname;
    register char *recdom;
    register MMOENU mode;
    	     char *recipient;
    register char *rcpt;
   
     
    session=sql_getvalue(sql,rs,i,"session");
    recname=sql_getvalue(sql,rs,i,"recname");
    recdom=sql_getvalue(sql,rs,i,"recdom");
    mode=(MMOENU)atoi(sql_getvalue(sql,rs,i,"mode"));
    (void) asprintf(&recipient,"%s@%s",recname,recdom);
    rcpt=sql_cleanstr(sql,recipient);
    if ((mode!=0)||(getemlogstatus(sql,session,rcpt)!=m_undefined)) {
      QUETYP *q;

      q=getonequeue(sql,rs,i);
      queues=(QUETYP **)vec_addlstlst((void **)queues,(void *)q);
      }
    (void) free(rcpt);
    (void) free(recipient);
    }
  (void) sql_dropresult(sql,rs);
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract old troube queue	*/
/*	inqueue trouble queue.			*/
/*						*/
/************************************************/
static QUETYP **gettroublequeue(SQLFILE *sql,time_t curtime,int max)

{
#define SELTRL	"SELECT * FROM %s WHERE upto<'%s' AND progress=2 ORDER BY upto ASC LIMIT %d"
QUETYP **queues;
SQLRESULT *rs;

queues=(QUETYP **)0;
curtime+=(24*3600);
if ((rs=sql_tupple(sql,SELTRL,INQUEUE,sql_fromunixtime(sql,curtime),max))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    QUETYP *q;

    q=getonequeue(sql,rs,i);
    queues=(QUETYP **)vec_addlstlst((void **)queues,(void *)q);
    }
  (void) sql_dropresult(sql,rs);
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove old reference from	*/
/*	inqueue trouble queue.			*/
/*						*/
/************************************************/
static void cleantrouble(SQLFILE *sql,time_t curtime)

{
#define	OPCL	"gesdbc.c:cleantrouble,"

static LCKTYP lck1={INQUEUE,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

QUETYP **queues;
int phase;
int proceed;

queues=(QUETYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Locking queue		*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPCL,INQUEUE);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*processing list	*/
      if ((queues=gettroublequeue(sql,curtime,10))!=(QUETYP **)0) {
        int i;

	
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  char *tracename;

  	  tracename=que_tracename(queues[i]);
	  (void) unlink(tracename);
          (void) dbc_delqueue(sql,queues[i]);
	  (void) free(tracename);
	  }
	queues=que_freequeues(queues);
        }
      break;
    case 2	:	/*unlocking data-base	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the alias list if	*/
/*	available.				*/
/*						*/
/************************************************/
static char **getalias(SQLFILE *sql,char *rcpt,char ***known)

{
#define	OPGT	"gesdbc.c:getalias,"

#define	SELALI	"%s %s WHERE domain %s AND username %s"

char **aliases;
int phase;
int proceed;
char *username;
char *domain;
SQLRESULT *rs;

aliases=(char **)0;
phase=0;
proceed=true;
username=getnsearch(sql,rcpt);
domain=getdsearch(rcpt);
rs=(SQLRESULT *)0;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*aliase already seen	*/
      if (*known!=(char **)0) {
	int i;

	for (i=0;(*known)[i]!=(char *)0;i++) {
	  if (strcmp((*known)[i],rcpt)==0) {
	    aliases=(char **)calloc(1,sizeof(char *));
	    phase=999;
	    break;
	    }
	  }
	}
      break;
    case 1	:	/*extracting aliase	*/
      if ((rs=sql_tupple(sql,SELALI,SELFROM,ALIASES,domain,username))!=(SQLRESULT *)0) {
	if (sql_nbrtupple(sql,rs)>0) {
	  aliases=clm_splitstring(sql_getvalue(sql,rs,0,"aliasnames"),',');
	  if (aliases==(char **)0)
	  /*to mark is an empty alias!		*/
	    aliases=(char **)calloc(1,sizeof(char *));
          *known=(char **)vec_addlstlst((void **)(*known),(void *)strdup(rcpt));
	  }
        (void) sql_dropresult(sql,rs);
  	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(domain);
(void) free(username);
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve aliase list	*/
/*						*/
/************************************************/
char **getaliaslist(SQLFILE *sql,char *rcpt,char ***known,int depth)

{
#define	OPGL	"gesdbc.c:getaliaslist,"
#define	MXADPTH	10

char **aliases;
char **sup;
int phase;
int proceed;

aliases=(char **)0;
sup=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase)  {
    case 0	:	/*critical depth reach!	*/
      if (depth>=MXADPTH) {
	(void) clm_alert(0,"%s, too many aliases recursions '%d>=%d'",OPGL,depth,MXADPTH);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*let get the alias list*/
      if ((sup=getalias(sql,rcpt,known))==(char **)0) {
	phase=999;	/*it was not an alias	*/
	}
      break;
    case 2	:	/*is alias empty?	*/
      if (sup[0]==(char *)0) {
	aliases=sup;
	phase=999;	/*alias empty		*/
	}
      break;
    case 3	:	/*lets scan alias list	*/
      if (sup!=(char **)0) {	/*always	*/
	int i;

	for (i=0;sup[i]!=(char *)0;i++) {
	  char **subalias;

	  subalias=(char **)0;
	  if (strcmp(sup[i],rcpt)==0) {
	    aliases=(char **)vec_addlstlst((void **)aliases,(void *)sup[i]);
	    continue;
	    }
	  if ((subalias=getaliaslist(sql,sup[i],known,depth+1))!=(char **)0) {
	    int j;

	    for (j=0;subalias[j]!=(char *)0;j++) {
	      aliases=(char **)vec_addlstlst((void **)aliases,(void *)subalias[j]);
	      }
	    aliases=(char **)vec_addlstlst((void **)aliases,(void *)0);
	    (void) free(subalias);
	    (void) free(sup[i]);
	    continue;
	    }
	  aliases=(char **)vec_addlstlst((void **)aliases,(void *)sup[i]);
	  }
	(void) free(sup);
	}
      break;
    case 4	:	/*sorting alias list	*/
      (void) qsort((void *)aliases,vec_sizelstlst((void **)aliases),sizeof(char *),cmpali);
      break;
    case 5	:	/*duplicate no double	*/
      if (aliases!=(char **)0) { /*always	*/
	char **unique;

	unique=(char **)calloc(vec_sizelstlst((void **)aliases)+1,sizeof(char *));
	unique[0]=aliases[0];
	if (unique[0]!=(char *)0) {
	  int i,j;

	  for (i=1,j=0;aliases[i]!=(char *)0;i++) {
	    if (strcmp(unique[j],aliases[i])==0)
	      (void) free(aliases[i]);
	    else {
	      j++;
	      unique[j]=aliases[i];
	      }
	    }
	  }
        (void) free(aliases);
	aliases=unique;
	} 
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve group record	*/
/*	related to a specific group ID.		*/
/*						*/
/************************************************/
static GRPTYP *getgroupbyid(SQLFILE *sql,int groupid)

{
#define	SELCLM	"%s %s WHERE groupid=%d"

GRPTYP *grp;
SQLRESULT *rs;

grp=(GRPTYP *)0;
if ((rs=sql_tupple(sql,SELCLM,SELFROM,CLMGRP,groupid))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    char *smtprmt;
    char *authrmt;

    grp=(GRPTYP *)calloc(1,sizeof(GRPTYP));
    grp->groupid=groupid;
    grp->active|=(atoi)(sql_getvalue(sql,rs,0,"active"));
    grp->prfid=(atoi)(sql_getvalue(sql,rs,0,"prfid"));
    grp->grpname=sql_strdup(sql_getvalue(sql,rs,0,"groupname"));
    grp->upto=getuptograce(sql,rs,0);
    grp->smtpport=strdup(SMTPPORT);
    smtprmt=sql_getvalue(sql,rs,0,"remotesmtp");
    if (smtprmt!=(char *)0) {
      char *strport;
      char *exitip;

      grp->smtprmt=strdup(smtprmt);
      if ((strport=sql_getvalue(sql,rs,0,"remoteport"))!=(char *)0) {
	if (atoi(strport)!=0) {
	  (void) free(grp->smtpport);
          grp->smtpport=strdup(strport);
	  }
	}
      if ((exitip=sql_getvalue(sql,rs,0,"exitip"))==(char *)0)
	exitip=ANYADDR;
      grp->exitip=strdup(exitip);
      }
    grp->authport=strdup(POP3SPORT);
    grp->authprot=strdup(POP3S);
    authrmt=sql_getvalue(sql,rs,0,"authserver");
    if (authrmt!=(char *)0) {
      char *authport;
      char *authprot;

      grp->authrmt=strdup(authrmt);
      if ((authport=sql_getvalue(sql,rs,0,"authport"))!=(char *)0) {
	if (atoi(authport)!=0) {
	  (void) free(grp->authport);
	  grp->authport=strdup(authport);
	  }
	}
      if ((authprot=sql_getvalue(sql,rs,0,"authproto"))!=(char *)0) {
	(void) free(grp->authprot);
	grp->authprot=strdup(authprot);
	}
      }
    (void) clm_strtolower(grp->authprot);
    }
  else
    (void) jrl_journalling(false,APJRL,"Unable to find goup definition for "
				       "groupid='%d' (DB trouble?!)",groupid);
  (void) sql_dropresult(sql,rs);
  }
return grp;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve data-base field	*/
/*	according one field name and email	*/
/*	address, field can be on more than on	*/
/*	table, search is done in a specific way.*/
/*						*/
/*	If field is not available a NULL ptr	*/
/*	is returned.				*/
/*						*/
/************************************************/
static char *getcommonfield(SQLFILE *sql,char *email,char *fieldname)

{
#define	PROBE0	"SELECT * FROM %s WHERE %s %s AND %s %s"
#define	PROBE1	"SELECT * FROM %s WHERE %s=%s"
#define	PROBE2	"SELECT * FROM %s WHERE %s %s"
#define	PROBE3	\
	"SELECT * FROM %s WHERE %s=(SELECT %s FROM %s WHERE domainname %s)"

char *field;
char *user;
char *cleanemail;
char *domain;
int phase;
int proceed;

field=(char *)0;
cleanemail=sql_cleanstr(sql,email);
user=getnsearch(sql,email);
domain=getdsearch(email);
phase=0;
proceed=true;
while (proceed==true) {
  char *probe;

  probe=(char *)0;
  switch (phase) {
    case 0	:	/*is it an aliases	*/
      (void) asprintf(&probe,PROBE0,ALIASES,"domain",domain,"username",user);
      break;
    case 1	:	/*is it in email table	*/
      (void) asprintf(&probe,PROBE1,USERS,"username",cleanemail);
      break;
    case 2	:	/*checking domain table	*/
      (void) asprintf(&probe,PROBE2,DOMS,"domainname",domain);
      break;
    case 3	:	/*checking group  table	*/
      (void) asprintf(&probe,PROBE3,CLMGRP,"groupid","groupid",DOMS,domain);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if (probe!=(char *)0) {
    SQLRESULT *rs;
    
    if ((rs=sql_tupple(sql,probe))!=(SQLRESULT *)0) {
      if (sql_nbrtupple(sql,rs)>0) {
	char *ptr;

	if ((ptr=sql_getvalue(sql,rs,0,fieldname))!=(char *)0) {
          field=strdup(ptr);
	  proceed=false;	/*Found IT!		*/
	  }
	}
      (void) sql_dropresult(sql,rs);
      }
    (void) free(probe);
    }
  phase++;
  }
(void) free(domain);
(void) free(user);
(void) free(cleanemail);
return field;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve data-base value	*/
/*	according one field name and and email	*/
/*	address					*/
/*	to make sure the SPF protection apply	*/
/*	to him.					*/
/*						*/
/************************************************/
static int getcommonvalue(SQLFILE *sql,char *email,char *fieldname,int defval)

{
char *field;

if ((field=getcommonfield(sql,email,fieldname))!=(char *)0) {
  defval=atoi(field);
  (void) free(field);
  }
return defval;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a structure to check	*/
/*	recipient name against a profile	*/
/*						*/
/************************************************/
static PRFTYP *getprofile(SQLFILE *sql,int prfid)

{
#define	GETPRF	"SELECT * FROM %s WHERE prfid=%d"

PRFTYP *profile;
SQLRESULT *rs;

profile=(PRFTYP *)0;
if ((rs=sql_tupple(sql,GETPRF,USEPRF,prfid))!=(SQLRESULT *)0) {
  profile=(PRFTYP *)calloc(1,sizeof(PRFTYP));
  if (sql_nbrtupple(sql,rs)>0) {
    const char *lang;

    profile->prfid=prfid;
    profile->system=atoi(sql_getvalue(sql,rs,0,"system"));
    profile->emailsz=atol(sql_getvalue(sql,rs,0,"emailsize"));
    profile->checkrmtspf=atoi(sql_getvalue(sql,rs,0,"checkrmtspf"));
    profile->checkrmtmx=atoi(sql_getvalue(sql,rs,0,"checkrmtmx"));
    profile->mxmgbytes=atol(sql_getvalue(sql,rs,0,"mxmgbytes"));
    profile->bulk=atoi(sql_getvalue(sql,rs,0,"bulk"));
    profile->blsid=atoi(sql_getvalue(sql,rs,0,"blsid"));
    profile->maxperday=atoi(sql_getvalue(sql,rs,0,"maxperday"));
    lang=sql_getvalue(sql,rs,0,"lang");
    profile->lang=DFLTLNG;
    if (lang!=(char *)0)
      profile->lang=atoi(lang);
    }
  else {	/*Lets set minimal values		*/
    profile->prfid=1;		/*external account	*/
    profile->emailsz=10*1024;	/*ten K max		*/
    profile->checkrmtspf=1;
    profile->checkrmtmx=1;
    profile->bulk=3;		/*very minimal value	*/
    (void) jrl_journalling(false,APJRL,"Unable to find profile definition for prfid='%d' (DB trouble?!)",
					prfid);
    }
  (void) sql_dropresult(sql,rs);
  }
return profile;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return one override record.*/
/*	Record is ajusted to address the '*'	*/
/*	field contents.				*/
/*						*/
/************************************************/
static OVRTYP *getoneoverride(SQLFILE *sql,char *orgname,char *orgdomain,
					   char *recname,char *recdomain,
					   SQLRESULT *rs,int num)

{
OVRTYP *ovr;
char *locname;
char *locdom;
int i;

ovr=(OVRTYP *)calloc(1,sizeof(OVRTYP));
locname=(char *)0;
locdom=(char *)0;
for (i=0;i<3;i++) {
  const char *dflt;
  const char *str;
  char **dst;

  dflt=(const char *)0;
  str=(const char *)0;
  dst=(char **)0;
  switch (i) {
    case 0	:
      str=sql_getvalue(sql,rs,num,"orgname");
      dflt=orgname;
      dst=&(locname);
      break;
    case 1	:
      str=sql_getvalue(sql,rs,num,"orgdom");
      dflt=orgdomain;
      dst=&(locdom);
      break;
    case 2	:
      str=sql_getvalue(sql,rs,num,"orgip");
      dflt="0/0";
      dst=&(ovr->orgip);
      break;
    }
  if (str==(char *)0)
    str="---";		/*NULL domain	*/
  if (strcmp(str,"*")==0) 
    str=dflt;
  *dst=sql_strdup(str);
  }
ovr->ovrid=(atol)(sql_getvalue(sql,rs,num,"ovrid"));
ovr->ovrride=(atoi)(sql_getvalue(sql,rs,num,"ovrride"));
(void) asprintf(&(ovr->orgemail),"%s@%s",locname,locdom);
(void) free(locdom);
(void) free(locname);
return ovr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a table was	*/
/*	later than a specific date.		*/
/*						*/
/************************************************/
static time_t checkupdst(SQLFILE *sql,const char *table)

{
#define MAXREC	1	/*just need one record	*/
#define	SELUPD	"%s %s ORDER BY updated DESC LIMIT %d"

time_t update;
SQLRESULT *rs;

update=(time_t)0;
if ((rs=sql_tupple(sql,SELUPD,SELFROM,table,MAXREC))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)==MAXREC) {
    update=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"updated"));
    }
  (void) sql_dropresult(sql,rs);
  }
return update;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update email logs		*/
/*						*/
/************************************************/
static int setlogs(SQLFILE *sql,DSPTYP *dsp,RCPTYP **rcpts)

{
#define	LOGSFL1	"session,rmtip,rmthostname,originator,localip,taille"
#define	LOGSFL2	"subject,recipient,status,why,grpin,grpout,sender"
#define	INLOGS "INSERT INTO %s (%s,%s) VALUES (%s,%s,%s,%d,%s,%u,%u,%s)"
int done;

done=0;
if ((dsp!=(DSPTYP *)0)&&(rcpts!=(RCPTYP **)0)) {
  register char *subject;
  register char *hfrom;
  register u_int grpin;
  char *str;
  int i;

  grpin=getgroupid(sql,dsp->efrom);
  hfrom=sql_cleanstr(sql,dsp->hfrom);
  subject=sql_cleanstr(sql,dsp->subject);
  (void) asprintf(&str,"'%s','%s','%s',%s,'%s',%lu",
		        dsp->session,dsp->peerip,dsp->peername,
			hfrom,dsp->locip,dsp->mtaille);
  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rloc;
    register char *rcpt;
    register char *inbehalf;
    register char *why;
    register char *locsub;
    register u_int grpout;

    rloc=rcpts[i];
    rcpt=sql_cleanstr(sql,rloc->rcpt);
    inbehalf=sql_cleanstr(sql,rloc->inbehalf);
    why=(char *)0;
    locsub=subject;
    grpout=getgroupid(sql,rloc->rcpt);
    switch (rloc->status) {
      case m_relayed	:
      case m_toforward	:
      case m_toresend	:
      case m_direct	:
        why=sql_cleanstr(sql,(char *)0);
	break;
      case m_rejected	:
      case m_empty	:
      case m_discarded	:
      case m_virusfound	:
      case m_trouble	:
        why=sql_cleanstr(sql,rloc->why);
	locsub=why;
	break;
      case m_quarantine	:
        why=sql_cleanstr(sql,rloc->why);
	if (dsp->subject==(char *)0) 
	  locsub=why;
	break;
      default		:
	break;
      }
    (void) sql_action(sql,INLOGS,EMLOGS,LOGSFL1,LOGSFL2,str,locsub,rcpt,rloc->status,why,grpin,grpout,inbehalf);
    why=clm_freestr(why);
    (void) free(inbehalf);
    (void) free(rcpt);
    done++;
    }
  (void) free(subject);
  (void) free(hfrom);
  (void) free(str);
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update email transit record*/
/*						*/
/************************************************/
int dbc_settrtdata(SQLFILE *sql,QUETYP *queue)

{
#define	OPSE	"gesdbc.c:dbc_settrtdata,"

#define	UPDTRT	\
  "UPDATE %s SET why=%s,last=%s,next=%s,status=%d WHERE session='%s' AND mfrom=%s AND recipient=%s"
#define	INSTRT	"INSERT INTO %s VALUES ('%s',%s,%s,%s,%d,%s,%s,%s,%s,%d)"

static LCKTYP lck1={EMLTRT,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
char *last;
char *next;
char *why;
char *recipient;
char *receml;
char *orgeml;
int phase;
int proceed;

done=true;
(void) asprintf(&last,"'%s'",sql_fromunixtime(sql,queue->lasttime));
(void) asprintf(&next,"'%s'",sql_fromunixtime(sql,queue->nexttime));
(void) asprintf(&recipient,"%s@%s",queue->recname,queue->recdom);
why=sql_cleanstr(sql,queue->why);
receml=sql_cleanstr(sql,recipient);
orgeml=sql_cleanstr(sql,queue->orgeml);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*updating record	*/
      if (sql_action(sql,UPDTRT,EMLTRT,why,last,next,queue->status,
				       queue->session,orgeml,receml)==1)
        phase=999;	/*record already present*/
      break;
    case 1	:	/*locking database table*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPSE,EMLTRT);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*updating record again	*/
      if (sql_action(sql,UPDTRT,EMLTRT,why,last,next,queue->status,
				       queue->session,orgeml,receml)==1)
        phase++;	/*record just created! 	*/
      break;
    case 3	:	/*creating record	*/
      if (done==true) {	/*always		*/
	char *first;
	char *subject;

        (void) asprintf(&first,"'%s'",sql_fromunixtime(sql,queue->start));
        subject=sql_cleanstr(sql,queue->subject);
        if (sql_action(sql,INSTRT,EMLTRT,queue->session,first,last,next,queue->retry,
				         orgeml,receml,subject,why,queue->status)!=1) {
          (void) clm_alert(0,"%s Unable to create transaction table session=<%s>",
			      OPSE,queue->session);
          (void) sql_unlock_rollback(sql);
	  done=false;
	  phase=999;	/*trouble trouble	*/
	  }
       (void) free(subject);
       (void) free(first);
       }
      break;
    case 4	:	/*unlocking db table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(orgeml);
(void) free(receml);
(void) free(why);
(void) free(recipient);
(void) free(next);
(void) free(last);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract an alias list if	*/
/*	there is one				*/
/*	available.				*/
/*						*/
/************************************************/
char **dbc_getaliaslist(SQLFILE *sql,char *rcpt)

{
char **known;
char **aliases;

known=(char **)0;
aliases=getaliaslist(sql,rcpt,&known,0);
known=(char **)vec_freelstlst((void **)known,VECFREE);
return aliases;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return user information if	*/
/*	available.				*/
/*						*/
/************************************************/
void dbc_getlocaluser(SQLFILE *sql,USRTYP *usr)

{
#define	OPGC	"gesdbc.c:dbc_getlocaluser,"

#define	SELUSR	"%s %s WHERE username=%s"

SQLRESULT *rs;
char *username;

usr->rcptflg=rcpt_unknown;
username=sql_cleanstr(sql,usr->username);
if ((rs=sql_tupple(sql,SELUSR,SELFROM,USERS,username))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    PRFTYP *prf;
    GRPTYP *grp;

    prf=getprofile(sql,getcommonvalue(sql,usr->username,"prfid",1));
    usr->prfid=prf->prfid;
    usr->rcptflg=rcpt_islocal;
    usr->active=mngactive(usr->username,atoi(sql_getvalue(sql,rs,0,"active")));
    if ((grp=dbc_getgroup(sql,usr->username))!=(GRPTYP *)0) {
      usr->active&=grp->active;
      grp=dbc_freegroup(grp);
      }
    else
      (void) jrl_journalling(true,APJRL,"Unable to find goup definition "
					"for user <%s> (DB trouble?!)",usr->username);
    usr->upto=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"upto"));
    if ((usr->upto>0)&&(usr->upto<time((time_t *)0))) {
      usr->active=false;
      usr->rcptflg=rcpt_inactive;
      }
    if (usr->active==true) {
      /*we set user password if account is allowed to be accessed	*/
      usr->password=sql_strdup(sql_getvalue(sql,rs,0,"password"));
      /*checking if Email space reach its maximun value			*/
      if (prf->mxmgbytes<atoi(sql_getvalue(sql,rs,0,"usedmgbytes")))
	usr->rcptflg=rcpt_boxtb;
      }
    else
      usr->password=strdup("!!");	/*user can't Be identified	*/
    prf=dbc_freeprofile(prf);
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(username);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get the available remote	*/
/*	mentor list.				*/
/*						*/
/************************************************/
SRVLST **dbc_getsrvlst(SQLFILE *sql,char *srvname)

{
#define	GETMNT	"SELECT * FROM %s WHERE srvname='%s' ORDER BY preference LIMIT 10"

SRVLST **srvlst;
SQLRESULT *rs;

srvlst=(SRVLST **)0;
if ((rs=sql_tupple(sql,GETMNT,SRVTBL,srvname))!=(SQLRESULT *)0) {
  int nbr;
  int i;

  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    SRVLST *srvloc;

    srvloc=(SRVLST *)calloc(1,sizeof(SRVLST));
    srvloc->srvname=strdup(srvname);
    srvloc->srvhost=sql_strdup(sql_getvalue(sql,rs,i,"srvhost"));
    srvloc->srvport=sql_strdup(sql_getvalue(sql,rs,i,"srvport"));
    srvloc->srvaccess=sql_strdup(sql_getvalue(sql,rs,i,"srvaccess"));
    srvlst=(SRVLST **)vec_addlstlst((void **)srvlst,(void *)srvloc);
    }
  (void) sql_dropresult(sql,rs);
  }
return srvlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free the list of available	*/
/*	mentor.					*/
/*						*/
/************************************************/
SRVLST **dbc_freesrvlst(SRVLST **srvlst)

{
if (srvlst!=(SRVLST **)0) {
  int i;

  for (i=0;srvlst[i]!=(SRVLST *)0;i++) {
    (void) free(srvlst[i]->srvname);
    srvlst[i]->srvport=clm_freestr(srvlst[i]->srvport);
    srvlst[i]->srvhost=clm_freestr(srvlst[i]->srvhost);
    srvlst[i]->srvaccess=clm_freestr(srvlst[i]->srvaccess);
    (void) free(srvlst[i]);
    }
  (void) free(srvlst);
  srvlst=(SRVLST **)0;
  }
return srvlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to store the session result	*/
/*	inside the clement logs.		*/
/*						*/
/************************************************/
int dbc_setlogs(SQLFILE *sql,DSPTYP *dsp,RCPTYP **rcpts)

{
static LCKTYP lck1={EMLOGS,dbd_lockwrite};
static LCKTYP lck2={DOMS,dbd_lockread};
static LCKTYP *lcklst[]={&lck1,&lck2,(LCKTYP *)0};

int done;
int phase;
int proceed;

done=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking data-base	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table","gesdbc.c:dbc_setlogs,",EMLOGS);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*updating logs		*/
      done=setlogs(sql,dsp,rcpts);	
      break;
    case 2	:	/*unlocking data base	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a log	*/
/*	record.					*/
/*						*/
/************************************************/
LOGTYP *dbc_freelogs(LOGTYP *mlog)

{
if (mlog!=(LOGTYP *)0) {
  mlog->subject=clm_freestr(mlog->subject);
  mlog->recipient=clm_freestr(mlog->recipient);
  mlog->hfrom=clm_freestr(mlog->hfrom);
  mlog->efrom=clm_freestr(mlog->efrom);
  mlog->rmtip=clm_freestr(mlog->rmtip);
  mlog->session=clm_freestr(mlog->session);
  (void) free(mlog);
  mlog=(LOGTYP *)0;
  }
return mlog;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a log 		*/
/*						*/
/************************************************/
LOGTYP *dbc_getlogs(SQLFILE *sql,char *session,char *recipient)

{
#define	OPLO	"gesdbc.c:dbc_getlogs,"

#define	SELLOG	"SELECT * FROM %s WHERE session='%s' AND recipient=%s"

LOGTYP *emlog;
char *rcpt;
SQLRESULT *rs;

emlog=(LOGTYP *)0;
rcpt=sql_cleanstr(sql,recipient);
if ((rs=sql_tupple(sql,SELLOG,EMLOGS,session,rcpt))!=(SQLRESULT *)0) {
  int nbr;

  nbr=sql_nbrtupple(sql,rs);
  if (nbr>0) {
    if (nbr>1)
      (void) clm_alert(0,"%s found %d %s records for session=<%s> and recipient<%s> (bug?)",
			  OPLO,nbr,EMLOGS,session,recipient);
    emlog=(LOGTYP *)calloc(1,sizeof(LOGTYP));
    emlog->session=strdup(session);
    emlog->rmtip=sql_strdup(sql_getvalue(sql,rs,0,"rmtip"));
    emlog->efrom=sql_strdup(sql_getvalue(sql,rs,0,"sender"));
    emlog->hfrom=sql_strdup(sql_getvalue(sql,rs,0,"originator"));
    emlog->subject=sql_strdup(sql_getvalue(sql,rs,0,"subject"));
    emlog->recipient=strdup(recipient);
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(rcpt);
return emlog;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update session logs within	*/
/*	data-base.				*/
/*						*/
/************************************************/
int dbc_updlogs(SQLFILE *sql,char *session,char *recipient,MENTYP status)

{
#define	UPDLOGS "UPDATE %s set status=%d WHERE session='%s' AND recipient=%s"

int done;
char *rcpt;

done=false;
rcpt=sql_cleanstr(sql,recipient);
done=sql_action(sql,UPDLOGS,EMLOGS,status,session,rcpt);
(void) free(rcpt);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if binding info have	*/
/*	been updated?				*/
/*						*/
/************************************************/
time_t dbc_checkbind(SQLFILE *sql)

{
return checkupdst(sql,BINDING);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get Binding information	*/
/*	request is done according a protocol	*/
/*	value.					*/
/*						*/
/************************************************/
BININF **dbc_getbinds(SQLFILE *sql,PROTYP proto)

{
#define	OPGE	"gesdbc.c:dbc_getbinds,"
#define SELBIN	"SELECT * FROM %s WHERE active=1 AND 	\
		  bindid IN \
		 (SELECT bindid FROM %s WHERE groupid IN \
		 (SELECT groupid FROM %s WHERE active=1))"

static	const char *strproto[]={
		"smtp","mentor","Unknown",(char *)0
		};

BININF **binds;
SQLRESULT *rs;

binds=(BININF **)0;
if ((rs=sql_tupple(sql,SELBIN,BINDING,CLMBIND,CLMGRP))!=(SQLRESULT *)0) {
  const char *needproto;
  int nbr;
  u_int i;

  needproto=strproto[(int)proto];
  nbr=sql_nbrtupple(sql,rs);
  for (i=0;i<nbr;i++) {
    char *hostname;
    char *port;
    u_int iteration;
    char **list;

    if (strcmp(needproto,sql_getvalue(sql,rs,i,"protocol"))!=0)
      continue;
    iteration=(atoi)(sql_getvalue(sql,rs,i,"iteration"));
    port=sql_getvalue(sql,rs,i,"port");
    hostname=sql_getvalue(sql,rs,i,"binding");
    if ((list=dns_getiplist(hostname))==(char **)0) {
      (void) clm_alert(0,"%s Unable to get iplist for host <%s>",OPGE,hostname);
      }
    else {
      u_int j;

      for (j=0;list[j]!=(char *)0;j++) {
        BININF *entre;

        entre=dev_mkbindinf(proto,list[j],port,iteration);
        binds=(BININF **)vec_addlstlst((void **)binds,(void *)entre);
	}
      list=(char **)vec_freelstlst((void **)list,VECFREE);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return binds;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to dynamicaly add a domain	*/
/*	to group.				*/
/*						*/
/************************************************/
int dbc_adddomtogroup(SQLFILE *sql,GRPTYP *grp,char *domain)

{
#define	OPAD	"gesdbc.c:dbc_adddomtogroup,"

#define	UPDDOM	"UPDATE %s set lang=(SELECT lang FROM %s WHERE groupid=%d) WHERE domainname='%s'"
#define INDOM	"INSERT INTO %s (domainname,groupid,prfid,active) VALUES('%s',%d,%d,%d)"

static LCKTYP lck1={DOMS,dbd_lockwrite};
static LCKTYP lck2={CLMGRP,dbd_lockread};
static LCKTYP *lcklst[]={&lck1,&lck2,(LCKTYP *)0};

int done;
char *action;
int phase;
int proceed;

done=true;
(void) asprintf(&action,UPDDOM,DOMS,CLMGRP,grp->groupid,domain);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*domain already added	*/
      if (sql_action(sql,action)>0)
	phase=999;	/*found at least one	*/
      break;
    case 1	:	/*locking the data-base	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPAD,DOMS);
	done=false;
	phase=999;
	}
      break;
    case 2	:	/*updating		*/
      if (sql_action(sql,action)>0) 
	phase++;	/*race condition 	*/
      break;
    case 3	:	/*creating		*/
      (void) free(action);
      (void) asprintf(&action,INDOM,DOMS,domain,grp->groupid,grp->prfid,(int)da_good);
      if (sql_action(sql,action)==0) {
	(void) clm_alert(0,"%s Unable to update %s table (bug?)",OPAD,DOMS);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      else
	(void) sql_action(sql,UPDDOM,DOMS,CLMGRP,grp->groupid,domain);
      break;
    case 4	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(action);
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a default domain	*/
/*	name. That domain is the oldest domain	*/
/*	defined within domain table in groupid 1*/
/*						*/
/************************************************/
char *dbc_getdfltdom(SQLFILE *sql)

{
#define	GETDFLD	"SELECT * FROM %s WHERE groupid=1 AND active=1 ORDER BY first ASC"
char *dfltdom;
SQLRESULT *rs;

dfltdom=(char *)0;
if ((rs=sql_tupple(sql,GETDFLD,DOMS))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) 
    dfltdom=sql_strdup(sql_getvalue(sql,rs,0,"domainname"));
  (void) sql_dropresult(sql,rs);
  }
return dfltdom;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a default originator*/
/*	according domain provided.		*/
/*	For now the default from is always the	*/
/*	same value.				*/
/*						*/
/************************************************/
char *dbc_getdfltfrom(SQLFILE *sql,char *domain)

{
char *dfltfrom;
char *dfltdom;

dfltdom=dbc_getdfltdom(sql);
(void) asprintf(&dfltfrom,"%s@%s",POSTM,dfltdom);
(void) free(dfltdom);
return dfltfrom;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a local*/
/*	user definition.			*/
/*						*/
/************************************************/
USRTYP *dbc_freelocaluser(USRTYP *localusr)

{
if (localusr!=(USRTYP *)0) {
  localusr->password=clm_freestr(localusr->password);
  localusr->username=clm_freestr(localusr->username);
  (void) free(localusr);
  localusr=(USRTYP *)0;
  }
return localusr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by group	*/
/*	definition structure.			*/
/*						*/
/************************************************/
GRPTYP *dbc_freegroup(GRPTYP *group)

{
if (group!=(GRPTYP *)0) {
  group->exitip=clm_freestr(group->exitip);
  group->authprot=clm_freestr(group->authprot);
  group->authport=clm_freestr(group->authport);
  group->authrmt=clm_freestr(group->authrmt);
  group->smtpport=clm_freestr(group->smtpport);
  group->smtprmt=clm_freestr(group->smtprmt);
  group->grpname=clm_freestr(group->grpname);
  (void) free(group);
  group=(GRPTYP *)0;
  }
return group;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract group information	*/
/*	from data-base.				*/
/*	If we are not able to retreive a group	*/
/*	from the domain, we return a NUL pointer*/
/*						*/
/************************************************/
GRPTYP *dbc_getgroup(SQLFILE *sql,char *rcpt)

{
GRPTYP *grp;
u_int groupid;

grp=(GRPTYP *)0;
groupid=0;
if ((groupid=getgroupid(sql,rcpt))!=0) {
  if ((grp=getgroupbyid(sql,groupid))==(GRPTYP *)0)
    (void) jrl_journalling(true,APJRL,"Unable to find goup definition for <%s> (DB trouble?!)",rcpt);
  }
return grp;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract group information	*/
/*	list from data-base.			*/
/*						*/
/************************************************/
GRPTYP **dbc_getgrouplst(SQLFILE *sql)

{
#define	SELGRP	"SELECT * FROM %s"

GRPTYP **grplst;
SQLRESULT *rs;

grplst=(GRPTYP **)0;
if ((rs=sql_tupple(sql,SELGRP,CLMGRP))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;

    for (i=0;i<nbr;i++) {
      int groupid;
      GRPTYP *grp;

      groupid=(atoi)(sql_getvalue(sql,rs,i,"groupid"));
      if ((grp=getgroupbyid(sql,groupid))!=(GRPTYP *)0) 
        grplst=(GRPTYP **)vec_addlstlst((void **)grplst,(void *)grp);
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return grplst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the BLSlists from	*/
/*	the data-base within an allocated memory*/
/*						*/
/************************************************/
char **dbc_getblssites(SQLFILE *sql,int blsid,int af_net)

{
#define	GETBLS	"SELECT * FROM %s WHERE blsid=%d"
char **sblsites;
SQLRESULT *rs;

sblsites=(char **)0;
if ((rs=sql_tupple(sql,GETBLS,BLSHST,blsid))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {
    int i;
 
    for (i=0;i<nbr;i++) {
      char *ipmode;

      ipmode=(char *)0;
      switch (af_net) {
	case 0		:	/*we take all BLS site	*/
	  ipmode="ipv";
	  break;
	case AF_INET	:	/*Only IPV4 Site	*/
	  ipmode="ipv4";
	  break;
	case AF_INET6	:	/*Only IPV6 Site	*/
	  ipmode="ipv6";
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpected af_net value '%d' (bug?)","gesdbc.c:dbc_getblssites",af_net);
	  break;
	}
      if (strstr(sql_getvalue(sql,rs,i,"ipmode"),ipmode)!=(char *)0) {
        char *ptr;

        ptr=strdup(sql_getvalue(sql,rs,i,"blsserver"));
        sblsites=(char **)vec_addlstlst((void **)sblsites,(void *)ptr);
	}
      }
    (void) sql_dropresult(sql,rs);
    }
  }
return sblsites;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to prepare a user info struct	*/
/*						*/
/************************************************/
USRTYP *dbc_mkuserinfo(char *rcpt,RCPTFLG status)

{
USRTYP *usr;

usr=(USRTYP *)calloc(1,sizeof(USRTYP));
usr->username=strdup(rcpt);
usr->rcptflg=status;
return usr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a specific user	*/
/*	is a local one.				*/
/*						*/
/************************************************/
void dbc_getuserinfo(SQLFILE *sql,USRTYP *usr)

{
int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*active domain?	*/
      if (dbc_getactive(sql,eml_getdomain(usr->username))!=1)
	phase=999;	/*domain NOT active	*/
      break;
    case 1	:	/*is user localy known	*/
      (void) dbc_getlocaluser(sql,usr);
      break;
    case 2	:	/*is user an alias	*/
      if ((usr->aliases=dbc_getaliaslist(sql,usr->username))!=(char **)0) {
	if ((usr->numalias=vec_sizelstlst((void **)usr->aliases))>0) {
	  if (usr->rcptflg==rcpt_unknown)
	    usr->rcptflg=rcpt_isalias;
	  else
	    usr->rcptflg=rcpt_islocalias;
	  }
	phase=999;	/*its an alias		*/
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by	*/
/*	user information			*/
/*						*/
/************************************************/
USRTYP *dbc_freeuserinfo(USRTYP *user)

{
if (user!=(USRTYP *)0) {
  user->aliases=(char **)vec_freelstlst((void **)user->aliases,VECFREE);
  user->password=clm_freestr(user->password);
  user->username=clm_freestr(user->username);
  (void) free(user);
  user=(USRTYP *)0;
  }
return user;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	profile definition			*/
/*						*/
/************************************************/
PRFTYP *dbc_freeprofile(PRFTYP *profile)

{
if (profile!=(PRFTYP *)0) {
  (void) free(profile);
  profile=(PRFTYP *)0;
  }
return profile;
}
/*
^L
*/
/************************************************/
/*	Procedure to return a profil, if a	*/
/*	specific profil is not found, send the	*/
/*	originator profile (profile 0)		*/
/*						*/
/************************************************/
PRFTYP *dbc_getemailprofile(SQLFILE *sql,char *user)

{
PRFTYP *prf;

prf=getprofile(sql,getcommonvalue(sql,user,"prfid",1));
prf->active=mngactive(user,getcommonvalue(sql,user,"active",0));
return prf;
}
/*
^L
*/
/************************************************/
/*	Procedure to free memory used by an	*/
/*	override structure.			*/
/*						*/
/************************************************/
OVRTYP **dbc_freeoverride(OVRTYP **ovrlst)

{
if (ovrlst!=(OVRTYP **)0) {
  int i;

  for (i=0;ovrlst[i]!=(OVRTYP *)0;i++) {
    register OVRTYP *ptr;

    ptr=ovrlst[i];
    ptr->orgip=clm_freestr(ptr->orgip);
    ptr->orgemail=clm_freestr(ptr->orgemail);
    (void) free(ptr);
    }
  (void) free(ovrlst);
  ovrlst=(OVRTYP **)0;
  }
return ovrlst;
}
/*
^L
*/
/************************************************/
/*	Procedure to insert an overrid record	*/
/*	within data-base.			*/
/*						*/
/************************************************/
int dbc_setoverride(SQLFILE *sql,char *recipient,char *originator,char *orgdom,char *iporg,int cidr,OVRENUM ovrride)

{
#define	OPOV	"gesdbc.c:dbc_setoverride,"

#define	INSOVR1	"INSERT INTO %s (ovrride,orgname,orgdom,orgip,recname,recdom) "
#define	INSOVR2	"VALUES (%d,%s,'%s','%s',%s,'%s')"
#define	UPDOVR1	"UPDATE %s SET ovrride=%d,lasthit=NOW() WHERE "
#define	UPDOVR2	"recdom='%s' AND recname=%s AND orgip='%s' AND orgdom='%s' and orgname=%s"
		 
static LCKTYP lck1={OVRRIDE,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
char *recname;
char *orgname;
char *recdomain;
char *ipclass;
char *action;
int phase;
int proceed;

done=true;
recdomain=eml_splitdomain(recipient);
recname=sql_cleanstr(sql,recipient);	/*MUST BE after splitdomain	*/
orgname=sql_cleanstr(sql,originator);
ipclass=dns_getipclass(iporg,cidr);
(void) asprintf(&action,UPDOVR1 UPDOVR2,OVRRIDE,ovrride,recdomain,recname,ipclass,orgdom,orgname);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets update teh record*/
      if (sql_action(sql,action)>0)
	phase=999;	/*found at least one	*/
      break;
    case 1	:	/*setting lock		*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPOV,OVRRIDE);
	done=false;
	phase=999;
	}
      break;
    case 2	:	/*updating		*/
      if (sql_action(sql,action)>0) 
	phase++;	/*race condition 	*/
      break;
    case 3	:	/*creating		*/
      (void) free(action);
      (void) asprintf(&action,INSOVR1 INSOVR2,OVRRIDE,ovrride,orgname,orgdom,ipclass,recname,recdomain);
      if (sql_action(sql,action)==0) {
	(void) clm_alert(0,"%s Unable to update %s table (bug?)",OPOV,OVRRIDE);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/	
	}
      break;
    case 4	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    } 
  phase++;
  }
(void) free(action);
(void) free(ipclass);
(void) free(orgname);
(void) free(recname);
return done;
}
/*
^L
*/
/************************************************/
/*	Procedure to return an override level	*/
/*	according special override directive	*/
/*	found with the data-base.		*/
/*						*/
/************************************************/
OVRTYP **dbc_getoverride(SQLFILE *sql,char *from,char *to)

{
#define	GETOVR	"SELECT * FROM %s WHERE recdom %s AND recname %s"

OVRTYP **ovrlst;
register const char *cd0;
char *orgdomain;
char *orgname;
char *recdomain;
char *recname;
int phase;
int proceed;

ovrlst=(OVRTYP **)0;
cd0="= '*'";
orgname=strdup(from);
orgdomain=eml_splitdomain(orgname);
recname=getnsearch(sql,to);
recdomain=getdsearch(to);
phase=0;
proceed=true;
while (proceed==true) {
  register const char *cd1;
  register const char *cd2;

  cd1=(char *)0;
  cd2=(char *)0;
  switch (phase) {
    case 0	:	/*overide for this user	*/
      cd1=recdomain;
      cd2=recname;
      break;
    case 1	:	/*overiding this username*/
      cd1=cd0;
      cd2=recname;
      break;
    case 2	:	/*overiding this domain	*/
      cd1=recdomain;
      cd2=cd0;
      break;
    case 3	:	/*overiding everything?	*/
      cd1=cd0;
      cd2=cd0;
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if ((cd1!=(char *)0)&&(cd2!=(char *)0)) {
    SQLRESULT *rs;

    if ((rs=sql_tupple(sql,GETOVR,OVRRIDE,cd1,cd2))!=(SQLRESULT *)0) {
      int nbr;

      if ((nbr=sql_nbrtupple(sql,rs))>0) {
        int i;
 
        for (i=0;i<nbr;i++) {
	  register OVRTYP *ptr;
	
	  ptr=getoneoverride(sql,orgname,orgdomain,recname,recdomain,rs,i);
	  ovrlst=(OVRTYP **)vec_addlstlst((void **)ovrlst,(void *)ptr);
	  }
	}
      (void) sql_dropresult(sql,rs);
      }
    }
  phase++;
  }
(void) free(recdomain);
(void) free(recname);
(void) free(orgname);
return ovrlst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update the override stamp	*/
/*	to mark the last hit.			*/
/*						*/
/************************************************/
void dbc_overhit(SQLFILE *sql,OVRTYP *over)

{
#define	OPOH	"gesdbc.c:dbc_overhit,"

#define	UPDOVR	"UPDATE %s SET lasthit=NOW() WHERE ovrid=%lu"

if (sql_action(sql,UPDOVR,OVRRIDE,over->ovrid)==0)
  (void) clm_alert(0,"%s Unable to update %s table (ovrid=%d)",OPOH,OVRRIDE,over->ovrid);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a list	*/
/*	file in quarantine.			*/
/*						*/
/************************************************/
EXTTYP **dbc_freeextract(EXTTYP **exts)

{
if (exts!=(EXTTYP **)0) {
  int i;

  for (i=0;exts[i]!=(EXTTYP *)0;i++) {
    (void) free(exts[i]->recipient); 
    (void) free(exts[i]->session); 
    (void) free(exts[i]);
    }
  (void) free(exts);
  exts=(EXTTYP **)0;
  }
return exts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract from the database	*/
/*	a list of session/recipient used to	*/
/*	extract quarantined email from		*/
/*	depository.				*/
/*						*/
/************************************************/
EXTTYP **dbc_getextract(SQLFILE *sql)

{
#define	OPGX	"gesdbc.c:dbc_getextract,"

#define	DELEXT	"DELETE FROM %s WHERE session='%s' AND recipient=%s"
#define	SELEXT	"SELECT * FROM %s ORDER BY stamp asc"

static LCKTYP lck1={EXTRACT,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

EXTTYP **exts;
SQLRESULT *rs;
int nbr;
int phase;
int proceed;

exts=(EXTTYP **)0;
rs=(SQLRESULT *)0;
nbr=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting lock access	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPGX,EXTRACT);
	}
      break;
    case 1	:	/*extracting record	*/
      if ((rs=sql_tupple(sql,SELEXT,EXTRACT))!=(SQLRESULT *)0){
	nbr=sql_nbrtupple(sql,rs);
	if (nbr>0) {
	  int i;

	  for (i=0;i<nbr;i++) {
	    char *session;
	    char *recipient;
	    char *rcpt;
	    MENTYP status;
	    int trouble;

	    trouble=false;
	    session=sql_getvalue(sql,rs,i,"session");
	    recipient=sql_getvalue(sql,rs,i,"recipient");
	    rcpt=sql_cleanstr(sql,recipient);
	    status=getemlogstatus(sql,session,rcpt);
	    if (status!=m_undefined) {
	      EXTTYP *extloc;

	      extloc=(EXTTYP *)calloc(1,sizeof(EXTTYP));
	      extloc->session=sql_strdup(session);
	      extloc->recipient=sql_strdup(recipient);
	      extloc->status=status;
	      exts=(EXTTYP **)vec_addlstlst((void **)exts,(void *)extloc);
      	      if (sql_action(sql,DELEXT,EXTRACT,session,rcpt)==0) {
        	(void) clm_alert(0,"%s Unable to delete record from table %s (session='%s',recipient='%s' bug?)",
				    OPGX,EXTRACT,session,recipient);
		trouble=true;
		}
	      if (dbc_updlogs(sql,session,recipient,m_progress)==false) {
        	(void) clm_alert(0,"%s Unable to update record from table %s (session='%s',recipient='%s' bug?)",
				    OPGX,EMLOGS,session,recipient);
		trouble=true;
		}
	      if (trouble==true) {
      		(void) sql_unlock_rollback(sql);
		phase=999;	/*trouble trouble*/
		break;
		}
	      }
	    (void) free(rcpt);
	    }
	  }
        (void) sql_dropresult(sql,rs);
	}
      break;
    case 2	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return exts;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to insert a volatil entry	*/
/*	about an internal user within the data	*/
/*	base. Purpose is to avoid to request	*/
/*	over and over if user is existing	*/
/*	to the internal server.			*/
/*						*/
/************************************************/
int dbc_setintusr(SQLFILE *sql,char *rcpt,RCPTFLG rcptflg)

{
#define	OPIN	"gesdbc.c:dbc_setintusr"

#define	UPDIUSR	"UPDATE %s SET status=%d,upto='%s',lastchk=NOW() WHERE username='%s'"
#define	INSIUSR	"INSERT INTO %s (username,status,upto,lastchk) VALUES('%s',%d,'%s',NOW())"

static LCKTYP lck1={INTUSR,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

int done;
register char *UPTO;
long offset;
int phase;
int proceed;

done=true;
UPTO="";
offset=31*H24;	/*1 Month			*/
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*updating record	*/
      switch (rcptflg) {
	case rcpt_delayed	:
	  offset=60;		/*One minute	*/
	  break;
	case rcpt_unknown	:
	  offset=(30*60);	/*30 minutes	*/
	  break;
	case rcpt_forward	:
	default			:
	  break;
	}
      UPTO=sql_fromunixtime(sql,time((time_t *)0)+offset);
      if (sql_action(sql,UPDIUSR,INTUSR,rcptflg,UPTO,rcpt)==1)
        phase=999;	/*record already present*/
      break;
    case 1	:	/*locking database table*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPIN,INTUSR);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*double checking record*/
      if (sql_action(sql,UPDIUSR,INTUSR,rcptflg,UPTO,rcpt)==1) 
        phase++;	/*record just created! 	*/
      break;
    case 3	:	/*creating record	*/
      if (sql_action(sql,INSIUSR,INTUSR,rcpt,rcptflg,UPTO)!=1) {
        (void) clm_alert(0,"%s Unable to create internal user =<%s>",OPIN,rcpt);
        (void) sql_unlock_rollback(sql);
	done=false;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*unlocking db table	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if we have already	*/
/*	an internal user status within the	*/
/*	database.				*/
/*						*/
/************************************************/
RCPTFLG dbc_getintusr(SQLFILE *sql,char *rcpt)

{
#define	GETINTUSR	"SELECT * FROM %s WHERE username='%s'"

RCPTFLG rcptflg;
SQLRESULT *rs;

rcptflg=rcpt_uknstatus;
if ((rs=sql_tupple(sql,GETINTUSR,INTUSR,rcpt))!=(SQLRESULT *)0) {
  int nbr;

  if ((nbr=sql_nbrtupple(sql,rs))>0) {

    rcptflg=(RCPTFLG)atoi(sql_getvalue(sql,rs,0,"status"));
    if (rcptflg==rcpt_forward) {
      time_t lastchk;

      lastchk=sql_tounixtime(sql,sql_getvalue(sql,rs,0,"lastchk"));
      if ((lastchk+H24)<time((time_t *)0))
	rcptflg=rcpt_uknstatus;	/*asking for a refresh	*/
      }
    }
  (void) sql_dropresult(sql,rs);
  }
return rcptflg;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to cleanup data-base dynamique*/
/*	table.					*/
/*	If the 'flush' variable is set, delete	*/
/*	all table contents, regardless of the	*/
/*	current time.				*/
/*						*/
/************************************************/
void dbc_dbcleanup(SQLFILE *sql)

{
#define	DELINU	"DELETE FROM %s WHERE upto<NOW()"

static LCKTYP lck1={INTUSR,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,(LCKTYP *)0};

time_t curtime;
int phase;
int proceed;

curtime=time((time_t *)0);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*cleaning trouble queue*/
      (void) cleantrouble(sql,curtime);
      break;
    case 1	:	/*locking database table*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table","gesdbc.c:dbc_dbcleanup,",INTUSR);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*cleaning old intusers	*/
      (void) sql_action(sql,DELINU,INTUSR);
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to create a set of queued	*/
/*	entry about Email to be transited.	*/
/*						*/
/************************************************/
int dbc_createqueues(SQLFILE *sql,QUETYP **queues)

{
#define	OPCR	"gesdbc.c:dbc_createqueue,"

static LCKTYP lck1={DOMS,dbd_lockread};
static LCKTYP lck2={INQUEUE,dbd_lockwrite};
static LCKTYP *lcklst[]={&lck1,&lck2,(LCKTYP *)0};

int numdone;
int proceed;
int phase;

numdone=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking database	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_locksoft %s table",OPCR,DOMS);
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*inserting record	*/
      if (queues!=(QUETYP **)0) {
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  if (setqueue(sql,queues[i])==false) {
	    numdone=0;
	    (void) sql_unlock_rollback(sql);
	    phase=999;	/*trouble trouble	*/
	    break;
	    }
	  numdone++;
	  }
        }
      break;
    case 2	:	/*committing result	*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return numdone;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to collect Email ready to be	*/
/*	transmitted. The Flag 'progress' is	*/
/*	set to 1 (inprogress).			*/
/*						*/
/************************************************/
QUETYP **dbc_getqueues(SQLFILE *sql,int max)

{
#define	OPGQ	"gesdbc.c:dbc_getqueues,"

#define	INPROG	"UPDATE %s SET progress=1,last=NOW(),retry=retry+1 WHERE session='%s' AND transit=%d"

static LCKTYP lck1={INQUEUE,dbd_lockwrite};
static LCKTYP lck2={EMLOGS,dbd_lockread};
static LCKTYP lck3={ALIASES,dbd_lockread};
static LCKTYP lck4={USERS,dbd_lockread};
static LCKTYP *lcklst[]={&lck1,&lck2,&lck3,&lck4,(LCKTYP *)0};

QUETYP **queues;
int proceed;
int phase;

queues=(QUETYP **)0;
proceed=true;
phase=0;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*locking EMLOG READ	*/
      if (sql_getlock(sql,lcklst)<0) {
        (void) clm_alert(0,"%s Unable to dbd_lockhard %s table",OPGQ,INQUEUE);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*extracting ready	*/
      if ((queues=getpendingqueue(sql,max))!=(QUETYP **)0) {
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  if (sql_action(sql,INPROG,INQUEUE,queues[i]->session,queues[i]->transit)!=1) {
            (void) clm_alert(0,"%s Unable flag queue 'progress' session=<%s> transit='%d' (bug?)",
				OPCR,queues[i]->session,queues[i]->transit);
            (void) sql_unlock_rollback(sql);
	    queues=que_freequeues(queues);
	    phase=999;	/*trouble trouble	*/
	    break;
	    }
	  }
	}
      break;
    case 2	:	/*unlocking		*/
      (void) sql_unlock_commit(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return queues;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update inqueue email data	*/
/*	transmitted. The Flag 'progress' is	*/
/*	set to 1 (inprogress).			*/
/*						*/
/************************************************/
int dbc_updqueue(SQLFILE *sql,QUETYP *q)

{
#define	UPDQUE	\
 "UPDATE %s SET next='%s',last='%s',retry=%d,progress=%d,status=%d,why=%s,numtrace=%lu WHERE session='%s' AND transit=%d"

int num;
char *last;
char *next;
char *why;

last=strdup(sql_fromunixtime(sql,q->lasttime));
next=strdup(sql_fromunixtime(sql,q->nexttime));
why=sql_cleanstr(sql,q->why);
num=sql_action(sql,UPDQUE,INQUEUE,next,last,q->retry,q->progress,q->status,why,q->numtrace,q->session,q->transit);
(void) free(why);
(void) free(next);
(void) free(last);
return num;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to move an entrie from the	*/
/*	inqueue list in ready mode.		*/
/*						*/
/************************************************/
int dbc_rdyqueue(SQLFILE *sql,QUETYP *queue)

{
#define	RDYQUE	"UPDATE %s SET progress=0 WHERE session='%s' AND transit=%d"

int done;

done=true;
if (queue!=(QUETYP *)0) {
  if (sql_action(sql,RDYQUE,INQUEUE,queue->session,queue->transit)<1) {
    (void) clm_alert(0,"%s Unable to make inqueue session=<%s> transit='%d' ready",
			"gesdbc.c:dbc_rdyqueue,",queue->session,queue->transit);
    done=false;
    }
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove an entrie from the	*/
/*	inqueue list.				*/
/*						*/
/************************************************/
int dbc_delqueue(SQLFILE *sql,QUETYP *queue)

{
#define	DELQUE	"DELETE FROM %s WHERE session='%s' AND transit=%d"

int done;

done=true;
if (queue!=(QUETYP *)0) {
  if (sql_action(sql,DELQUE,INQUEUE,queue->session,queue->transit)<1) {
    (void) clm_alert(0,"%s Unable to delete inqueue session=<%s> transit='%d'",
			"gesdbc.c:dbc_delqueue,",queue->session,queue->transit);
    done=false;
    }
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if the proposed	*/
/*	domain is in fact a domain alias.	*/
/*						*/
/************************************************/
char *dbc_isitadomainalias(SQLFILE *sql,char *domain)

{
#define GETDOMAL	"SELECT * FROM %s WHERE alias=%s"
char *domalias;
char *cleandom;
SQLRESULT *rs;

domalias=(char *)0;
cleandom=sql_cleanstr(sql,domain);
if ((rs=sql_tupple(sql,GETDOMAL,DOMAL,cleandom))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) 
    domalias=sql_strdup(sql_getvalue(sql,rs,0,"domainname"));
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleandom);
return domalias;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to set list of quarantined 	*/
/*	email to extract queue when local user	*/
/*	is sending Email to a remote user.	*/
/*						*/
/************************************************/
void dbc_purgequar(SQLFILE *sql,char *originator,char *recipient)


{
#define	OPPU	"gesdbc.c:dbc_purgequar,"
#define	FPUR1	"session,recipient"
#define	PURGE	"INSERT INTO %s (%s) (SELECT %s FROM %s WHERE status=%d AND recipient=%s AND originator=%s)"

char *orgeml;
char *receml;

orgeml=sql_cleanstr(sql,originator);
receml=sql_cleanstr(sql,recipient);
if (sql_action(sql,PURGE,EXTRACT,FPUR1,FPUR1,EMLOGS,m_quarantine,orgeml,receml)<0) {
  char *cmd;
 
  (void) asprintf(&cmd,PURGE,EXTRACT,FPUR1,FPUR1,EMLOGS,m_quarantine,orgeml,receml);
  (void) clm_alert(0,"%s Unable to do SQL request=<%s> (bug?)",OPPU,cmd);
  (void) free(cmd);
  }
(void) free(receml);
(void) free(orgeml);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return true if the 	*/
/*	authenticated originator sent too many	*/
/*	outside its own group.			*/
/*						*/
/************************************************/
int dbc_toomanymail(SQLFILE *sql,char *authemail,int *sent)

{
#define	OPOO	"gesdbc.c:dbc_toomanymail,"

int toomany;
PRFTYP *prf;
u_int grpid;
int phase;
int proceed;

toomany=false;
*sent=0;
prf=(PRFTYP *)0;
grpid=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting user profile	*/
      if ((prf=dbc_getemailprofile(sql,authemail))==(PRFTYP *)0) {
        (void) clm_alert(0,"%s Unable to find profile for user <%s> (bug?)",OPOO,authemail);
	proceed=false;
        }
      break;
    case 1	:	/*getting group id	*/
      if ((grpid=getgroupid(sql,authemail))==0) {
        (void) clm_alert(0,"%s Unable to find groupid for user <%s> (bug?)",OPOO,authemail);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*checking Email count	*/
      *sent=getcount(sql,authemail,grpid);
      if (*sent>prf->maxperday)
	toomany=true;
      break;
    default	:	/*safe Guard		*/
      prf=dbc_freeprofile(prf);
      proceed=false;
      break;
    }
  phase++;
  }
return toomany;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to retrieve the domain	*/
/*	activation level			*/
/*						*/
/************************************************/
DACTYP dbc_getactive(SQLFILE *sql,char *domain)

{
DACTYP active;
time_t upto;
char *cleandom;
SQLRESULT *rs;

active=da_unknown;
cleandom=sql_cleanstr(sql,domain);
if ((rs=sql_tupple(sql,SELDOM,SELFROM,DOMS,cleandom))!=(SQLRESULT *)0) {
  if (sql_nbrtupple(sql,rs)>0) {
    switch ((atoi)(sql_getvalue(sql,rs,0,"active"))) {
      case 0	:	/*domain is locked	*/
	active=da_lock;
	break;
      case 1	:	/*domain is accepted	*/
	active=da_good;
        upto=getuptograce(sql,rs,0);
        if ((upto>0)&&(upto<time((time_t *)0)))
	  active=da_lock; 	/*domain expired*/
	break;
      case 2	:	/*domain is to resend	*/
	active=da_resend;
	break;
      default	:	/*unexpected status	*/
	(void) clm_alert(0,"%s Unexpected active status '%d' on domain <%s>",
			   "gesdbc.c:dbc_getactive,",(atoi)(sql_getvalue(sql,rs,0,"active")),domain);
	break;
      }
    }
  (void) sql_dropresult(sql,rs);
  }
(void) free(cleandom);
return active;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a DSP structure from	*/
/*	the data collationned within emlogs	*/
/*	table.					*/
/*						*/
/************************************************/
DSPTYP*dbc_getsessiondsp(SQLFILE *sql,char *session,char *recipient)

{
DSPTYP *dsp;
LOGTYP *emlog;
int phase;
int proceed;

dsp=(DSPTYP *)0;
emlog=(LOGTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting log info	*/ 
      if ((emlog=dbc_getlogs(sql,session,recipient))==(LOGTYP *)0)
	phase=999;	/*no log? trouble	*/
      break;
    case 1	:	/*converting log to dsp	*/
      dsp=(DSPTYP *)calloc(1,sizeof(DSPTYP));
      dsp->spmlvl=NAN;
      dsp->spmlim=NAN;
      dsp->session=emlog->session;
      dsp->peerip=emlog->rmtip;
      dsp->efrom=emlog->efrom;
      dsp->hfrom=emlog->hfrom;
      dsp->subject=emlog->subject;
      break;
    case 2	:	/*cleaning log		*/
      emlog->session=(char *)0;
      emlog->rmtip=(char *)0;
      emlog->efrom=(char *)0;
      emlog->hfrom=(char *)0;
      emlog->subject=(char *)0;
      emlog=dbc_freelogs(emlog);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return dsp;
}
/*
 * ^L
 * */
/************************************************/
/*						*/
/*	Procedure to setup a clean from data	*/
/*	for each recipient, this can be not	*/
/*	the same within the same email, if an	*/
/*	alias recipient expension is used.	*/
/*						*/
/************************************************/
void dbc_setinbehalf(SQLFILE *sql,RCPTYP **rcpts,char *inbehalf)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) 
    rcpts[i]->inbehalf=strdup(inbehalf);
  }
}

