/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	Define module level routines to handle	*/
/*	clement data-base.			*/
/*						*/
/************************************************/
#ifndef GESDBC
#define GESDBC
#include	<time.h>
#include	"unieml.h"
#include	"unidev.h"
#include	"unisql.h"
#include	"unique.h"

#define	EMAILSZ	0	/*index to email size	*/
#define	BOXESZ	1	/*index to box size	*/

typedef	struct	{
	u_int maxemail;	/*maximum email size	*/
	u_int maxbox;	/*maximum mail box size	*/
	}MAXSIZ;

typedef	struct	logtyp {/*log information	*/
	char *session;	/*unique session id	*/
	char *rmtip;	/*remote ipnum		*/
	char *efrom;	/*extracted SMTP from	*/
	char *hfrom;	/*email header from	*/
	char *recipient;/*email recipient	*/
	char *subject;	/*email subject		*/
	}LOGTYP;

typedef	struct	usrtyp {
	RCPTFLG rcptflg;/*user rcpt flag	*/
	char *username;	/*full user name	*/
	int domover;	/*domain override.	*/
	int active;	/*is user active	*/
	time_t upto;	/*account date validity	*/
	int prfid;	/*profile name		*/
	int grpid;	/*group id		*/
	int lang;	/*user language		*/
	char *password;	/*user current password	*/
	u_int numalias;	/*aliases expension size*/
	char **aliases;	/*user aliases list	*/
	}USRTYP;

typedef	struct grptyp	{
	int groupid;	/*group unique uumber	*/
	char *grpname;	/*group name		*/
	char *smtprmt;	/*auxilary SMTP server	*/
	char *smtpport;	/*auxilary SMTP port	*/
	char *authrmt;	/*auxilary authenticator*/
	char *authport;	/*rmt authenticator port*/
	char *authprot;	/*authenticator protocol*/
	char *exitip;	/*exiting IP		*/
	char *subject;	/*log subject		*/
	time_t upto;	/*group expiration date	*/
	int prfid;	/*group profile ID	*/
	int active;	/*is domain active	*/
	}GRPTYP;

typedef	struct prftyp	{
	int system;	/*empty profile	flag	*/
	u_int prfid;	/*profile ID		*/
	u_long emailsz;	/*max email size	*/
	int blsid;	/*BLS list to check	*/
	int checkrmtspf;/*Check remote SPF	*/
	int checkrmtmx;	/*check remote MX	*/
	u_int mxmgbytes;/*eml max space allowed	*/
	u_int maxperday;/*max # of email allowed*/
	u_int bulk;	/*number of RCPT allowed*/
	u_int active;	/*is the account active	*/
	int lang;	/*status language	*/
	}PRFTYP;

typedef enum {tr_banned,tr_bypass,tr_accepted,tr_maybe} OVRENUM;

typedef	struct ovrtyp	{
	u_long ovrid;	/*override unique id	*/
	OVRENUM ovrride;/*override value	*/
	char *orgemail;	/*originator email	*/
	char *orgip;	/*IP allowed origine	*/
	}OVRTYP;

typedef	struct		{
	char *srvname;	/*service reference name*/
	int preference;	/*service preference	*/
	char *srvhost;	/*service server name	*/
	char *srvport;	/*service access port	*/
	char *srvaccess;/*service access data	*/
	}SRVLST;

/*quarantine extraction list			*/
typedef	struct	{
	char *session;	/*session name		*/
	char *recipient;/*recipient Email	*/
	MENTYP status;	/*transaction status	*/
	}EXTTYP;

/*Transaction data				*/
typedef	struct	{
	char *session;	/*session unique ID	*/
	time_t first;	/*first insert		*/
	time_t last;	/*last action date	*/
	time_t next;	/*next action date	*/
	int retried;	/*number of retry	*/
	char *mfrom;	/*email originator	*/	
	char *recipient;/*email recipient	*/	
	char *subject;	/*email subject		*/
	char *why;	/*transit explanaition	*/
	MENTYP status;	/*transaction status	*/
	}TRDTYP;

extern int dbc_settrtdata(SQLFILE *sql,QUETYP *queue);
extern char **dbc_getaliaslist(SQLFILE *sql,char *rcpt);
extern void dbc_getlocaluser(SQLFILE *sql,USRTYP *usr);
extern SRVLST **dbc_getsrvlst(SQLFILE *sql,char *srvname);
extern SRVLST **dbc_freesrvlst(SRVLST **srvlst);
extern int dbc_setlogs(SQLFILE *sql,DSPTYP *dsp,RCPTYP **rcpts);
extern LOGTYP *dbc_freelogs(LOGTYP *mlog);
extern LOGTYP *dbc_getlogs(SQLFILE *sql,char *session,char *recipient);
extern int dbc_updlogs(SQLFILE *sql,char *session,char *recipient,MENTYP status);
extern MAXSIZ *dbc_getmaxes(SQLFILE *sql,char *rcpt);
extern time_t dbc_checkbind(SQLFILE *sql);
extern BININF **dbc_getbinds(SQLFILE *sql,PROTYP proto);
extern int dbc_adddomtogroup(SQLFILE *sql,GRPTYP *grp,char *domain);
extern char *dbc_getdfltdom(SQLFILE *sql);
extern char *dbc_getdfltfrom(SQLFILE *sql,char *domain);
extern GRPTYP *dbc_freegroup(GRPTYP *group);
extern GRPTYP *dbc_getgroup(SQLFILE *sql,char *rcpt);
extern GRPTYP **dbc_getgrouplst(SQLFILE *sql);
extern char **dbc_getblssites(SQLFILE *sql,int blsid,int af_net);
extern USRTYP *dbc_mkuserinfo(char *rcpt,RCPTFLG status);
extern void dbc_getuserinfo(SQLFILE *sql,USRTYP *usr);
extern USRTYP *dbc_freeuserinfo(USRTYP *user);
extern PRFTYP *dbc_freeprofile(PRFTYP *profile);
extern PRFTYP *dbc_getemailprofile(SQLFILE *sql,char *user);
extern OVRTYP **dbc_freeoverride(OVRTYP **ovrlst);
extern int dbc_setoverride(SQLFILE *sql,char *recipient,char *orgname,char *orgdom,char *iporg,int cidr,OVRENUM ovrride);
extern OVRTYP **dbc_getoverride(SQLFILE *sql,char *from,char *to);
extern void dbc_overhit(SQLFILE *sql,OVRTYP *over);
extern EXTTYP **dbc_freeextract(EXTTYP **exts);
extern EXTTYP **dbc_getextract(SQLFILE *sql);
extern int dbc_setintusr(SQLFILE *sql,char *rcpt,RCPTFLG rcptflg);
extern RCPTFLG dbc_getintusr(SQLFILE *sql,char *rcpt);
extern void dbc_dbcleanup(SQLFILE *sql);
extern int dbc_createqueues(SQLFILE *sql,QUETYP **queues);
extern QUETYP **dbc_getqueues(SQLFILE *sql,int max);
extern int dbc_updqueue(SQLFILE *sql,QUETYP *queue);
extern int dbc_rdyqueue(SQLFILE *sql,QUETYP *queue);
extern int dbc_delqueue(SQLFILE *sql,QUETYP *queue);
extern char *dbc_isitadomainalias(SQLFILE *sql,char *domain);
extern void dbc_purgequar(SQLFILE *sql,char *originator,char *recipient);
extern int dbc_toomanymail(SQLFILE *sql,char *authemail,int *sent);
extern DACTYP dbc_getactive(SQLFILE *sql,char *domain);
extern DSPTYP*dbc_getsessiondsp(SQLFILE *sql,char *session,char *recipient);
extern void dbc_setinbehalf(SQLFILE *sql,RCPTYP **rcpts,char *inbehalf);
#endif
