/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLCNF:					*/
/*	Take care of all clement configuration	*/
/*	structure reading from the data-base.	*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#include	<netdb.h>
#include	<arpa/inet.h>
#include	<ctype.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"gesdbc.h"
#include	"gescnf.h"
#include	"lvlcnf.h"
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a config structure	*/
/*						*/
/************************************************/
CNFTYP *cnf_chkconfig(CNFTYP *cnf)

{
/*checking database update every UPDCYCLE sec	*/
#define	UPDCYLE	60	

const char *empty;
time_t curtime;
time_t curupd;
time_t lstupd;
time_t update;
SQLFILE *sql;
int phase;
int proceed;

empty=(const char *)0;
curtime=time((time_t *)0);
curupd=(time_t)0;
lstupd=(time_t)0;
update=(time_t)0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*config already loaded?*/
      if (cnf!=(CNFTYP *)0) {
	cnf->updated=false;
	update=cnf->update;
	if ((curtime+UPDCYLE)>cnf->lastscan) {
	  proceed=false;/*no need to recheck	*/
	  }
        }
      break;
    case  1	:	/*do wee need to reload	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	(void) clm_alert(0,"gestbl.c,tbl_getconfig: unable open database");
	cnf=cnf_freeconfig(cnf);
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case  2	:	/*was a binding update?	*/
      lstupd=dbc_checkbind(sql);
      if (lstupd>update)
	curupd=lstupd;
      break;
    case  3	:	/*do we need to update?	*/
      if (curupd<=update)
	phase=999;	/*no need to go further	*/
      break;
    case  4	:	/*reloading config	*/
      cnf=cnf_freeconfig(cnf);
      cnf=(CNFTYP *)calloc(1,sizeof(CNFTYP));
      cnf->updated=true;
      cnf->lastscan=curtime;
      cnf->update=curupd;
      break;
    case  5	:	/*loading the new config*/
      cnf->binds=dbc_getbinds(sql,pro_smtp);
      break;
    default	:	/*SAFE guard		*/
      sql=sql_close(sql);
      proceed=false;
      break;
    }
  phase++;
  }
return cnf;
}
