/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLTRT:					*/
/*	Email transit management.		*/
/*						*/
/************************************************/
#include	<time.h>
#include	<errno.h>
#include	<syslog.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<pwd.h>
#include	<sys/stat.h>
#include	<string.h>
#include	"dbgmem.h"
#include	"unijrl.h"
#include	"gesdbc.h"
#include	"gesque.h"
#include	"lvltrt.h"

#define	OKQ	0	/*queue ready to be pick*/
#define	INUSEQ	1	/*queue in usage	*/
#define	TRBLQ	2	/*queue moved to trouble*/
/*

*/
/************************************************/
/*						*/
/*	Procedure to move a data-file within	*/
/*	the trouble queue			*/
/*						*/
/************************************************/
static int totrblq(SQLFILE *sql,QUETYP *q,time_t curtime)

{
#define	OPTR	"lvltrt.c:totrblq,"


int done;
char *datafile;
char *newfile;

done=true;
datafile=que_getdatafilename(d_transit,q);
newfile=que_getdatafilename(d_trblq,q);
if (link(datafile,newfile)<0) {
  done=false;		/*oups trouble		*/
  switch (errno) {
    case EXDEV	:
      done=clm_dupfile(datafile,newfile);
      break;
    default	:
      (void) clm_alert(0,"%s Unable to move file <%s> to <%s> (error=<%s>)",
			  OPTR,datafile,newfile,strerror(errno));
      break;
    }
  }
(void) unlink(datafile);
(void) free(newfile);
(void) free(datafile);
q->nexttime=curtime+(360*24*3600);
q->lasttime=curtime;
q->progress=TRBLQ;
q->retry++;
if (q->why!=(char *)0) {
  char *ptr;

  while ((ptr=strchr(q->why,'\n'))!=(char *)0)
    *ptr=' ';
  while ((ptr=strchr(q->why,'\r'))!=(char *)0)
    *ptr=' ';
  }
if (done==true) {
  (void) jrl_journalling(false,EVTRT,"Moving transaction %s,<%s@%s> to trouble queue",
	      			      q->session,q->recname,q->recdom);
  (void) dbc_updqueue(sql,q);
  }
else
  (void) dbc_delqueue(sql,q);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove a transaction from	*/
/*	Queuing reference.			*/
/*						*/
/************************************************/
static int delqueue(SQLFILE *sql,QUETYP *queue)

{
#define	OPDL	"lvltrt.c:delqueue,"

int done;
char *datafile;

done=true;
datafile=que_getdatafilename(d_transit,queue);
if (unlink(datafile)<0) {
  (void) clm_alert(0,"%s Unable to remove file <%s> (error=<%s>) (bug?)",
		      OPDL,datafile,strerror(errno));
  }
(void) lvl_delqueue(sql,queue);
(void) free(datafile);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage queue status.	*/
/*	Removing data file if email transit	*/
/*	reached a final status.			*/
/*						*/
/************************************************/
QUETYP **trt_cleanqueues(SQLFILE *sql,QUETYP **queues)

{
#define	OPCL	"lvltrt.c:trt_cleanqueues,"

if (queues!=(QUETYP **)0) {
  time_t curtime;
  int i;

  curtime=time((time_t *)0);
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    switch (queues[i]->status) {
      case RETRY	:	/*rmt not ready	*/
      case INRELAY	:	/*no contact	*/
	queues[i]->progress=OKQ;
        (void) dbc_updqueue(sql,queues[i]);
	break;
      case FAILED	:	/*big failure	*/
      case BOUNCE	:	/*bounced	*/
	queues[i]->progress=TRBLQ;
        (void) dbc_updqueue(sql,queues[i]);
	break;
      case CMDOK	:	/*transmitted	*/
	(void) delqueue(sql,queues[i]);
	break;
      default		:	/*move it to trbl queue*/
	(void) clm_alert(0,"%s Unexpected status='%d' for transaction <%s-%d> (bug?)",
			    OPCL,queues[i]->status,queues[i]->session,queues[i]->transit);
	(void) totrblq(sql,queues[i],curtime);
	break;
      }
    queues[i]=que_freequeue(queues[i]);
    }
  (void) free(queues);
  queues=(QUETYP **)0;
  }
return queues;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove trace and queue log	*/
/*						*/
/************************************************/
void lvl_delqueue(SQLFILE *sql,QUETYP *queue) 

{
if (queue!=(QUETYP *)0) {
  char *tracename;

  tracename=que_tracename(queue);
  (void) unlink(tracename);
  (void) free(tracename);
  (void) dbc_delqueue(sql,queue);	/*removing queue*/
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to write down the resend file	*/
/*	within the transit directory.		*/
/*						*/
/************************************************/
u_int trt_transit(SQLFILE *sql,char *outip,char *smtpmx,char *smtpport,
		  char *from,DSPTYP *dsp,char *infile,RCPTYP **rcpts)

{
#define	OPSI	"lvltrt.c:trt_transit,"

int numdone;
QUETYP **queues;
int phase;
int proceed;

numdone=0;
queues=(QUETYP **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*preparing data		*/
      queues=que_mkqueues(outip,smtpmx,smtpport,dsp->session,mm_std,dsp->subject,dsp->rejectto,from,rcpts);
      break;
    case   1	:	/*checking for trouble		*/
      if (queues==(QUETYP **)0) {
	(void) clm_alert(0,"%s Unable to make QUETYP (bug!)",OPSI);
	phase=999;	/*we are really in big trouble	*/
  	}
      break;
    case   2	:	/*dumping transit record	*/
      if (dbc_createqueues(sql,queues)==0) {
	(void) clm_alert(0,"%s Unable to update database with queue  (bug!)",OPSI);
	phase=999;	/*we are really in big trouble	*/
	}
      break;
    case   3	:	/*dumping transit record	*/
      if (queues!=(QUETYP **)0) {	/*always	*/
	char *newshead;
	int i;

        (void) asprintf(&newshead,"Set New transit %s",dsp->session);
	(void) clm_fasprintf(&newshead,"\n\t\t%08s from <%s>","",from);
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  int ready;
	  char *cmt;
	  char *lfile;

	  ready=true;
	  cmt=""; 
	  lfile=que_getdatafilename(d_transit,queues[i]);
	  if (link(infile,lfile)<0) {
	    (void) clm_alert(0,"%s Unable to link file <%s> to file <%s> (errno=<%s>))",
				OPSI,infile,lfile,strerror(errno));
	    cmt=" not queued (trouble?)";
	    ready=false;
	    }
	  (void) clm_fasprintf(&newshead,"\n\t\t%08d to   <%s@%s>%s",
					  queues[i]->transit,queues[i]->recname,queues[i]->recdom,cmt);
	  (void) free(lfile);
	  if (ready==true)
            (void) dbc_rdyqueue(sql,queues[i]);
	  numdone++;
          }
	(void) jrl_journalling(false,EVCKR,"%s",newshead);
	(void) free(newshead);
	}
      break;
    default	:	/*SAFE guard			*/
      queues=que_freequeues(queues);
      proceed=false;
      break;
    }
  phase++;
  }
return numdone;
}
