/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODPRO:					*/
/*	Module doing protocol handling		*/
/*						*/
/************************************************/
#include	<string.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subprc.h"
#include	"subclm.h"
#include	"bascnv.h"
#include	"unijrl.h"
#include	"unidev.h"
#include	"uniper.h"
#include	"unieml.h"
#include	"lvlsta.h"
#include	"lvleml.h"
#include	"lvlcla.h"
#include	"modpro.h"

/*flagged error messages			*/
#define	SOULALR	"ALR:"
#define	SOULNRL	"NRL:"
#define	SOULUKN	"UKN:"
#define	SOULDNS	"DNS:"
#define	SOULSPF	"SPF:"
#define	SOULBLS	"BLS:"
/*

*/
/************************************************/
/*						*/
/*	Procedure to log E-mail header		*/
/*	information within the journal logs.	*/
/*						*/
/************************************************/
static void reporthdrs(EVLTYP *evl)

{
static char *tetes[]={
		"E-From",
		"Subject",
		"Message-Id",
		"Spam-lvl",
		"Email size",
	        (char *)0
		};

if (evl!=(EVLTYP *)0) {
  register int i;

  for (i=0;tetes[i]!=(char *)0;i++) {
    register char *ptr;

    ptr=(char *)0;
    switch(i) {
      case 0		:
	ptr=evl->realfrom;
        break;
      case 1		:
	ptr=evl->subject;
        break;
      case 2		:
	ptr=evl->msgid;
        break;
      case 3		:
	ptr=evl->spamlvl;
        break;
      case 4		:
        (void) jrl_journalling(false,EVJRL,"%s: %lu",tetes[i],evl->mtaille);
        break;
      default		:
	break;
      }
    if (ptr!=(char *)0) 
      (void) jrl_journalling(false,EVJRL,"%s: %s",tetes[i],ptr);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to ask extended advice	*/
/*						*/
/************************************************/
static void askadvices(BPTR *bptr,u_long mntpool,CTTYP *cta)

{
#define	OPMN	"modpro.c:askadvices,"

#define MXWMEN	10/*max time waiting for mentor	*/

char *rqst;
char *msgref;
char *answ;
int phase;
int proceed;

rqst=(char *)0;
msgref=(char *)0;
answ=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*makeking advice rqst	*/
      if ((rqst=eml_cttoadvmsg(cta))==(char *)0) {
	(void) clm_alert(0,"%s Unable to make an ADV msg (bug?)",OPMN);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*sending message	*/
      if (rqst!=(char *)0) {	/*always	*/
	TIMESPEC timeout;

	timeout.tv_sec=(time_t)MXWMEN;
	timeout.tv_nsec=(u_long)0;
        if ((msgref=ipc_postmsg(bptr,&timeout,mntpool,rqst))==(char *)0) {
	  (void) clm_alert(0,"%s Unable to send ADV msg <%s> in due time ",OPMN,rqst);
          (void) free(rqst);
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    case 2	:	/*expecting answer	*/
      if (msgref!=(char *)0) {	/*always	*/
	TIMESPEC timeout;

	timeout.tv_sec=(time_t)MXWMEN*2;
	timeout.tv_nsec=(u_long)0;
	if ((answ=ipc_getreply(bptr,&timeout,msgref))==(char *)0) {
	  (void) clm_alert(0,"%s Unable to get ADV reply tp msg <%s> in due time ",OPMN,rqst);
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(msgref);
	}
      (void) free(rqst);
      break;
    case 3	:	/*updating rcpt status	*/
      (void) eml_advmsgupdrcpt(answ,cta->rcpts);
      (void) free(answ);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;	
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to make an empty rcpt, used	*/
/*	when connection is rejected before any	*/
/*	RCPT request recieved.			*/
/*						*/
/************************************************/
static void dummyrcpt(CTTYP *contact,char *trouble)

{
RCPTYP *rcpt;

rcpt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
rcpt->status=m_trouble;
rcpt->rcpt=strdup("Unknown");
(void) asprintf(&(rcpt->why),"%s %s",SOULALR,trouble);
contact->rcpts=(RCPTYP **)vec_addlstlst((void **)contact->rcpts,(void *)rcpt);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to ask extended advice	*/
/*						*/
/************************************************/
static void feednote(BPTR *bptr,u_long mntpool,CTTYP *cta)

{
#define	OPFE	"modpro.c:feednote,"

#define MXWMEN	10/*max time waiting for mentor	*/

char *rqst;
char *msgref;
int phase;
int proceed;

rqst=(char *)0;
msgref=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*makeking advice rqst	*/
      if ((rqst=eml_cttotknmsg(cta))==(char *)0) {
	(void) clm_alert(0,"%s Unable to make an ADV msg (bug?)",OPFE);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*sending message	*/
      if (rqst!=(char *)0) {	/*always	*/
	TIMESPEC timeout;

	timeout.tv_sec=(time_t)MXWMEN;
	timeout.tv_nsec=(u_long)0;
        if ((msgref=ipc_postmsg(bptr,&timeout,mntpool,rqst))==(char *)0) {
	  (void) clm_alert(0,"%s Unable to send FEED msg <%s> in due time ",OPFE,rqst);
          (void) free(rqst);
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    case 2	:	/*no answer expected	*/
      (void) free(msgref);
      (void) free(rqst);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;	
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to force a local decision	*/
/*	advice if Mentor was not available.	*/
/*						*/
/************************************************/
static void locadvices(EVLTYP *evl,RCPTYP **rcpts)

{
float pourcent;
float score;
float limit;

pourcent=0.0;
if (evl->spamlvl!=(char *)0)
  if (sscanf(evl->spamlvl,"%f/%f",&score,&limit)==2)
    pourcent=(score/limit)*100.0f;
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rc;

    rc=rcpts[i];
    switch (rc->status) {
      case m_indata	:
	rc->status=m_acceptable;
				/*NO BREAK		*/
      case m_toforward	:
	if (pourcent>=100.0f) {
	  rc->why=clm_freestr(rc->why);	
	  (void) asprintf(&(rc->why),"local score=%4.2f%% (was %s)",
			  pourcent,eml_strmntstatus(rc->status));
	  switch (rc->status) {
	    case m_acceptable	:
	      rc->status=m_quarantine;
	      break;
	    case m_toforward	:
	      rc->status=m_stayput;
	      break;
	    default	:
    	      (void) clm_alert(0,"%s User <%s> unexpected status='%d'",rc->rcpt,rc->status);
	      break;
	    }
	  }
        break;
      case m_mustfwd	:	/*forced forwarding	*/
	rc->status=m_toforward;
        break;
      default		:
        break;
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set some specific targeted	*/
/*	to a new value.				*/
/*						*/
/************************************************/
static void setsomestatus(CTTYP *contact,MENTYP targstatus,MENTYP newstatus)

{
if ((contact!=(CTTYP *)0)&&(contact->rcpts!=(RCPTYP **)0)) {
  int i;

  for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rcptloc;

    rcptloc=contact->rcpts[i];
    if (rcptloc->status==targstatus) {
        rcptloc->status=newstatus; 
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set all user status to	*/
/*	collect statistic later on.		*/
/*						*/
/************************************************/
static void setallstatus(CTTYP *contact,MENTYP status,char *fmt,...)

{
#define	OPST	"modpro.c:setallstatus,"

va_list args;
char *info;

va_start(args,fmt);
vasprintf(&info,fmt,args);
if ((contact!=(CTTYP *)0)&&(contact->rcpts!=(RCPTYP **)0)) {
  int i;

  for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rcptloc;

    rcptloc=contact->rcpts[i];
    switch (rcptloc->status) {
      case m_rejected	:	
      case m_discarded	:	
      case m_empty	:	
      case m_trouble	:	
      case m_relayed	:
	break;		/*nothing to do		*/
      case m_toresolv	:
      case m_toforward	:
      case m_toresend	:
      case m_indata	:
	rcptloc->why=clm_freestr(rcptloc->why);
	rcptloc->why=strdup(info);
        rcptloc->status=status; 
	break;		/*force all status	*/
      default		:	/*unexpected	*/
        (void) clm_alert(0,"%s status='%d' unexpected; session <%s> for <%s> new status='%d', info=<%s>",
			   OPST,rcptloc->status,contact->session,rcptloc->rcpt,status,info);
	break;
      }
    }
  }
(void) free(info);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report virus detection to	*/
/*	remote originator AND logs it.		*/
/*						*/
/************************************************/
static void virusfound(CTTYP *cta,int status)

{
char *info;
char *forlogs;

info=eml_ascstat(status,cta->virname);
(void) eml_printfcrlf(cta,info);
forlogs=eml_forlogs(info,(char *)0);
(void) jrl_journalling(false,EVJRL,"%s\n\t  %s",CLAVIR,forlogs);
(void) free(forlogs);
(void) free(info);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request to checker about 	*/
/*	virus in file.				*/
/*						*/
/************************************************/
static int checkvirus(BPTR *bptr,u_long pool,CTTYP *cta)

{
#define	OPVI	"modpro.c:checkvirus,"
int status;
char *rsp;

status=ERRPROC;
if ((rsp=eml_xchgmsg(bptr,pool,"%s %s",CHKVIR,cta->session))!=(char *)0) {
  CLMCHKTYP virstat;
  int pos;
  register char *ptr;

  (void) sscanf(rsp,"%d%n",(int *)(&virstat),&pos);
  ptr=rsp+pos;
  (void) memmove(rsp,ptr,strlen(ptr)+1);
  (void) clm_skipblank(rsp);
  switch (virstat) {
    case cla_clean	:	/*no virus, E-mail fine	*/
      status=CMDOK;
      break;
    case cla_virus	:	/*Virus found		*/
      status=BADVIR;
      (void) asprintf(&(cta->virname),"%s",rsp);
      (void) virusfound(cta,status);
      break;
    case cla_nocheck	:	/*Virus not check	*/
      (void) jrl_journalling(false,EVJRL,"Email not checked for Virus");
      status=ERRPROC;
      break;
    default	:		/*trouble!		*/
      (void) clm_core_dump("%s, virstat '%d' unexpected (bug?)",OPVI,virstat);
      break;
    }
  (void) free(rsp);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send collected state to 	*/
/*	parent process and reset the session	*/
/*	the file spam level.			*/
/*						*/
/************************************************/
static void sessionstats(MXCTYP *mxc,CTTYP *contact)

{
#define	OPSS	"modpro.c:sessionstats,"

char *msg;
int phase;
int proceed;

msg=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have a contact	*/
      if (contact==(CTTYP *)0) {
        char str[200];

        (void) snprintf(str,sizeof(str),"%s, No contact structure (bug?)",OPSS);
	(void) clm_alert(0,"%s",str);
        (void) jrl_journalling(true,APJRL,"%s",str);
	phase=999;
	}
      break;
    case 1	:	/*do we have email info	*/
      if ((contact->frm==(FRMTYP *)0)||(contact->rcpts==(RCPTYP **)0)) {
	phase=999;	/*contact not successful*/
        }
      break;
    case 2	:	/*set msg contents	*/
      (void) asprintf(&msg,"%s %s %lu %s %s\n",STATFRM,
					 contact->frm->efrom,
					 contact->mtaille,
					 contact->locip,
					 contact->servport);
      if (msg!=(char *)0) {
        int i;

        for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
          char *ext;
          register RCPTYP *loc;
  
          loc=contact->rcpts[i];
          (void) asprintf(&ext,"%s %d %s\n",DATA,loc->status,loc->rcpt);
          msg=clm_astrcat(msg,ext);
          (void) free(ext);
          }
        }
      break;
    case 3	:	/*send message to collector	*/
      if (msg!=(char *)0) {
        char *rsp;

        if ((rsp=eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s",msg))!=(char *)0) {
          int status;
	
	  status=0;
          (void) sscanf(rsp,"%d",&status);
	  switch (status) {
	    case CMDOK	:	/*everything fine	*/
	      break;
	    default	:	/*unexpected trouble	*/
	      (void) clm_alert(0,"%s Unexpected status msg returned=<%s> (pid=%d bug?)",OPSS,rsp,getpid());
	      break;
	    }
	  (void) free(rsp);
	  }
        (void) free(msg);
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request log to be stored	*/
/*	within the SQL data-base.		*/
/*						*/
/************************************************/
static void sessionlogs(MXCTYP *mxc,CTTYP *contact)

{
#define	OPSL	"modpro.c:sessionlogs,"

DSPTYP *dsp;
char **dspstr;
char *strloc;
char *rsp;
int phase;
int proceed;

dsp=(DSPTYP *)0;
dspstr=(char **)0;
strloc=(char *)0;
rsp=(char *)0;
phase=0;
proceed=((contact->founddata==false)&&(contact->rcpts!=(RCPTYP **)0));
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting data	*/
      if ((dsp=eml_contacttodsp(contact))==(DSPTYP *)0) {
	(void) clm_alert(0,"%s Unable to extract dispatch session <%s> (bug?)",
			    OPSL,contact->session);
	proceed=false;	/*BIG trouble		*/
	}
      break;
    case 1	:	/*converting dispatch	*/
      if ((dspstr=eml_setdispatch(dsp))==(char **)0) {
	(void) clm_alert(0,"%s Unable to setdispatch session <%s> (bug?)",
			    OPSL,contact->session);
	phase=999;	/*BIG trouble		*/
	}
      break;
    case 2	:	/*merging dspstr	*/
      if ((strloc=clm_mergestring(dspstr,'\n'))==(char *)0) {
	(void) clm_alert(0,"%s Unable to setdispatch session <%s> (bug?)",
			    OPSL,contact->session);
	proceed=false;	/*BIG trouble		*/
	}
      dspstr=(char **)vec_freelstlst((void **)dspstr,VECFREE);
      break;
    case 3	:	/*sending information	*/
      if ((rsp=eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s %s",REPTLOG,strloc))==(char *)0) {
	(void) clm_alert(0,"%s No answer to %s <\n%s\n> (overload?)",OPSL,REPTLOG,strloc);
	phase=999;
	}
      break;
    case 4	:	/*cleaning act		*/
      if (rsp!=(char *)0) {	/*always true	*/
        int status;
	
	status=0;
        (void) sscanf(rsp,"%d",&status);
	switch (status) {
	  case CMDOK	:	/*everything fine	*/
	    break;
	  default	:	/*unexpected trouble	*/
	    (void) clm_alert(0,"%s Unexpected status msg returned=<%s> to <\n%s\n>"
			       " (overload?)",OPSL,rsp,strloc);
	    break;
	  }
        (void) free(rsp);
	}
      (void) free(strloc);
      break;
    default	:	/*SAFE guard		*/
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to log all session information*/
/*						*/
/************************************************/
static void sessionwrapup(MXCTYP *mxc,CTTYP *contact,SCODTYP scode)

{
#define	OPSW	"modpro.c:sessionwrapup,"

if ((contact!=(CTTYP *)0)&&(contact->rcpts!=(RCPTYP **)0)) {
  int i;

  for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rcptloc;

    rcptloc=contact->rcpts[i];
    switch (rcptloc->status) {
      case (m_relayed)	:
	if (contact->founddata==false) {
	  rcptloc->status=m_empty;
	  rcptloc->why=clm_freestr(rcptloc->why);
	  (void) asprintf(&(rcptloc->why),"ALR: %s","no email data received");
	  }
	break;
      case (m_indata)	:	/*rcpt set?	*/
	switch (scode) 	{
	  case s_data	:
	  case s_timeout:
	  case s_reset	:
	  case s_quit	:
	    rcptloc->status=m_empty;
	    break;
	  default	:	/*Hmmm Strange	*/
	    (void) clm_alert(0,"%s session <%s> rcpt=<%s> status/scode='%d/%d' Unexpected (bug?)",
				OPSW,contact->session,rcptloc->rcpt,rcptloc->status,scode);
	    break;
	  }
	break;
      default		:	/*		*/
	break;
      }
    }
  }
(void) sessionlogs(mxc,contact);
(void) sessionstats(mxc,contact);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request to checker about 	*/
/*	the file spam level.			*/
/*						*/
/************************************************/
static char *getspamlvl(BPTR *bptr,u_long pool,CTTYP *cta)

{
char *spamlvl;
char *commande;

(void) asprintf(&commande,"%s %s %s %s %s",GETSPM,
			  cta->locip,cta->servport,
			  cta->rcpts[0]->rcpt,cta->session);
spamlvl=eml_xchgmsg(bptr,pool,"%s",commande);
(void) free(commande);
return spamlvl;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to dispatch Mail to sending	*/
/*	process.				*/
/*						*/
/************************************************/
static int rmtdispatch(BPTR *bptr,u_long pool,CTTYP *cta)

{
#define	OPRT	"modpro.c:rmtdispatch,"

int completion;
char *rsp;
int phase;
int proceed;

completion=false;
rsp=(char *)0;
phase=0;
proceed=true;
cta->info=clm_freestr(cta->info);
(void) asprintf(&(cta->info),"%s","Email LOST!, unable to dispatch email");
while (proceed==true) {
  switch (phase) {
    case 0	:	/*dispatching file	*/
      if (eml_dispatch(cta)==false) 
	phase++;	/*no need for dispatch	*/
      break;
    case 1	:	/*rqst to do sending	*/
      if ((rsp=eml_xchgmsg(bptr,pool,"%s %s",RQSTSND,cta->session))!=(char *)0) {
        int status;
        int pos;
	
	status=0;
        (void) sscanf(rsp,"%d%n",&status,&pos);
        (void) memmove(rsp,rsp+pos,strlen(rsp+pos)+1);
	(void) clm_skipblank(rsp);
        cta->info=clm_freestr(cta->info);
        cta->info=strdup(rsp);
	switch (status) {
	  case CMDOK	:
	    completion=true;
	    break;
	  default	:
	    (void) clm_alert(0,"%s got status='%d' to dispatch request (bug?)",OPRT,status);
	    break;
	  }
	(void) free(rsp);
	}
      break;
    case 2	:	/*removing dispatch file*/
      if (completion!=true) {
        int i;
	char *alr;

	alr="ALR: email lost";	
	(void) clm_alert(0,"%s Unable to dispatch session <%s>, email lost!",OPRT,cta->session);
	for (i=0;cta->rcpts[i]!=(RCPTYP *)0;i++) {
	  RCPTYP *rcptloc;

	  rcptloc=cta->rcpts[i];
	  switch (rcptloc->status) {
	    case m_trouble   :	/*reject entry	*/
	    case m_rejected  :
	    case m_discarded :
	      break;
	    default	     :
	      rcptloc->why=clm_freestr(rcptloc->why);
	      (void) asprintf(&(rcptloc->why),alr);
              rcptloc->status=m_trouble; 
	      break;
	    }
	  }
	}
      (void) eml_remdispatch(cta);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return completion;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set a reject note as	*/
/*	subject in the logs.			*/
/*						*/
/************************************************/
static void updwhy(RCPTYP *rmt,const char *fmt,...)

{
va_list args;
char *newwhy;
char *ptr;

va_start(args,fmt);
vasprintf(&newwhy,fmt,args);
if ((ptr=strchr(newwhy,'\n'))!=(char *)0) {
  *ptr='\000';
  (void) clm_skiprtlf(newwhy);
  }
rmt->why=clm_freestr(rmt->why);
rmt->why=newwhy;
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request RCPT status from	*/
/*	Checker.				*/
/*						*/
/************************************************/
static int rmtchecking(BPTR *bptr,u_long pool,CTTYP *cta,char *rcpt,char **answ)

{
int status;
FRMTYP *frm;
char *rsp;

status=ERRPROC;
frm=cta->frm;
rsp=(char *)0;
if (frm!=(FRMTYP *)0) {
  char *msg;
  CKPTYP scn;

  (void) memset(&scn,'\000',sizeof(CKPTYP));
  scn.peerip=frm->peerip;
  scn.heloname=cta->heloname;
  scn.efrom=frm->efrom;
  scn.taille=frm->mtaille;
  scn.locip=frm->locip;
  scn.locport=frm->servport;
  scn.rcpt=rcpt;
  scn.ctxflag=frm->frmstatus;
  if ((msg=eml_mkckrrqst(&scn))!=(char *)0) {
    if ((rsp=eml_xchgmsg(bptr,pool,"%s %s",CKRCPT,msg))!=(char *)0) {
      int pos;
      register char *ptr;

      (void) sscanf(rsp,"%d%n",&status,&pos);
      ptr=rsp+pos;
      (void) memmove(rsp,ptr,strlen(ptr)+1);
      }
    (void) free(msg);
    }
  }
switch (status) {
  case RELAYOK	:	/*Keeping actual status		*/
  case FWRDOK	:	
  case MUSTFWD	:	
  case ALIOK	:	
  case SENDOK	:	
  case CMDOK	:
    (void) asprintf(answ,"%d%s",CMDOK,rsp);
    break;
  case SPAMMER	:
  case NOSPAMS	:
  case NOLOC	:
  case BADMX	:
  case NODNS	:
  case BLCKLST	:
  case BADSPF	:
  case NEWSITE	:
  case CHKUSR	:
  case NOSNDR	:
    (void) asprintf(answ,"%d%s",status,rsp);
    break;
  case ERRPROC	:
    (void) asprintf(answ,"%d%s",status,rsp);
    break;
  case REJECT	:
  default	:
    if (strlen(rsp)>0)
      (void) asprintf(answ,"%d%s",status,rsp);
    else
      *answ=eml_ascstat(status,"");
    break;
  }
rsp=clm_freestr(rsp);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to cross check rcpt address	*/
/*	with a little bit of remote help.	*/
/*	If there is no need to go further,	*/
/*	return a false status.			*/
/*						*/
/************************************************/
static int crosschecking(MXCTYP *mxc,CTTYP *cta,RCPTYP *rmt)

{
int letsroll;
char *answ;
int status;

letsroll=true;
answ=(char *)0;
switch ((status=rmtchecking(mxc->bptr,mxc->ckrpool,cta,rmt->rcpt,&answ))) {
  case CMDOK	:	/*everything OK			*/
    break;
  case RELAYOK	:	/*remote IP is relayable	*/
    cta->relayable=true;
    rmt->status=m_relayed;
    status=CMDOK;	/*Lets say everything OK	*/
    break;
  case MUSTFWD	:	/*mail must be forwarded anyway	*/
    rmt->status=m_mustfwd;
    status=CMDOK;	/*Lets say everything OK	*/
    break;
  case FWRDOK	:	/*mail to forward to an Internal*/
    rmt->status=m_toforward;
    status=CMDOK;	/*Lets say everything OK	*/
    break;
  case ALIOK	:	/*mail to be aliased later on	*/
    rmt->status=m_toresolv;
    status=CMDOK;	/*Lets say everything OK	*/
    break;
  case SENDOK	:	/*Not Primary MX, lets resend 	*/
    cta->relayable=true;/*the Email			*/
    rmt->status=m_toresend;
    status=CMDOK;	/*Lets say everything OK	*/
    break;
  case MAXDAY	:	/*sending too many email!	*/
  case NORELAY	:	/*We do not relay		*/
  case NOSPAMS	:	/*No SPAM thanks.		*/
  case BADSPF	:	/*no need to go further		*/
  case MISSMX	:	/*recipient without MX		*/
  case BADMX	:	/*originator without MX		*/
  case BLCKLST	:	/*black list by recipient	*/
    rmt->status=m_rejected;
    break;
  case BULKOVR	:	/*too many rcpt  stop process	*/
  case ERRPROC	:	/*we are really in deep trouble	*/
    rmt->status=m_rejected;
    letsroll=false;
    sigterm=true;	/*lets get out fast		*/
    break;
  case CHKUSR	:	/*Recipient need to be checked	*/
  case TOOMANY	:	/*To Many E_mail in queue	*/
  case NEWSITE	:	/*server first encountance	*/
  case ALREAD	:	/*server already connected	*/
  case NOSNDR	:	/*Sender is not connected	*/
  case NOCHKR	:	/*Checker not available		*/
  case SPAMMER	:	/*Sender seems to be a spammer	*/
    rmt->status=m_rejected;
    break;
  default	:	/*too bad for other status	*/
    rmt->status=m_rejected;
    break;
  }
rmt->smtpstatus=status;
rmt->why=answ;
switch (rmt->smtpstatus) {
  case NEWSITE	:	/*no need to go further	*/
  case NOSPAMS	:	/*site is new or spammer*/
    letsroll=false;	/*forget about remote	*/
    break;
  default	:	/*lets continue	session	*/
     break;		/*nothing to do		*/
  }
return letsroll;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request the clamav version	*/
/*	used by the clement checker.		*/
/*	within the event journal		*/
/*						*/
/************************************************/
static void getclamvers(MXCTYP *mxc,CTTYP *cta)

{
char *rsp;
char *clamvs;

clamvs=(char *)0;
if ((rsp=eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s",CLAMVS))!=(char *)0) {
  int status;
  int sofar;

  status=0;
  if (sscanf(rsp,"%d%n",&status,&sofar)==1) {
    char *ptr;
  
    ptr=rsp+sofar; 
    while (*ptr==' ')
      ptr++;
    switch (status) {
      case CMDOK	:
	clamvs=strdup(ptr);
	break;
      default		:
	(void) asprintf(&clamvs,"Error? <%s>",ptr);
	break;
      }
    }
  (void) free(rsp);
  }
if (clamvs==(char *)0)
  clamvs=strdup("Virus checker version undefined");
cta->clamvers=clamvs;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display ALL from status	*/
/*	within the event journal		*/
/*						*/
/************************************************/
static void reportfrom(CTTYP *contact,int status)

{
#define	FRMOK	"%d <%s>... sender ok"
#define	ADDFRM	"\n\t%*s%s"

register u_int frmstatus;
char *curlog;
int phase;
int proceed;

frmstatus=contact->frm->frmstatus;
(void) asprintf(&curlog,"%s: <%s>",MAILF,contact->frm->mfrom);
if (contact->frm->mtaille>0)
  (void) clm_fasprintf(&curlog," SIZE=%lu",contact->frm->mtaille);
phase=0;
proceed=true;
while (proceed==true) {
  char *addptr;

  addptr=(char *)0;
  switch (phase) {
    case 0	: /*first report		*/
      switch (status) {
	case CMDOK	:
          (void) asprintf(&addptr,FRMOK,status,contact->frm->efrom);
	  break;
	case STOLEN	:
          addptr=eml_ascstat(status,contact->frm->efrom);
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpected status '%d' (bug?)","modpro.c:reportfrom,",status);
	  status=ERRPROC;
			/*NO BREAK		*/
	case ERRPROC	:
          addptr=eml_ascstat(status,contact->frm->efrom);
	  break;
	}
      (void) eml_printfcrlf(contact,"%s",addptr);
      (void) clm_fasprintf(&curlog,ADDFRM,2,"",addptr);
      (void) free(addptr);
      addptr=(char *)0;
      break;
    case 1	:	/*Checking big trouble	*/
      if ((frmstatus&BADFROM)!=0)
	addptr=strdup("Bad FROM: request");
      break;
    case 2	:	/*Checking AUTH		*/
      if ((frmstatus&AUTH)!=0)
	addptr="Originator Authenticated";
      break;
    case 3	:	/*Checking Relaying	*/
      if ((frmstatus&RELAY)!=0)
	addptr="Peer IP relayed";
      break;
    case 4	:	/*Checking if bad MX	*/
      if ((frmstatus&NOMX)!=0)
	addptr="Domain without MX";
      break;
    case 5	:	/*Checking bad SPF	*/
      if ((frmstatus&SPFFAIL)!=0)
	addptr="Peer IP SPF restricted";
      break;
    case 6	:	/*checking for stolen ID*/
      if ((frmstatus&AUTHALR)!=0)
	addptr="ID Missmatch";
      break;
    case 7	:	/*checking for bad EHLO	*/
      if ((frmstatus&BADEHLO)!=0)
	addptr="Bad Helo data";
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  if (addptr!=(char *)0) {
    (void) clm_fasprintf(&curlog,ADDFRM,6,"",addptr);
    }
  phase++;
  }
(void) jrl_journalling(false,EVJRL,"%s",curlog);
(void) free(curlog);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle MAIL FROM request	*/
/*	Request to check all originator IP	*/
/*	parameters.				*/
/*						*/
/************************************************/
static int checkmail(MXCTYP *mxc,CTTYP *cta)

{
int status;
char *msg;
char *rsp;
int phase;
int proceed;

status=ERRPROC;
(void) asprintf(&msg,"%s %s %s %s %d",cta->frm->efrom,cta->frm->peerip,
				     cta->frm->locip,cta->frm->servport,
				     cta->autok);
rsp=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking for mismatch	*/
      if (cta->autok==true) {
	if (strcmp(cta->authname,cta->frm->efrom)!=0) {
	  cta->frm->frmstatus|=AUTHALR;
	  status=STOLEN;
	  phase=999;	/*no need to go further	*/
	  }
	}
      break;
    case 1	:	/*asking to double check*/
      if ((rsp=eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s %s",CKFROM,msg))!=(char *)0) {
	int maxbulk;
	int sofar;

	if (sscanf(rsp,"%d %d %n",&status,&maxbulk,&sofar)==2) {
	  char *ptr;

	  ptr=rsp+sofar; 
	  while (*ptr==' ')
	    ptr++;
	  cta->frm->stainfo=strdup(ptr);
	  cta->frm->frmstatus|=status;
	  cta->frm->maxbulk=maxbulk;
	  status=CMDOK;
	  }
        else {
	  status=ERRPROC;
	  (void) clm_alert(0,"%s <%s> is not a status (bug?)","modpro.c:checkmail,",rsp); 
	  }
        (void) free(rsp);
        }
      break;
    default	:	/*SAFE guard		*/
      (void) reportfrom(cta,status);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(msg);
return (status==CMDOK);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if data-contents is	*/
/*	acceptable and store data within the	*/
/*	sender queue.				*/
/*						*/
/************************************************/
static int checkdata(MXCTYP *mxc,CTTYP *cta)

{
#define MXRCVD 100	/*Email LOOP detection	*/

int letsroll;
int virstat;
char *cmt;
u_long startmilli;
u_long stopmilli;
int phase;
int proceed;

letsroll=true;
virstat=ERRPROC;
cmt=(char *)0;
startmilli=(u_long)0;
stopmilli=(u_long)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*collecting data		*/
      phase=999;	/*trouble expected(?)		*/
      startmilli=clm_getmillisec();
      switch (eml_collectdata(cta)) {
 	case s_data	:	
	  phase=0;	/*lets continue process		*/
          stopmilli=clm_getmillisec();
          (void) jrl_journalling(false,EVJRL,"Collected %ld bytes within %5.3f seconds",
					     cta->mtaille,(float)(stopmilli-startmilli)/1000.0);
	  break;
 	case s_baddata	:	
	  cmt="Bad data stream";
          (void) jrl_journalling(false,EVJRL,cmt);
	  (void) setallstatus(cta,m_trouble,cmt);
	  break;
 	case s_timeout	:	
	  cmt="Data-collection timeout";
          (void) jrl_journalling(false,EVJRL,cmt);
	  (void) setallstatus(cta,m_empty,cmt);
	default		:
	  letsroll=false;
	  break;
        }
      break;
    case  1	:	/*parsing E-mail header		*/
      if ((cta->evl=hdr_parseheader(cta->session,cta->headers))==(EVLTYP *)0) {
	char *strloc;

	strloc="Incorrect E-mail Header";
	(void) eml_report(cta,MALABRT,strloc);
	(void) setallstatus(cta,m_trouble,"%s",strloc);
	phase=999;
	}
      break;
    case  2	:	/*From: missing from header?	*/
      cta->founddata=true;
      if (cta->evl->realfrom==(char *)0)
	cta->evl->realfrom=strdup(cta->frm->efrom);
      break;
    case  3	:	/*cheking emaill loop		*/
      if (cta->evl->nbrreceived>MXRCVD) {
	char *info;
	char *forlogs;

	info=eml_ascstat(HDRLOOP,cta->evl->nbrreceived);
	(void) eml_printfcrlf(cta,info);
	forlogs=eml_forlogs(info,(char *)0);
	(void) jrl_journalling(false,EVJRL,"%s\n\t  %s","Mail rejected: Email loop",forlogs);
	(void) setallstatus(cta,m_discarded,"%s","Delivery loop detected in Email header");
	(void) free(forlogs);
	(void) free(info);
	cta->founddata=false;
	phase++;	/*no need to check about virus	*/
	phase++;	/*no need to check spam level	*/
	}
      break;
    case  4	:	/*checking for virus		*/
      virstat=checkvirus(mxc->bptr,mxc->ckrpool,cta);
      switch (virstat) {
	case ERRPROC	:
	  (void) eml_changeheader(cta,VCHK,"Warning! Virus Free Contents NOT validated");
				/*NO BREAK but trouble	*/
	case CMDOK	:
          cta->info=clm_freestr(cta->info);
          (void) asprintf(&(cta->info),"%d Message accepted for delivery (Session ID=<%s>)",CMDOK,cta->session);
          (void) eml_printfcrlf(cta,cta->info);
          (void) jrl_journalling(false,EVJRL,cta->info);
	  break;
	case BADVIR	:
	  cta->founddata=false;
	  (void) setallstatus(cta,m_virusfound,"Virus found <%s>",cta->virname);
	  break;
	default		:	/*virus found		*/
	  (void) clm_alert(0,"%s, sessions <%s> unexpected virus status (bug?)","modpro.c,checkdata,",
			      cta->session);
	  phase=999;		/*no need to go further	*/
	  break;
	}
      if (virstat!=CMDOK)
	phase++;	/*no need to check spam level	*/
      (void) prc_settitle("%s Cross-checking email sessions <%s>",appname,cta->session);
      break;
    case  5	:	/*checking spamlevel		*/
      if ((cta->relayable==false)&&(cta->autok==false))
	cta->evl->spamlvl=getspamlvl(mxc->bptr,mxc->ckrpool,cta);
      else
	(void) setsomestatus(cta,m_indata,m_acceptable);
      break;
    case  6	:	/*reporting Header		*/
      cta->evl->mtaille=cta->mtaille;
      (void) reporthdrs(cta->evl);
      break;
    case  7	:	/*asking checker advice		*/
      (void) askadvices(mxc->bptr,mxc->ckrpool,cta);
      break;
    case  8	:	/*set missing advice		*/
      (void) locadvices(cta->evl,cta->rcpts);
      break;
    case  9	:	/*asking mentor advice		*/
      (void) feednote(mxc->bptr,mxc->mntpool,cta);
      break;
    case 10	:	/*dispatching			*/
      if (cta->founddata==true) { /*data to dispatch?	*/
        if (rmtdispatch(mxc->bptr,mxc->ckrpool,cta)==false) {
          char *inf;

          (void) asprintf(&inf,"%s %s","Warning!",cta->info);
          (void) jrl_journalling(false,EVJRL,inf);
	  (void) setallstatus(cta,m_trouble,"%s",inf);
	  (void) free(inf);
	  }
	}
      break;
    case 11	:	/*logging rcpt final status	*/
      (void) pro_reportrcpt(cta->rcpts);
      break;
    default	:	/*SAFE guard			*/
      proceed=false;
      break;
    }
  phase++;
  }
return letsroll;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle user authentication	*/
/*	request.				*/
/*						*/
/************************************************/
static int checkusrauth(MXCTYP *mxc,CTTYP *cta)

{
#define	OPCH "modpro.c:checkusrauth,"

int letsroll;

letsroll=false;
if ((cta->info!=(char *)0)&&(strlen(cta->info)>0)) {
  int checkstat;
  char *username;
  char *password;
  char *msg;
  char *rsp;
  int phase;
  int proceed;

  checkstat=ERRPROC;
  username=(char *)calloc(strlen(cta->info),sizeof(char));
  password=(char *)calloc(strlen(cta->info),sizeof(char));
  msg=(char *)0;
  rsp=(char *)0;
  phase=0;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case  0	:	/*extracting info	*/      
        if (sscanf(cta->info,"%s %s",username,password)!=2) {
          (void) clm_alert(0,"%s wrong format <%s> (bug?)",OPCH,cta->info);
          cta->authname=strdup("was_not_specified");
	  phase=999;	/*trouble trouble	*/
	  }
	break;
      case  1	:	/*sending auth request	*/
	cta->authname=strdup(username);
        (void) asprintf(&msg,"%s %s %s %s",CKID,
			     cta->info,cta->locip,cta->servport);
        if ((rsp=eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s",msg))==(char *)0) {
          (void) eml_report(cta,ERRPROC,"Time out in authentication process");
	  phase=999;	/*trouble trouble	*/
	  }
        (void) free(msg);
	break;
      case  2	:	/*getting auth status	*/
        (void) sscanf(rsp,"%d",&checkstat); 
	switch (checkstat) {
	  case IDOK	:
            (void) jrl_journalling(false,EVJRL,"Auth Accepted user=<%s>",
						username);
	    letsroll=true;
	    break;
	  case BADID	:{
	    char *strloc;

	    (void) asprintf(&strloc,"Auth Rejected user=<%s> pass=<%s>",username,password);
	    (void) eml_dummyfrm(cta); 
	    (void) dummyrcpt(cta,strloc);
            (void) jrl_journalling(false,EVJRL,strloc);
	    (void) free(strloc);
	    (void) sleep(2);	/* to further delay retry	*/	
	    }
	    break;
	  default	:
            (void) jrl_journalling(false,EVJRL,"Auth wrong status='%d' "
					       "user=<%s> pass=<%s>",
						checkstat,username,password);
	    break;
	  }
        (void) jrl_journalling(false,EVJRL,"%s",rsp);
        (void) eml_printfcrlf(cta,"%s",rsp);
	(void) free(rsp);
	break;
      default	:	/*SAFE guard		*/
	(void) free(password);
	(void) free(username);
        proceed=false;
	break;
      }
    phase++;
    }
  }
else 
  (void) clm_alert(0,"%s no data (bug?)",OPCH);
return letsroll;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report status info about	*/
/*	RCPT request.				*/
/*						*/
/************************************************/
static void logsrcpt(CTTYP *cta,char *rcpt,int smtpstatus,char *why,char *note)

{
char *forlogs;

forlogs=eml_forlogs(why,note);
(void) jrl_journalling(false,EVJRL,"%s: <%s>\n\t  %d%s",RCPT,rcpt,smtpstatus,forlogs);
forlogs=clm_freestr(forlogs);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request default domain	*/
/*	value to checker.			*/
/*						*/
/************************************************/
static char *getdfltdom(MXCTYP *mxc)

{
return eml_xchgmsg(mxc->bptr,mxc->ckrpool,"%s",DFLTDOM);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle RCPT check, first	*/
/*	check the RCPT localy and then do	*/
/*	a deep checking by asking to mentor.	*/
/*						*/
/************************************************/
static int checkrcpt(MXCTYP *mxc,CTTYP *cta)

{
#define	OPCE	"modpro.c:checkrcpt,"

int letsroll;
const char *frm;
RCPTYP *rmt;
int proceed;
int phase;

letsroll=true;
frm="%s %s";
rmt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
rmt->status=m_indata;
rmt->rcpt=cta->info;
cta->info=(char *)0;
proceed=true;
phase=1;
while (proceed==true) {
  switch (phase) {
    case   0:		/*trouble, trouble			*/
      proceed=false;
      break;
    case   1	: 	/*adding domain				*/
      if (strcmp(rmt->rcpt,POSTM)==0) {
	char *dfltdom;

	if ((dfltdom=getdfltdom(mxc))!=(char *)0) {
	  rmt->note=clm_freestr(rmt->note);
	  (void) asprintf(&(rmt->note),"Domain <%s> assigned to recipient <%s>",dfltdom,rmt->rcpt);
	  (void) free(rmt->rcpt);
	  (void) asprintf(&(rmt->rcpt),"%s@%s",POSTM,dfltdom);
	  (void) free(dfltdom);
	  }
	}
      break;
    case   2	: 	/*is user already in rcpt		*/
      if (cta->rcpts!=(RCPTYP **)0) {
        register int i;

	for (i=0;cta->rcpts[i]!=(RCPTYP *)0;i++) {
	  if (strcmp(cta->rcpts[i]->rcpt,rmt->rcpt)==0) {
	    rmt->status=m_discarded;
	    rmt->smtpstatus=REJECT;
	    (void) asprintf(&(rmt->why),"%d %s",cta->rcpts[i]->smtpstatus,"Recipient already specified");
	    break;
	    }
	  }
        }
      break;
    case   3	: 	
      cta->frm->curbulk++;
      if (cta->frm->curbulk>cta->frm->maxbulk) {
	rmt->status=m_discarded;
	rmt->smtpstatus=BULKOVR;
        rmt->why=eml_ascstat(BULKOVR,cta->frm->maxbulk);
	}
      switch (rmt->status) {
	case m_rejected		:
	case m_discarded	:
	  phase++;	/*no need to do remote checking		*/
	  break;
	default			:
	  break;
	}
      break;
    case   4	: 	/*checking via remote connection	*/
      letsroll=crosschecking(mxc,cta,rmt);
      break;
    case   5	: 	/*checking for too many returned bad	*/
      if (rmt!=(RCPTYP *)0) {	/*always			*/
	int pos;

	pos=0;
        if (cta->badrcpt>BADRCPT) {
	  rmt->why=clm_freestr(rmt->why);
          rmt->why=eml_ascstat(MANYRJT);
	  rmt->status=m_rejected;
	  }
        if (sscanf(rmt->why,"%d%n",&rmt->smtpstatus,&pos)==1) 
          (void) memmove(rmt->why,rmt->why+pos,strlen(rmt->why+pos)+1);
	}
      break;
    case   6	:	/*adding an answer delay if rejected	*/
      if (debug==0) {
        switch(rmt->status) {
	  case m_rejected		:
	    (void) sleep(2);	/*error penalty for remote server */
	    break;
	  default			:
	    break;
	  }
	}		/*sending status and logging		*/
      (void) eml_printfcrlf(cta,"%d%s",rmt->smtpstatus,rmt->why);
      (void) logsrcpt(cta,rmt->rcpt,rmt->smtpstatus,rmt->why,rmt->note); 
      break;
    case   7	:	/*counting status			*/
      switch (rmt->smtpstatus) {
	case DLYUSR	:
	case MANYRJT	:
	case MBFULL	:
	case BULKOVR	:
	case BANNED	:
	  cta->badrcpt++;
	  (void) updwhy(rmt,frm,SOULALR,rmt->why+1);
	  break;
        case MAXDAY	:
	case NORELAY	:
	  cta->badrcpt++;
	  (void) updwhy(rmt,frm,SOULNRL,rmt->why+1);
	  break;
	case EXPIRED	:
	case NOLOC	:
	  cta->badrcpt++;
	  (void) updwhy(rmt,frm,SOULUKN,rmt->why+1);
	  break;
	case MISSMX	:
	case BADMX	:
	case NODNS	:
	  cta->badrcpt++;
	  (void) updwhy(rmt,frm,SOULDNS,rmt->why+1);
	  break;
	case BADSPF	:
	  cta->badrcpt+=3;	/*SPF penalty			*/
	  (void) updwhy(rmt,frm,SOULSPF,rmt->why+1);
	  break;
	case BLCKLST	:	/*system black listed		*/
	  cta->badrcpt+=3;	/*black listing penalty		*/
	  (void) updwhy(rmt,frm,SOULBLS,rmt->why+7);
	  break;
	case CMDOK	:	/*good status			*/
	  if (cta->goodrcpt==0)
	    cta->foremail=strdup(rmt->rcpt);
	  cta->goodrcpt++;
	  break;
	default	:	/*unexpected error	*/
	  (void) clm_alert(0,"%s Unexpected status error <%d %s> (session=<%s> rcpt=<%s> bug?)",
			      OPCE,rmt->smtpstatus,rmt->why,cta->session,rmt->rcpt);
	  break;
	}
      break;
    case   8	:	/*final status management		*/
      switch (rmt->status) {
	case m_discarded:
	  rmt=eml_freercpt(rmt);
	  break;
	default		:
          cta->rcpts=(RCPTYP **)vec_addlstlst((void **)cta->rcpts,(void *)rmt);
	  break;
	}
      if (letsroll==false)
        (void) jrl_journalling(false,EVJRL,"STOP: disconnecting session");
      break;
    default	:	/*SAFE Guard				*/
      proceed=false;
      break;
    }
  phase++;
  }
return letsroll;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle SMTP protocole	*/
/*						*/
/************************************************/
static void dosmtp(MXCTYP *mxc,LSTTYP *channel,int iteration)

{
#define	OPTL	"modpro.c:dosmtp,"
#define ACTCYC	"%s now %s [%s|%s] (%02d/%02d) (pid='%05d')"

int cycle;

cycle=0;
(void) jrl_journalling(true,APJRL,ACTCYC,"Proto_smtp","Starting",
					  channel->binding->ip,
			  		  channel->binding->port,
					  iteration,
					  channel->binding->iteration,
					  getpid());
while (sigterm==false) {
  CTTYP *contact;

  (void) prc_settitle("%s waiting on [%s|%s] (%02d/%02d/%04d)",
		       appname,channel->binding->ip,
		       channel->binding->port,iteration,
		       channel->binding->iteration,cycle);
  if ((contact=eml_acceptcontact(channel))!=(CTTYP *)0) {
    int proceed;

    proceed=true;
    (void) getclamvers(mxc,contact);
    (void) eml_opensession(contact);
    while (proceed==true) {
      int letsgo;
      SCODTYP scode;

      letsgo=true;
      scode=eml_docontact(contact);
      switch (scode) {
        case s_mail	:
	  if ((letsgo=checkmail(mxc,contact))==false) {
	    char *strloc;
	
	    strloc="Disconnect: from ID mismatch auth ID";
	    (void) dummyrcpt(contact,strloc);
            (void) jrl_journalling(false,EVJRL,strloc);
	    (void) sleep(1);	/*trouble delay	*/
	    }
	  break;
        case s_rcpt	:
          letsgo=checkrcpt(mxc,contact);
	  break;
        case s_reset	:
	  contact->reset=true;
	  (void) sessionwrapup(mxc,contact,scode);
	  (void) eml_resetsession(contact);
	  (void) sleep(1);	/*artificial delay	*/
          if (contact->totbad<BADRCPT)
	    (void) eml_printfcrlf(contact,"%d Reset state",CMDOK);
          else {
	    char *msg;

            (void) asprintf(&msg,"%d Too many RCPT errors already, must disconnect",CMDABRT);
            (void) jrl_journalling(false,EVJRL,"%s",msg);
	    (void) eml_printfcrlf(contact,"%s",msg);
	    (void) free(msg);
	    proceed=false;	/*no need to go further	*/
	    }
	  break;
        case s_auth	:
	  contact->autok=true;	/*lets say it is good	*/
	  if ((letsgo=checkusrauth(mxc,contact))==false) 
            (void) jrl_journalling(false,EVJRL,"Disconnect: bad authentication password");
	  break;
	case s_unknown	:
	  (void) clm_alert(0,"%s Unknown received command '%d'",OPTL,scode);
	  letsgo=false;
	  break;
	case s_data	:	/*good data contents	*/
	  if ((proceed=checkdata(mxc,contact))==true) {
	    (void) sessionwrapup(mxc,contact,scode);
	    (void) eml_resetsession(contact);
	    /*we can now accept a new "MAIL FROM:"	*/
	    }
	  break;
	case s_baddata	:	/*bad data contents	*/
          (void) jrl_journalling(false,EVJRL,"Disconnect, bad data contents");
	  proceed=false;
	  break;
	case s_timeout	:	/*contact lost		*/
          (void) jrl_journalling(false,EVJRL,"Disconnect by Timeout");
				/*NO BREAK		*/
	case s_quit	:
	  proceed=false;
	  break;
	default		:
	  (void) clm_alert(0,"%s Unexpected received command '%d'( bug?)",
			      OPTL,scode);
	  break;
        }
      if (letsgo==false) {
        (void) sleep(1);
	proceed=false;
        }
      if (proceed==false) {
        (void) sessionwrapup(mxc,contact,scode);
	}
      }
    (void) eml_closesession(contact);
    (void) eml_lostcontact(contact);
    (void) jrl_journalling(true,EVJRL,(char *)0);
    cycle++;
    }
  }
(void) jrl_journalling(true,APJRL,ACTCYC,"Proto_smtp","exiting",
					  channel->binding->ip,
			  		  channel->binding->port,
					  iteration+1,
					  channel->binding->iteration,
					  getpid());
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to log recipient clement final*/
/*	say.					*/
/*						*/
/************************************************/
void pro_reportrcpt(RCPTYP **rcpts)

{
#define	OPSE	"modpro.c:reportrcpt"
#define	RCFM	"RCPT: <%s>\n\t\t%s%s%s"

if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rc;
    const char *code;
    char *sep;
    char *info;

    rc=rcpts[i];
    code=eml_strmntstatus(rc->status);
    sep=" ";
    info=strdup("");
    switch (rc->status) {
      case m_toresolv	:
	code="";
        info=clm_freestr(info); 
	(void) asprintf(&info," %s",rc->why);
        break;
      case m_discarded	:
      case m_virusfound	:
	code=(const char *)0;
        break;
      case m_trouble	:
      case m_acceptable	:
      case m_direct	:
      case m_relayed	:
      case m_toforward	:
      case m_toresend	:
      case m_rejected	:
      case m_quarantine	:
      case m_stayput	:
      case m_mustfwd	:
	if ((rc->why!=(char *)0)&&(strlen(rc->why)>0)) {
          info=clm_freestr(info); 
	  (void) asprintf(&info," %s",rc->why);
	  sep=":";
	  }
        break;
      default		:
        (void) clm_alert(0,"%s Unknown status '%d' for user <%s>",
		            OPSE,rc->status,rc->rcpt);
        break;
      }
    if (code!=(const char *)0)
      (void) jrl_journalling(false,EVJRL,RCFM,rc->rcpt,code,sep,info);
    (void) free(info);
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle protocole management*/
/*						*/
/************************************************/
void pro_doproto(MXCTYP *mxc,LSTTYP *channel,int iteration)

{
#define OPDO	"modpro.c:pro_doproto,"

int phase;
int proceed;

phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*set rdy to accept msg	*/
      if (ipc_subscribe(mxc->bptr,(u_long)0,true)==false) {
        (void) clm_alert(0,"%s Unable to subscribe to message exchg",OPDO);
	proceed=false;
	}
      break;
    case 1	:	/*doing real job	*/
      switch (channel->binding->proto) {
        case pro_smtp	:
          (void) dosmtp(mxc,channel,iteration);
          break;
        default	:
          (void) clm_core_dump("%s Unknown protocol '%d'",
				OPDO,channel->binding->proto);
          break;
        }
      break;
    default	:	/*SAFE guard		*/
      (void) ipc_subscribe(mxc->bptr,(u_long)0,false);
      proceed=false;
      break;
    }
  phase++;
  }
(void) jrl_closejournals();
}
