/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODCKR:					*/
/*	Module doing mail checking process	*/
/*						*/
/************************************************/
#include	<sys/poll.h>
#include	<sys/stat.h>
#include	<errno.h>
#include	<time.h>
#include	<math.h>
#include	<string.h>
#include	<unistd.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unidir.h"
#include	"unijrl.h"
#include	"unisql.h"
#include	"geseml.h"
#include	"gesque.h"
#include	"gesclm.h"
#include	"geseml.h"
#include	"gesdbc.h"
#include	"lvleml.h"
#include	"lvltrt.h"
#include	"modtrt.h"

#define	TITLE	"%s Transitr (%04lu/%08lu)"

/*to call the Email delivery shell script	*/
#define	SHELL	"shell"
#define	DELIVER	"deliver.sh"

#define	NOREPLY	"No_reply"
#define	TRANSIT	"Transit"
/*

*/
/************************************************/
/*						*/
/*	Procedure to create a transit tracelog	*/
/*						*/
/************************************************/
static void tracelog(char *dbgname,QUETYP *queue) 

{
if (dbgname!=(char *)0) {
  char *sendlog;

  sendlog=que_tracename(queue);
  (void) unlink(sendlog);
  if (link(dbgname,sendlog)<0)
    (void) clm_alert(0,"%s Unable to link file <%s> to file <%s> (error=<%s> bug?)",
		       "modtrt.c:tracelog,",dbgname,sendlog,strerror(errno));
  (void) free(sendlog);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to log queue entry status	*/
/*	destination.				*/
/*						*/
/************************************************/
static void logqueue(SQLFILE *sql,QUETYP *queue) 

{
	 char *recipient;
register char **split;

(void) asprintf(&recipient,"%s@%s",queue->recname,queue->recdom);
if ((split=clm_splitstring(queue->why,'\n'))!=(char **)0) {
  char *rpt;
  register int s;

  rpt=(char *)0;	
  (void) clm_fasprintf(&rpt,"%s",split[0]);
  for (s=1;split[s]!=(char *)0;s++)
    (void) clm_fasprintf(&rpt,"\n\t%s%s","  ",split[s]);
  (void) jrl_journalling(false,EVTRT,"%s",rpt);
  rpt=clm_freestr(rpt);
  split=(char **)vec_freelstlst((void **)split,VECFREE);
  }
else
  (void) jrl_journalling(false,EVTRT,"%s",queue->why);
(void) dbc_updlogs(sql,queue->session,recipient,eml_getcomstat(queue->status,(MENTYP)queue->orgstatus));
(void) dbc_updqueue(sql,queue);
(void) free(recipient);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to sort queue according	*/
/*	destination.				*/
/*						*/
/************************************************/
static QUETYP **sortdest(QUETYP **queues) 

{
QUETYP **dest;

dest=(QUETYP **)0;
if ((queues!=(QUETYP **)0)&&(queues[0]!=(QUETYP *)0)) {
  register int qulong;
  register int i;

  qulong=vec_sizelstlst((void **)queues);
  dest=(QUETYP **)calloc(qulong+1,sizeof(QUETYP *));
  dest[0]=queues[0];
  for (i=1;i<qulong;i++) {
    if ((strcmp(dest[0]->recdom,queues[i]->recdom)==0)&&
	 strcmp(dest[0]->smtpmx,queues[i]->smtpmx)==0) {
      dest[i]=queues[i];
      continue;
      }
    break;
    }
  qulong++;
  (void) memmove(queues,queues+i,(qulong-i)*sizeof(QUETYP *));
  }
return dest;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the transit status	*/
/*	explainantion.				*/
/*	allow Email transit.			*/
/*						*/
/************************************************/
static void setwhy(QUETYP **queues,int status,char *frm, ...)

{
/*ATTENTION Note: bug within vasprintf!		*/
/*		  va_list is lost, do not call	*/
/*		  it twice with the same args!	*/
va_list args;

va_start(args,frm);
if (queues!=(QUETYP **)0) {
  char *strloc;
  int i;

  (void) vasprintf(&strloc,frm,args);
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    queues[i]->status=status;
    queues[i]->why=strdup(strloc);
    }
  (void) free(strloc);
  }
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to establish a contact with	*/
/*	one of the available MX			*/
/*						*/
/************************************************/
static CTTYP *makecontact(MXTYP **mxlst,char *smtpport,time_t sectimeout)

{
CTTYP *contact;

contact=(CTTYP *)0;
if (mxlst!=(MXTYP **)0) {
  register int i;

  for (i=0;mxlst[i]!=(MXTYP *)0;i++) {
    contact=eml_opencontact(pro_smtp,mxlst[i]->mxname,smtpport,sectimeout);
    if ((contact!=(CTTYP *)0)&&(contact->established==true)) {
      (void) jrl_journalling(false,EVTRT,"Connected to <%s/[%s]> port '%s'",
					  contact->peername,contact->peerip,
					  contact->servport);
      break;
      }
    contact=eml_closecontact(contact);
    (void) jrl_journalling(false,EVTRT,"Unable to make SMTP contact with <%s> port '%s'",
			      		mxlst[i]->mxname,smtpport);
    }
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to know about the email to be	*/
/*	send size.				*/
/*						*/
/************************************************/
static u_long getmailsize(QUETYP *queue)

{
u_long msize;
char *datafile;
struct stat mailstat;

msize=(u_long)0;
datafile=que_getdatafilename(d_transit,queue);
if (stat(datafile,&mailstat)<0) 
  (void) clm_alert(0,"%s Unable to get file <%s> stat infomation (error=<%s> bug?)",
		     "modtrt.c:getmailsize",datafile,strerror(errno));
else
  msize=(u_long)mailstat.st_size;
(void) free(datafile);
return msize;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send Email contents on	*/
/*	connected SMTP channel.			*/
/*						*/
/************************************************/
static int senddata(CTTYP *contact,QUETYP *queue)

{
#define	TBUF	2048

int status;
u_long totbytes;
u_long startmilli;
u_long stopmilli;
FILE *fdata;
char *buffer;
char *datafile;
int phase;
int proceed;

status=ERRPROC;
totbytes=(u_long)0;
startmilli=(u_long)0;
stopmilli=(u_long)0;
fdata=(FILE *)0;
buffer=(char *)0;
datafile=que_getdatafilename(d_transit,queue);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((fdata=fopen(datafile,"r"))==(FILE *)0) {
	status=FAILED;
	contact->info=clm_freestr(contact->info);
	(void) asprintf(&(contact->info),"unable to open file <%s> (error=<%s>)",datafile,strerror(errno));
	proceed=false;	/*Big Trouble		*/
	}
      break;
    case 1	:	/*is remote ready to get*/
      if (eml_sendcmd(contact,DATAOK,"%s",DATA)!=DATAOK) 
	phase=999;	/*get out fast		*/
      break;
    case 2	:	/*sending file contents	*/
      (void) clm_pushdebug(0);	/*no debug level in data mode	*/
      startmilli=clm_getmillisec();
      if ((buffer=(char *)calloc(TBUF+1,sizeof(char)))!=(char *)0) {
	while (fgets(buffer,TBUF,fdata)!=(char *)0) {
          register u_int taille;
	
	  taille=strlen(buffer);
          if (taille>0) {
	    register char *ptr;

	    if (buffer[0]=='.')	/*doubling '.'	*/
	      (void) tls_write(contact->tls,buffer,1);
	    ptr=&(buffer[taille-1]);
	    if ((*ptr=='\n')&&(*(ptr-1)!='\r')) {
	      (void) strcpy(ptr,"\r\n");
	      taille++; 
	      }
            (void) tls_write(contact->tls,buffer,taille);
	    totbytes+=taille;
	    }
	  }
        (void) free(buffer);
        }
      stopmilli=clm_getmillisec();
      (void) clm_popdebug();
      break;
    case 3	:	/*report data-transfer	*/
      {
      char *strloc;

      (void) asprintf(&strloc,"Transmitted %lu bytes within %5.3f seconds ",
			       totbytes,(float)(stopmilli-startmilli)/1000.0);
      (void) jrl_journalling(false,EVTRT,"%s",strloc);
      strloc[strlen(strloc)-1]='\n';
      (void) tls_traceout(contact->tls,"#--",strloc,strlen(strloc));
      (void) free(strloc);
      }
      break;
    case 4	:	/*asking acknowledge	*/
      status=eml_sendcmd(contact,CMDOK,"%s",DATAEND);
      break;
    default	:	/*SAFE guard		*/
      (void) fclose(fdata);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(datafile);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send queued email on SMTP	*/
/*	channel					*/
/*						*/
/************************************************/
static void sendemail(SQLFILE *sql,QUETYP **q,CTTYP *contact)

{
int nxt;
int batch;

nxt=0;
batch=nxt;
while (q[nxt]!=(QUETYP *)0) {
  char *session;
  u_long msize;
  int fromstat;
  int num;

  session=q[nxt]->session;
  msize=getmailsize(q[nxt]);
  fromstat=ERRPROC;
  num=0;
  if (nxt>0) {
    if (eml_sendcmd(contact,CMDOK,"%s","RSET")!=CMDOK) {
      int i;

      for (i=batch;q[i]!=(QUETYP *)0;i++) {
	q[i]->status=ERRPROC;
        q[i]->why=strdup(contact->info);
	}
      break;		/*no need to go further	*/
      }
    }
  (void) jrl_journalling(false,EVTRT,"%s %s",TRANSIT,session);
  (void) jrl_journalling(false,EVTRT,"From: <%s>",q[nxt]->orgeml);
  (void) jrl_journalling(false,EVTRT,"Subject: %s",q[nxt]->subject);
  fromstat=eml_sendmfrom(contact,q[nxt]->orgeml,msize);
  (void) clm_cleanstring(contact->info);
  while (q[nxt]!=(QUETYP *)0) {
    register int status;

    status=fromstat;
    if (strcmp(q[nxt]->session,session)!=0) 
      break; 
    (void) tracelog(contact->dbgname,q[nxt]);
    if (status==CMDOK) {
      status=eml_sendcmd(contact,CMDOK,"%s: <%s@%s>",RCPT,q[nxt]->recname,q[nxt]->recdom);
      if ((status>=CMDOK)&&(status<(CMDOK+10))) {
	status=CMDOK;
        num++;
	}
      else
        q[nxt]->why=strdup(clm_cleanstring(contact->info));
      }
    else 
      q[nxt]->why=strdup(contact->info);
    q[nxt]->status=status;
    if (status!=CMDOK)
      (void) logqueue(sql,q[nxt]);
    nxt++;
    }
  if (num>0) {
    register int status;
    register int i;

    status=senddata(contact,q[batch]);
    (void) clm_cleanstring(contact->info);
    for (i=batch;i<nxt;i++) {
      if (q[i]->status==CMDOK) {
	register const char *coms;

	coms="Accepted";
	if (status!=CMDOK)
	  coms="In trouble";
	q[i]->why=clm_freestr(q[i]->why);
	(void) asprintf(&(q[i]->why),"<%s@%s> %s:\n%s",q[i]->recname,q[i]->recdom,coms,contact->info);
	(void) logqueue(sql,q[i]);
	}
      }
    }
  batch=nxt;
  } 
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to quarantine an E-mail in the*/
/*	user directory within the right domain	*/
/*	Return result within the Queue element.	*/
/*						*/
/************************************************/
static void quaremail(QUETYP *q)

{
#define OPST	"modtrt.c:quaremail,"

int done;
char *newfile;
char *mailfile;
char *newdir;
int phase;
int proceed;

done=false;
(void) asprintf(&newdir,"%s/%s/%s/%s",clm_appdir(d_mails),q->recdom,q->recname,QUARDIR);
(void) asprintf(&newfile,"%s/%s",newdir,q->session);
mailfile=que_getdatafilename(d_transit,q);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*need to create newdir	*/
      (void) clm_createdirs(newdir,0755);
      break;
    case 1	:	/*moving  filename	*/
      done=true;
      if (link(mailfile,newfile)<0) {
	done=false;		/*oups trouble		*/
	switch (errno) {
	  case EXDEV	:
	    done=clm_dupfile(mailfile,newfile);
	    break;
	  default	:
	    q->status=CMDOK;
            (void) asprintf(&(q->why),"unable to move file <%s> to <%s> (error=<%s>)",
				       mailfile,newfile,strerror(errno));
	    break;
	  }
	}
      if (done==true) {
	q->status=CMDOK;
        (void) asprintf(&(q->why),"mail now available in <%s@%s> quarantine box",q->recname,q->recdom);
	}
      break;
    case 2	:	/*set quar file mode	*/
      if (chmod(newfile,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)<0) 
	(void) clm_alert(0,"%s unable to change file <%s> permission mode (error=<%s>)",
			    OPST,newfile,strerror(errno));
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(mailfile);
(void) free(newfile);
(void) free(newdir);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do check if recipient	*/
/*	domain need to be overridden		*/
/*						*/
/************************************************/
static void checkdomover(SQLFILE *sql,QUETYP *queue)

{
char *dom;

if ((dom=dbc_isitadomainalias(sql,queue->recdom))!=(char *)0) {
  (void) jrl_journalling(false,EVTRT,"Domain override from <%s@%s> to <%s@%s>",
				      queue->recname,queue->recdom,queue->recname,dom);
  (void) free(queue->recdom);
  queue->recdom=dom;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email local quarantine	*/
/*	storage					*/
/*						*/
/************************************************/
static void localquarantine(SQLFILE *sql,QUETYP **queues)

{
char *session;
int i;

session="";
for (i=0;queues[i]!=(QUETYP *)0;i++) {
  if (strcmp(session,queues[i]->session)!=0) {
    session=queues[i]->session;
    (void) jrl_journalling(false,EVTRT,"%s %s",TRANSIT,queues[i]->session);
    (void) jrl_journalling(false,EVTRT,"From: <%s>",queues[i]->orgeml);
    (void) jrl_journalling(false,EVTRT,"Subject: %s",queues[i]->subject);
    }
  (void) checkdomover(sql,queues[i]);
  (void) quaremail(queues[i]);
  (void) jrl_journalling(false,EVTRT,"%s",queues[i]->why);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email local delivery	*/
/*	using the delivery shell.		*/
/*						*/
/************************************************/
static void localdelivery(SQLFILE *sql,QUETYP **queues)

{
#define	MXBXSOR	"Unable to store mail in <%s> email mailbox (status=%d), Sorry."

char *script;
char *session;
int i;

(void) asprintf(&script,"%s%s/%s",clm_appdir(d_usrlib),SHELL,DELIVER);
session="";
for (i=0;queues[i]!=(QUETYP *)0;i++) {
  char *filename;
  char *orgeml;
  char *rcpt;
  char *cmd;
  int status;

  if (strcmp(session,queues[i]->session)!=0) {
    session=queues[i]->session;
    (void) jrl_journalling(false,EVTRT,"%s %s",TRANSIT,queues[i]->session);
    (void) jrl_journalling(false,EVTRT,"From: <%s>",queues[i]->orgeml);
    (void) jrl_journalling(false,EVTRT,"Subject: %s",queues[i]->subject);
    }
  (void) checkdomover(sql,queues[i]);
  filename=que_getdatafilename(d_transit,queues[i]);
  orgeml=strdup(queues[i]->orgeml);
  (void) eml_standardize(orgeml);
  (void) asprintf(&rcpt,"%s@%s",queues[i]->recname,queues[i]->recdom);
  (void) asprintf(&cmd,"%s %s \"%s\" \"%s\"",script,filename,orgeml,rcpt);
  switch (status=system(cmd)) {
    case -1		:	/*big system trouble	*/
      queues[i]->status=FAILED;
      (void) asprintf(&(queues[i]->why),"system trouble to store mail in <%s> email mailbox",rcpt);
      (void) jrl_journalling(false,EVTRT,"Bug! %s",queues[i]->why);
      break;
    case  0		:	/*everything fine	*/
      queues[i]->status=CMDOK;
      (void) asprintf(&(queues[i]->why),"mail now available in <%s> email mailbox",rcpt);
      (void) jrl_journalling(false,EVTRT,"%s",queues[i]->why);
      break;
    default		:	/*status trouble	*/	
      queues[i]->status=CMDBAD;
      (void) asprintf(&(queues[i]->why),"Unable to store mail in <%s> in mailbox (status=%d)",rcpt,status);
      (void) jrl_journalling(false,EVTRT,"WARNING! %s",queues[i]->why);
      break;
    }
  (void) free(cmd);
  (void) free(rcpt);
  (void) free(orgeml);
  (void) free(filename);
  }
(void) free(script);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare an Email to be	*/
/*	bounced back to originator.		*/
/*						*/
/************************************************/
static void rebouncing(SQLFILE *sql,QUETYP *queue)

{
#define OPRB	"modtrt.c:rebouncing,"

char *ptr;
char *ext;
char *destination;
MMOENU mode;
char *recipient;
char *dataname;
char *mailname;
char *rejectto;
char *bouncename;
char *subject;
char **explain;
QUETYP *qloc;
int phase;
int proceed;

ptr=(char *)0;
ext="RJCT";
destination=WITHMX;
mode=mm_rebounce;
recipient=(char *)0;
dataname=(char *)0;
mailname=(char *)0;
rejectto=(char *)0;
bouncename=(char *)0;
subject=(char *)0;
explain=(char **)0;
qloc=(QUETYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*rebounce of rebounce? */
      recipient=strdup(queue->rejectto);
      (void) asprintf(&rejectto,"%s@%s",POSTM,eml_getdomain(recipient));
      if (queue->mode==mm_rebounce) {
	ext="LASTHOPE";
	mode=mm_lasthope;
	(void) free(rejectto);
	rejectto=strdup("");
	}
      (void) asprintf(&dataname,"%s-%s",queue->session,ext);
      (void) asprintf(&bouncename,"%s/%s",clm_appdir(d_transit),dataname);
      mailname=que_getdatafilename(d_transit,queue);
      (void) asprintf(&subject,"Error report about your Email (%s)",queue->subject);
      break;
    case 1	:	/*preparing new queue	*/
      if ((ptr=strchr(recipient,'@'))!=(char *)0) {
        *ptr='\000';
	qloc=que_mkqueue(ANYADDR,destination,SMTPPORT,dataname,0,subject,rejectto,rejectto,recipient,ptr+1,mode,(int)m_toforward);
	qloc->transit=que_getuniqtransit();
	}
      (void) free(rejectto);
      break;
    case 2	:	/*setting data in queue	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	QUETYP *qs[2];

	qs[0]=qloc;
        qs[1]=(QUETYP *)0;
        if (dbc_createqueues(sql,qs)==0) {
	  (void) clm_alert(0,"%s Unable to update database with queue (bug!)",OPRB);
	  qloc=que_freequeue(qloc); /*we are really in big trouble	*/
	  }
	}
      if (qloc==(QUETYP *)0) {
	(void) clm_alert(0,"%s Unable to make rebounce to originator <%s> (queue session=<%s-%d>)",
			    OPRB,queue->orgeml,queue->session,queue->transit);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 3	:	/*preparing attachment	*/
      explain=(char **)vec_addlstlst((void **)0,(void *)strdup(queue->why));
      if (eml_attache(mailname,bouncename,queue->session,NOREPLY,queue->orgeml,subject,explain)==false) {
	(void) jrl_journalling(false,EVTRT,"%s Unable to attache orginal mail <%s-%d> (bug?)",
					   OPRB,queue->session,queue->transit);
        (void) unlink(bouncename);	/*back tracking	*/
	(void) lvl_delqueue(sql,qloc);	/*removing queue*/
	qloc=que_freequeue(qloc);
	phase=999;		/*getting out fast	*/
	}
      explain=(char **)vec_freelstlst((void **)explain,VECFREE);
      break;
    case 4	:	/*preparing attachment	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	char *cmt;
	int ready;
	char *lfile;
	char *newshead;

	cmt="";
	ready=true;
        (void) asprintf(&newshead,"Set rebounce report %s to <%s@%s>",
				   qloc->session,qloc->recname,qloc->recdom);
	lfile=que_getdatafilename(d_transit,qloc);
	if (link(bouncename,lfile)<0) {
	  (void) clm_alert(0,"%s Unable to link file <%s> to file <%s> (errno=<%s>))",
			      OPRB,bouncename,lfile,strerror(errno));
	  cmt=" not queued (trouble?)";
	  ready=false;
	  }
	(void) clm_fasprintf(&newshead,"\n\t\tabout <%s@%s>",
					queue->recname,queue->recdom,cmt);
	(void) free(lfile);
	if (ready==true)
          (void) dbc_rdyqueue(sql,qloc);
	else
          (void) lvl_delqueue(sql,qloc);
	(void) jrl_journalling(false,EVTRT,"%s",newshead);
        (void) unlink(bouncename);
	}
      (void) jrl_journalling(true,EVTRT,(char *)0);
      break;
    case 5	:	/*reporting the event	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	DSPTYP dsp;
	RCPTYP rcpt;
	RCPTYP *rcpts[2];

	dsp.session=qloc->session;
	dsp.peerip=LOCALIP;
	dsp.peername=LOCALHOST;
	dsp.locip=LOCALIP;
	dsp.subject=qloc->subject;
	dsp.inbehalf=POSTM;
	dsp.efrom=POSTM;
	dsp.hfrom=POSTM;
	dsp.mtaille=1000;
	rcpt.status=m_relayed;
	rcpt.inbehalf=POSTM;
	(void) asprintf(&(rcpt.rcpt),"%s@%s",qloc->recname,qloc->recdom);
	rcpts[0]=&rcpt;
	rcpts[1]=(RCPTYP *)0;
	(void) dbc_setlogs(sql,&dsp,rcpts);
	(void) free(rcpt.rcpt);
        (void) jrl_journalling(false,EVJRL,"%-*s%s -> Rebouncing start",6,SESSION,qloc->session);
        (void) jrl_journalling(false,EVJRL,"%s: <%s>",MAILF,qloc->rejectto);
        (void) jrl_journalling(false,EVJRL,"%s: <%s@%s>",RCPT,qloc->recname,qloc->recdom);
        (void) jrl_journalling(false,EVJRL,"%s: %s","Subject",qloc->subject);
        (void) jrl_journalling(false,EVJRL,"Rebouncing end");
        (void) jrl_journalling(true,EVJRL,(char *)0);
        qloc=que_freequeue(qloc);
	}
      break;
    default	:	/*SAFE guard		*/
      queue->status=BOUNCE;	/*queue bounced	*/
      (void) free(subject);
      (void) free(bouncename);
      mailname=clm_freestr(mailname);
      (void) free(dataname);
      (void) free(recipient);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transit E-mail to remote	*/
/*	SMTP server.				*/
/*						*/
/************************************************/
static void transitsmtp(SQLFILE *sql,QUETYP **queues)

{
#ifdef	DEBUG	
#define MXTIME	 15	/*15 to contact in debug*/
#else
#define MXTIME	180	/*180s to make contact	*/
#endif
#define	NEXTRY	15*60	/*Next try in 15 minutes*/
MXTYP **mxlst;
CTTYP *contact;
int phase;
int proceed;

mxlst=(MXTYP **)0;
contact=(CTTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets prepare next try	*/
      if (queues!=(QUETYP **)0) {
	register time_t curtime;
	register time_t nexttime;
	int i;

	curtime=time((time_t *)0);
	nexttime=curtime+NEXTRY;
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  queues[i]->lasttime=curtime;
	  queues[i]->nexttime=nexttime;
	  queues[i]->numtrace=que_getuniqtransit();
	  (void) dbc_updqueue(sql,queues[i]);
	  }
	}
      break;
    case 1	:	/*is E-mail relayed	*/
     if (strstr(queues[0]->smtpmx,WITHMX)!=queues[0]->smtpmx) {
       MXTYP *mx;

       mx=(MXTYP *)calloc(1,sizeof(MXTYP));
       mx->preference=1;
       mx->mxname=strdup(queues[0]->smtpmx);
       mxlst=(MXTYP **)vec_addlstlst((void *)0,(void *)mx);
       phase++;		/*no need to look for MX*/
       }
      break;
    case 2	:	/*getting serveur list	*/
      if ((mxlst=dns_getmxorip(queues[0]->recdom))==(MXTYP **)0) {
	(void) setwhy(queues,BADMX,"Unable to get an MX or an IP for domain <%s>",queues[0]->recdom);
	proceed=false;
	}
      break;
    case 3	:	/*looking for one server*/
      (void) jrl_journalling(false,EVTRT,"SMTP Transit to domain %s",queues[0]->recdom);
      contact=makecontact(mxlst,queues[0]->smtpport,MXTIME);
      break;
    case 4	:	/*reporting trouble	*/
      if ((contact==(CTTYP *)0)||(contact->established==false)) {
	int nummx;
        int i;

	nummx=vec_sizelstlst((void **)mxlst);
	if (contact==(CTTYP *)0) {
	  if (nummx>1)
	    (void) setwhy(queues,MISSMX,
			         "Unable to make SMTP contact with any of the %d MX for domain "
			         "<%s> within '%d' seconds",
		 		  nummx,queues[0]->recdom,MXTIME);
	  else
	    (void) setwhy(queues,NOTGOOD,"Unable to make SMTP contact with domain <%s> MX",queues[0]->recdom);
	  }
	else {
	  (void) setwhy(queues,FAILED,"contact with domain <%s> MX returned <%s>",queues[0]->recdom,contact->info);
	  }
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  queues[i]->status=INRELAY;
	  (void) logqueue(sql,queues[i]);
	  }
	phase=999;	/*going out quickly	*/
	}
      break;
    case 5	:	/*sending Email		*/
      (void) sendemail(sql,queues,contact);
      break;
    default	:	/*SAFE guard		*/
      contact=eml_closecontact(contact);
      mxlst=dns_cleanmx(mxlst);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transit received Email,	*/
/*	either store the Email localy, or	*/
/*	send it to remote SMTP server		*/
/*						*/
/************************************************/
static void dotransit(SQLFILE *sql,QUETYP **queues)

{
typedef enum	{
	a_locdel,	/*local delivery	*/
	a_locquar,	/*local quarantine	*/
	a_smtp		/*smtp relaying/delivery*/
	}LOCTYP;

static VOCTYP vocable[]={
        {a_locdel,LOCALDEL,(const void *)0},
        {a_locquar,LOCALQUAR,(const void *)0},
        {a_smtp,(char *)0,(const void *)0}
        };

VOCTYP *voc;
char *smtpmx;

smtpmx=strdup(queues[0]->smtpmx);
voc=clm_getvoca(vocable,smtpmx);
switch(voc->code) {
  case a_locdel	:	/*Local delivery	*/
    (void) localdelivery(sql,queues);
    break;
  case a_locquar:	/*Local quarantibe	*/
    (void) localquarantine(sql,queues);
    break;
  case a_smtp	:	/*smtp delivery/relay	*/
    (void) transitsmtp(sql,queues);
    break;
  }
(void) free(smtpmx);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to rebounce Email to		*/
/*	originator if Email got a reject final	*/
/*	status.					*/
/*						*/
/************************************************/
static void dorebounce(SQLFILE *sql,QUETYP **queues)

{
if (queues!=(QUETYP **)0) {
  int i;

  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    if (queues[i]->status!=CMDOK) {
      if (queues[i]->upto<queues[i]->nexttime) {
	queues[i]->status=FAILED;
	queues[i]->why=clm_freestr(queues[i]->why);
	queues[i]->why=eml_ascstat(queues[i]->status,"Unable to transmit email in due time");
        (void) logqueue(sql,queues[i]);
	}
      }
    if (queues[i]->status<TMPERR)
      continue;
    switch (queues[i]->mode) {
      case (mm_std)	:
      case (mm_rebounce):
	if (queues[i]->status>=SYNERR) {
          (void) logqueue(sql,queues[i]);
          (void) rebouncing(sql,queues[i]);
	  }
	else
	  queues[i]->status=RETRY;
	break;
      default		:
	break;
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email quarantine recall	*/
/*						*/
/************************************************/
static int quarantinerecall(SQLFILE *sql,char *script,EXTTYP *ext)

{
#define	OPQU	"modtrt.c:quarantinerecall,"

int done;
char *user;
char *domain;
char *ptr;
char *filename;
char *cmd;
int status;

done=false;
(void) jrl_journalling(false,EVTRT,"Quarantined ID %s/%s",ext->session,ext->recipient);
user=strdup(ext->recipient);
domain=eml_getdomain(user);
if ((ptr=strchr(user,'@'))!=(char *)0)
  *ptr='\000';
(void) asprintf(&filename,"%s/%s/%s/%s/%s",clm_appdir(d_mails),domain,user,QUARDIR,ext->session);
(void) asprintf(&cmd,"%s %s %s %s@%s",script,filename,"postmaster",user,domain);
switch (status=system(cmd)) {
  case -1		:	/*big system trouble	*/
    (void) jrl_journalling(false,EVTRT,"Bug! system trouble to recover quarantine");
    break;
  case  0		:	/*everything fine	*/
    (void) jrl_journalling(false,EVTRT,"mail now available in <%s> email mailbox",ext->recipient);
    if (unlink(filename)<0) 
      (void) clm_alert(0,"%s Unable to remove file <%s> (error=<%s>) bug?",OPQU,filename,strerror(errno));
    (void) dbc_updlogs(sql,ext->session,ext->recipient,m_direct);
    done=true;
    break;
  default		:	/*status trouble	*/	
    (void) jrl_journalling(false,EVTRT,"WARNING! Unable to store quarantine to <%s> email mailbox (status=%d)",
					ext->recipient,WEXITSTATUS(status));
     break;
  }
(void) free(cmd);
(void) free(filename);
(void) free(user);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email quarantine recall	*/
/*	when the email must be sent remote	*/
/*						*/
/************************************************/
static int stayputrecall(SQLFILE *sql,EXTTYP *ext)

{
#define	OPSR	"modtrt.c:stayputrecall,"

int done;
DSPTYP *dsp;
RCPTYP **rcpts;
GRPTYP *group;
char *infile;
int phase;
int proceed;

done=false;
dsp=(DSPTYP *)0;
rcpts=(RCPTYP **)0;
group=(GRPTYP *)0;
infile=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting DSP		*/
      if ((dsp=dbc_getsessiondsp(sql,ext->session,ext->recipient))==(DSPTYP *)0){
        (void) clm_alert(0,"%s Unable to get data session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	proceed=false;
	}
      break;
    case 1	:	/*making RCPT		*/
      if (dsp!=(DSPTYP *)0) { /*always		*/
	RCPTYP *rcpt;

        rcpt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
        rcpt->rcpt=strdup(ext->recipient);
        rcpts=(RCPTYP **)vec_addlstlst((void *)rcpts,(void *)rcpt);
	}
      break;
    case 2	:	/* getting group	*/
      if ((group=dbc_getgroup(sql,ext->recipient))==(GRPTYP *)0) {
        (void) clm_alert(0,"%s Unable to find group for session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	phase=999;
	}
      break;
    case 3	:	/*preparing infile	*/
      infile=eml_getquarfilename(dsp->session,ext->recipient);
      break;
    case 4	:	/*making transfer rqst	*/
      done=true;
      if (trt_transit(sql,group->exitip,group->smtprmt,group->smtpport,dsp->hfrom,dsp,infile,rcpts)==0) {
        (void) clm_alert(0,"%s Unable to recall quarantine for session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	done=false;
	}
      break;
    default	:	/*SAFE Guard		*/
      infile=clm_freestr(infile);
      group=dbc_freegroup(group);
      rcpts=eml_freercpts(rcpts);
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check AND send an Email	*/
/*	to remote site				*/
/*						*/
/************************************************/
static QUETYP **tracker(QUETYP **queues)

{
#define	OPSE	"modtrt.c:tracker,"
#define	TRACKER	"%s Tracker (%04lu)"

u_int batch;
SQLFILE *sql;
const char *empty;
int phase;
int proceed;

sql=(SQLFILE *)0;
empty=(const char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database",OPSE);
	phase=999;	/*TROUBLE		*/
	}
      break;
    case 1	:	/*displaying process	*/
      batch=vec_sizelstlst((void **)queues);
      (void) prc_settitle(TRACKER,appname,(u_long)batch);
      break;
    case 2	:	/*transiting Email	*/
      (void) dotransit(sql,queues);
      break;
    case 3	:	/*check for rebounce	*/
      (void) dorebounce(sql,queues);
      break;
    case 4	:	/*final queue update	*/
      if (queues!=(QUETYP **)0) {
	int nxt;

	nxt=0;
	while (queues[nxt]!=(QUETYP *)0) {
	  (void) dbc_settrtdata(sql,queues[nxt]);
	  nxt++;
	  }
	}
      queues=trt_cleanqueues(sql,queues);
      break;
    default	:	/*SAFE guard		*/
      sql=sql_close(sql);
      (void) jrl_journalling(true,EVTRT,(char *)0);
      proceed=false;
      break;
    }
  phase++;
  if (sigterm==true) 
    phase=999;		/*going out fast	*/
  }
return queues;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan directories extract	*/
/*	smtp file to be forwarded.		*/
/*						*/
/************************************************/
static SQLFILE *scanqueues(SQLFILE *sql,PRCTYP **prclst,u_long *sofar)

{
#define	OPSC	"modtrt.c:scanqueues,"

#define MXQUE	20	/*maximun queue crunch size	*/
#define	MXFORK	10	/*maximun fork number by level	*/
#define	SNDTTL	"%s Transitr sending <%s>"

QUETYP **queues;
u_int batch;
const char *empty;
int phase;
int proceed;

queues=(QUETYP **)0;
batch=0;
empty=(const char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*child still working	*/
      *prclst=prc_childcheck(*prclst);
      if ((*prclst!=(PRCTYP *)0)&&((*prclst)->nbr>=MXFORK))
	phase=999;	/*busy, see you later	*/
      break;
    case 1	:	/*something in queue?	*/
      if ((queues=dbc_getqueues(sql,MXQUE))==(QUETYP **)0) 
	phase=999;	/*no need to go further	*/
      break;
    case 2	:	/*there is something	*/
      batch=vec_sizelstlst((void **)queues);
      (void) prc_settitle(TITLE,appname,(u_long)batch,*sofar);
      sql=sql_close(sql);
      if (queues!=(QUETYP **)0) {	/*always*/
        QUETYP **dest;
 
        while ((dest=sortdest(queues))!=(QUETYP **)0) {
	  pid_t transpid;

	  switch (transpid=fork()) {
	    case -1	:	/*trouble!	*/
	      (void) clm_alert(0,"%s Unable to fork transit process (error=<%s>)",OPSC,strerror(errno));
	      break;
	    case  0	:	/*child process	*/
	      (void) sql_close(sql);
	      (void) dbg_setnoleak(1);
	      (void) dbg_dumpmem(appname,"sender");
	      dest=tracker(dest);
	      (void) sleep(1);
	      (void) exit(0);
	      break;
	    default	:	/*parent process*/
	      *prclst=prc_addtoprc(*prclst,transpid);
	      break;
	    }
          dest=que_freequeues(dest);
	  }
  	(void) free(queues); 
	*sofar+=batch;
        (void) prc_settitle(TITLE,appname,(u_long)0,*sofar);
	}
      break;
    case 3	:	/*reconnecting		*/
      (void) sleep(1);
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to re-open database",OPSC);
	phase=999;	/*BIG TROUBLE		*/
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return sql;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract Email to be	*/
/*	extracted from quarantine.		*/
/*						*/
/************************************************/
static void doextract(SQLFILE *sql)

{
#define	OPEX	"modtrt.c:doextract,"

EXTTYP **exts;

exts=(EXTTYP **)0;
if ((exts=dbc_getextract(sql))!=(EXTTYP **)0) {
  char *script;
  int i;

  (void) asprintf(&script,"%s%s/%s",clm_appdir(d_usrlib),SHELL,DELIVER);
  for (i=0;exts[i]!=(EXTTYP *)0;i++) {
    LOGTYP *emlog;
    int done;

    emlog=(LOGTYP *)0;
    done=false;
    switch (exts[i]->status) {
      case m_quarantine	:
        done=quarantinerecall(sql,script,exts[i]);
	break;
      case m_stayput	:
        done=stayputrecall(sql,exts[i]);
	break;
      default		:
	(void) clm_alert(0,"%s Unexpected status '%d' for quarantine recall session=<%s> recipient=<%s> (Bug?)",
			   "modtrt.c:doextract,",exts[i]->status,exts[i]->session,exts[i]->recipient);
	break;
      }
    if (done==false)
      continue;
    if ((emlog=dbc_getlogs(sql,exts[i]->session,exts[i]->recipient))!=(LOGTYP *)0) {
      char *orgdom;

      orgdom=eml_getdomain(emlog->hfrom);
      if (dbc_setoverride(sql,emlog->recipient,"*",orgdom,emlog->rmtip,16,tr_accepted)==false) {
        (void) clm_alert(0,"%s Unable to set override recipient <%s> origin=<%s> rmtip=<%s> (bug?)",
			   OPEX,emlog->recipient,emlog->hfrom,emlog->rmtip);
	}
      emlog=dbc_freelogs(emlog);
      }
    }
  (void) free(script);
  exts=dbc_freeextract(exts);
  (void) jrl_journalling(true,EVTRT,(char *)0);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for email transit	*/
/*	order.					*/
/*						*/
/************************************************/
static void scanner()

{
#define OPSA	"modtrt.c:scanner,"


const char *empty;
time_t curtime;
time_t lastclean;
u_long sofar;
PRCTYP *prclst;
SQLFILE *sql;
int phase;
int proceed;

empty=(const char *)0;
curtime=(time_t)0;
lastclean=(time_t)0;
sofar=(u_long)0;
prclst=(PRCTYP *)0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database",OPSA);
	phase=999;	/*TROUBLE		*/
	}
      break;
    case 1	:	/*cleaning		*/
      curtime=time((time_t *)0);
      if ((lastclean+60)<curtime) {
	(void) dbc_dbcleanup(sql);
	lastclean=curtime;
	}
      break;
    case 2	:	/*doing the real job	*/
      if ((sql=scanqueues(sql,&prclst,&sofar))==(SQLFILE *)0)
	phase=999;	/*trouble trouble	*/
      break;
    case 3	:	/*extracting data	*/
      (void) doextract(sql);
      (void) sleep(2);	/*lets rest a little bit*/
      phase=0;		/*lets loop		*/
      break;
    default	:	/*SAFE Guard		*/
      prclst=prc_freeprc(prclst);
      proceed=false;
      break;
    }
  phase++;
  if (sigterm==true) {
    sql=sql_close(sql);
    proceed=false;
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to fork a transit process, 	*/
/*	this process whill handle going-out	*/
/*	SMTP queue.				*/
/*						*/
/************************************************/
pid_t trt_cycle()


{
#define	OPTR	"modtrt.c:trt_cycle,"

pid_t trt;

switch (trt=fork()) {
  case -1	:
    (void) clm_alert(0,"%s unable to fork (error=<%s>)",OPTR,strerror(errno));
    break;
  case  0	:
    /*forget about previously allocated memory leak */
    (void) dbg_setnoleak(1);
    (void) clm_setjournal(EVJRL,d_log,"events");
    (void) clm_setjournal(EVTRT,d_log,"transit");
    (void) prc_settitle(TITLE,appname,(u_long)0,(u_long)0);
    (void) scanner();
    (void) jrl_closejournals();
    (void) dbg_dumpmem(appname,"transiter");
    (void) sleep(1);
    (void) exit(0);
    break;
  default	:
    break;
  }
return trt;
}


