/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	UNIQUE:					*/
/*	Take Care of all access to devices	*/
/*	could be system dependent.		*/
/*	Take care about email transit queue 	*/
/*	management.				*/
/*						*/
/************************************************/
#include	<stdlib.h>
#include	<string.h>
#include	"unique.h"
#include	"subclm.h"
#include	"dbgmem.h"
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a very unique	*/
/*	transit number				*/
/*						*/
/************************************************/
u_long que_getuniqtransit()

{
#define	TRTCYCLE 1000

static u_long transit;

transit++;
transit%=TRTCYCLE;
return (getpid()*TRTCYCLE)+transit;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the queue trace	*/
/*	file.					*/
/*						*/
/************************************************/
char *que_tracename(QUETYP *queue)

{
char *tracename;

(void) asprintf(&(tracename),"%s/%s-t%lu",clm_appdir(d_tobesend),queue->session,queue->numtrace);
return tracename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a dataname build	*/
/*	around session ID and transit number	*/
/*						*/
/************************************************/
char *que_getdatafilename(DIRENUM dir,QUETYP *queue)

{
char *datafilename;

(void) asprintf(&datafilename,"%s/%s-%05lu",clm_appdir(dir),queue->session,queue->transit);
return datafilename;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create a email transit	*/
/*	queue entry.				*/
/*						*/
/************************************************/
QUETYP *que_mkqueue(char *outip,char *smtpmx,char *smtpport,char *session,int transit,
		    char *subject,char *rejectto,char *orgeml,char *recname,char *recdom,MMOENU mode,int orgstatus)

{
QUETYP *q;

q=(QUETYP *)calloc(1,sizeof(QUETYP));
q->outip=strdup(outip);
q->smtpmx=strdup(smtpmx);
q->smtpport=strdup(smtpport);
q->session=strdup(session);
q->transit=transit;
q->mode=mode;
q->orgstatus=orgstatus;
q->subject=strdup(subject);
if (rejectto!=(char *)0) 
  q->rejectto=strdup(rejectto);
else
  q->rejectto=strdup(orgeml);
q->orgeml=strdup(orgeml);
q->recname=strdup(recname);
q->recdom=strdup(recdom);
return q;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	transit queue entry.			*/
/*						*/
/************************************************/
QUETYP *que_freequeue(QUETYP *queue)

{
if (queue!=(QUETYP *)0) {
  queue->why=clm_freestr(queue->why);
  queue->recdom=clm_freestr(queue->recdom);
  queue->recname=clm_freestr(queue->recname);
  queue->orgeml=clm_freestr(queue->orgeml);
  queue->rejectto=clm_freestr(queue->rejectto);
  queue->subject=clm_freestr(queue->subject);
  queue->session=clm_freestr(queue->session);
  queue->smtpport=clm_freestr(queue->smtpport);
  queue->smtpmx=clm_freestr(queue->smtpmx);
  queue->outip=clm_freestr(queue->outip);
  (void) free(queue);
  queue=(QUETYP *)0;
  }
return queue;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a 	*/
/*	transite email queue list		*/
/*						*/
/************************************************/
QUETYP **que_freequeues(QUETYP **queues)

{
if (queues!=(QUETYP **)0) {
  int i;
  
  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    queues[i]=que_freequeue(queues[i]);
    }
  (void) free(queues);
  queues=(QUETYP **)0;
  }
return queues;
}
