/************************************************/
/*						*/
/*	UNIRLY					*/
/*	Define relaying authorisation module.	*/
/*	All procedure involved to allow an Email*/
/*	relaying are located in this module	*/
/*						*/
/************************************************/
#include	<sys/stat.h>
#include	<arpa/inet.h>
#include	<errno.h>
#include	<time.h>
#include	<string.h>
#include	"dbgmem.h"
#include	"lowtyp.h"
#include	"basafn.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unirly.h"

#define	RLYENV	"RELAY"	/*Relay file ENV var	*/
#define	RLYLST	"alwaysrelay:popbsmtp"

typedef	struct	{
	AFNTYP *afnnum;	/*IP number		*/
	char *from;	/*originator email addr	*/
	time_t upto;	/*maximun relayable	*/
	}RLYENT;

typedef	struct	{
	char *fullname;	/*file name to scan	*/
	time_t lastscan;/*file lastscan		*/
	RLYENT **relais;/*relais list info	*/
	}RLYFILE;

typedef	struct	{
	time_t lastscan;/*last general scan	*/
	RLYFILE **infos;/*file scanned contents	*/
	}RLYINF;
/*

*/
/************************************************/
/*						*/
/*	Procedure to free memory used by a	*/
/*	relais array.				*/
/*						*/
/************************************************/
static RLYENT **freerelais(RLYENT **relais)

{
if (relais!=(RLYENT **)0) {
  int i;

  for (i=0;relais[i]!=(RLYENT *)0;i++) {
    relais[i]->afnnum=afn_freeipnum(relais[i]->afnnum);
    relais[i]->from=clm_freestr(relais[i]->from);
    (void) free(relais[i]);
    }
  (void) free(relais);
  relais=(RLYENT **)0;
  }
return relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a minimal RLYINF	*/
/*	structure contents.			*/
/*						*/
/************************************************/
static RLYINF *preprelay()

{
RLYINF *relay;
const char *ptr;
char *list;
char **split;
int phase;
int proceed;

relay=(RLYINF *)0;
ptr=(const char *)0;
list=(char *)0;
split=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting file list	*/
      if ((ptr=getenv(RLYENV))==(const char *)0) {
        ptr=RLYLST;
	}
      list=strdup(ptr); 
      break;
    case 1	:	/*do we have a list	*/
      if (strlen(list)==0)
        phase=999;	/*no list! get out quick*/
      break;
    case 2	:	/*preparing the struct	*/
      relay=(RLYINF *)calloc(1,sizeof(RLYINF));
      relay->lastscan=(time_t)0;
      break;
    case 3	:	/*list of file to scan	*/
      if ((split=clm_splitstring(list,':'))!=(char **)0) {
        register const char *etcpath;
	int i;
        
	etcpath=clm_appdir(d_etc);
	for (i=0;split[i]!=(char *)0;i++) {
	  RLYFILE *info;
          

	  info=(RLYFILE *)calloc(1,sizeof(RLYFILE));
	  (void) asprintf(&(info->fullname),"%s/%s",etcpath,split[i]);
          info->lastscan=(time_t)0;
	  relay->infos=(RLYFILE **)vec_addlstlst((void **)relay->infos,(void *)info);
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      split=(char **)vec_freelstlst((void **)split,VECFREE);
      (void) free(list);
      proceed=false;
      break;
    }
  phase++;
  }
return relay;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update one relay info list	*/
/*						*/
/************************************************/
static RLYFILE *updateinfos(RLYFILE *infos)

{
#define	OPUP	"unirly.c:updateinfos,"

#define	MXB	1024
#define	IP4MASK	32

time_t updtime;
FILE *fichier;
char buffer[MXB];
int phase;
int proceed;

updtime=(time_t)0;
fichier=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Was the file updated?	*/
      if (infos->fullname!=(char *)0) {
	struct stat bufstat;

	if (stat(infos->fullname,&bufstat)<0) {
	  (void) clm_alert(1,"%s Unable to get stat for <%s>, (error=<%s>)",
			     OPUP,infos->fullname,strerror(errno));
          infos->relais=freerelais(infos->relais);
	  proceed=false;/*trouble trouble	*/
	  }
	else {
	  if (bufstat.st_mtime<=infos->lastscan) {
	    updtime=bufstat.st_mtime;
	    phase=999;	/*no need to go further	*/
	    }
	  }
        }
      break;
    case 1	:	/*cleaning infos record	*/
      infos->relais=freerelais(infos->relais);
      break;
    case 2	:	/*opening file		*/
      if ((fichier=fopen(infos->fullname,"r"))==(FILE *)0) {
	(void) clm_alert(1,"%s Unable to open file <%s>, (error=<%s>)",
			    OPUP,infos->fullname,strerror(errno));
	phase=999;	/*no need to go further	*/
	}
      break;
    case 3	:	/*reading file		*/
      while (clm_getstr(fichier,buffer,sizeof(buffer),'#')!=(char *)0) {
	time_t upto;
        char *from;
	char ip[50];
	int mask;
	int num;
	char strip[MXB];
	char strfrom[MXB];
	char strupto[MXB];

	upto=(time_t)0;
        from=(char *)0;
	mask=IP4MASK;
	num=sscanf(buffer,"%s %s %s",strip,strfrom,strupto);
	switch (num)	{	/*order matter	*/
	  case	3	:	/*get upto	*/
	    (void) sscanf(strupto,"%ld",&upto);
				/*NO BREAK!	*/
	  case	2	:	/*get from	*/
	    from=strdup(strfrom);
				/*NO BREAK!	*/
	  case	1	:	/*get IP	*/
	    if (sscanf(strip,"%49[^/]/%d",ip,&mask)>=1) {
	      RLYENT *entry;
	      AFNTYP *afnnum;

	      if ((afnnum=afn_getipnum(ip))==(AFNTYP *)0) {
	        (void) clm_alert(0,"%s Unable to convert IP <%s> in line <%s> from file <%s> (error=<%s>)",
			            OPUP,ip,buffer,infos->fullname,strerror(errno));
	 	break;
		}
	      (void) afn_maskip(afnnum,mask);
	      entry=(RLYENT *)calloc(1,sizeof(RLYENT));
	      entry->afnnum=afnnum;
	      entry->from=from;
	      entry->upto=upto;
	      infos->relais=(RLYENT **)vec_addlstlst((void **)infos->relais,(void *)entry);
	      }
	    break;
	  case  0	:	/*trouble!	*/
	    (void) clm_alert(1,"%s Unable to scan line <%s> from file <%s>",
			    OPUP,buffer,infos->fullname);
	    break;
	  }
	}
      break;
    case 4	:	/*closing file		*/
      (void) fclose(fichier);
      infos->lastscan=updtime;
      break;
    default	:	/*SAFE GUARD		*/
      proceed=false;
      break;
    }
  phase++;
  }
return infos;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to load the relay structure	*/
/*						*/
/************************************************/
static int loadrelay(RLYINF **relay)

{
int ready;
time_t curtime;
int phase;
int proceed;

ready=false;
curtime=time((time_t *)0);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*already a struct ready*/
      if (*relay==(RLYINF *)0) {
	*relay=preprelay();
        if (*relay==(RLYINF *)0)
	  proceed=false;
	}
      break;
    case 1	:	/*already a struct ready*/
      if ((*relay)->lastscan>=curtime)
	phase=999;	/*no need to refresh	*/
      break;
    case 2	:	/*lets refresh entries	*/
      if ((*relay)->infos!=(RLYFILE **)0) {
        int i;
	
	for (i=0;(*relay)->infos[i]!=(RLYFILE *)0;i++) {
	  (*relay)->infos[i]=updateinfos((*relay)->infos[i]);
	  }
	}
      (*relay)->lastscan=curtime;
      break;
    default	:	/*SAFE guard		*/
      if ((*relay)->infos!=(RLYFILE **)0)
        ready=true;
      proceed=false;
      break;
    }
  phase++;
  }
return ready;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific	*/
/*	originator is relayable.		*/
/*						*/
/************************************************/
static int isrelayable(RLYFILE *relay,char *from,AFNTYP *afnnum,time_t curtime)

{
int relayable;

relayable=false;
if ((relay!=(RLYFILE *)0)&&(relay->relais!=(RLYENT **)0)) {
  int i;

  for (i=0;relay->relais[i]!=(RLYENT *)0;i++) {
    RLYENT *ptr;
    int phase;
    int proceed;

    ptr=relay->relais[i];
    phase=0;
    proceed=true;
    while (proceed==true) {
      switch (phase) {
	case 0		:	/*checking IP	*/
	  if (afn_cmpipnum(ptr->afnnum,afnnum,ptr->afnnum->afnmask)==false)
	    proceed=false;	/*no luck	*/
	  break;
	case 1		:	/*checking from	*/
	  if ((ptr->from!=(char *)0)&&(strcmp(ptr->from,from)!=0))
	    proceed=false;	/*no luck	*/
	  break;
	case 2		:	/*checking date	*/
	  if ((ptr->upto>0)&&(curtime>ptr->upto))
	    proceed=false;	/*no luck	*/
	  break;
        default		:	/*SAFE guard	*/
	  relayable=true;	/*Bingo!	*/
          proceed=false;
	  break;
	}
      phase++;
      }
    }
  }
return relayable;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to fee memory used by relay	*/
/*	data structure.				*/
/*	is allow to relay an Email		*/
/*						*/
/************************************************/
RLYTYP *rly_freerelais(RLYTYP *relais)

{
if (relais!=(RLYTYP *)0) {
  RLYINF *ptr;

  ptr=(RLYINF *)relais;
  if (ptr->infos!=(RLYFILE **)0) {
    int i;

    for (i=0;ptr->infos[i]!=(RLYFILE *)0;i++) {
      ptr->infos[i]->relais=freerelais(ptr->infos[i]->relais);
      ptr->infos[i]->fullname=clm_freestr(ptr->infos[i]->fullname);
      (void) free(ptr->infos[i]);
      }
    (void) free(ptr->infos);
    }
  (void) free(ptr);
  relais=(RLYTYP *)0;
  }
return relais;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a originator	*/
/*	is allow to relay an Email		*/
/*						*/
/************************************************/
int rly_isrelayable(RLYTYP **relais,char *from,char *peerip)

{
#define	OPLY	"unirly.c:rly_isrelayable,"

int relayable;
AFNTYP *afnnum;
time_t curtime;
int phase;
int proceed;

relayable=false;
afnnum=(AFNTYP *)0;
curtime=time((time_t *)0);
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*converting IP address	*/
      if ((afnnum=afn_getipnum(peerip))==(AFNTYP *)0) {
	(void) clm_alert(0,"%s Unable to scan IP <%s> (error=<%s>)",OPLY,peerip,strerror(errno));
	proceed=false;
	}
      break;
    case 1	:	/*checking relais update*/
      proceed=loadrelay((RLYINF **)relais);
      break;
    case 2	:	/*scanning for relay	*/
      if (((RLYINF *)*relais)->infos!=(RLYFILE **)0) {
        int i;
	
	for (i=0;((RLYINF *)*relais)->infos[i]!=(RLYFILE *)0;i++) {
	  relayable=isrelayable(((RLYINF *)*relais)->infos[i],from,afnnum,curtime);
	  if (relayable==true)
	    break;
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      afnnum=afn_freeipnum(afnnum);
      proceed=false;
      break;
    }
  phase++;
  }
return relayable;
}
